package se;

import java.util.*;

public class AgentFunction extends Function {
	private String name;
	private Tuple args;
	private boolean invertibleFunction;
	private boolean publicFunction;
	private ParametricAgentFunction paramF;
	//structure with
	//[ ((assignment1,constantizedTuple1),constant1) ;...; ((assignmentN,constantizedTupleN),constantN) ]
	//assignments contains Variables while in the corresponding ParametricFunction it'll contain parameters
	private ArrayList<Pair<Pair<ArrayList<Pair<Message,ParametricMessage>>,ParametricTuple>,ParametricConstant>> constantiz;
	
	public AgentFunction(String name, Tuple args){
		this.name=name;
		this.args=args;
		this.invertibleFunction=false;
		this.publicFunction=true;
		this.paramF=new ParametricAgentFunction(this);
	}
	
	public AgentFunction(String name, Tuple args, String messageString){
		this(name, args);
		this.messageString = messageString;
	}
	
	public Evaluable substitute(ArrayList<Pair<Message, Message>> mex_newMex){
		return new AgentFunction(this.name,(Tuple)this.args.substitute(mex_newMex),this.invertibleFunction,this.publicFunction);
	}
	
	public AgentFunction(String name, Tuple args, boolean invertibleFunction, boolean publicFunction){
		this(name,args);
		this.invertibleFunction=invertibleFunction;
		this.publicFunction=publicFunction;
	}
	
	public AgentFunction(String name, Tuple args, boolean invertibleFunction, boolean publicFunction, String messageString){
		this(name, args, invertibleFunction, publicFunction);
		this.messageString = messageString;
	}	
	
	public ArrayList<Pair<String,String>> interpolantFunctionString(Environment env){
		ArrayList<Pair<String,String>> ret=new ArrayList<Pair<String,String>>();
		String tmp="";
		
		for(int i=0;i<this.constantiz.size();i++){
			
			for(int j=0;j<this.constantiz.get(i).getElement0().getElement0().size();j++){
				tmp+="(= "+this.constantiz.get(i).getElement0().getElement0().get(j).getElement0().interpolantString(env)+" "+
						this.constantiz.get(i).getElement0().getElement0().get(j).getElement1().satString();
			}
			
			ret.add(new Pair<String,String>(tmp,this.constantiz.get(i).getElement1().satString()));	
		}
		
		return ret;
	}
	
	

	
	public ParametricAgentFunction eval(Environment env){
		return this.paramF;
	}
	
	public String print(){
		return this.getName()+"("+this.getArgs().print()+")"; 
	}
	
	public ArrayList<ArrayList<Message>> getVars(ArrayList<ArrayList<Message>> list){
		return this.getArgs().getVars(list);	
	}
	
	//gets and sets
	public String getName(){return this.name;}
	public Tuple getArgs(){return this.args;}
	public boolean isInvertible(){return this.invertibleFunction;}
	public boolean isPublic(){return this.publicFunction;}
	public void setConstantiz(ArrayList<Pair<Pair<ArrayList<Pair<Message,ParametricMessage>>,ParametricTuple>,ParametricConstant>> constantiz){
		this.constantiz=constantiz;
	}


	//interpolantString
	public String interpolantString(Environment env,ArrayList<Pair<Message, ParametricMessage>> assignment) {
		String ret="";
		
		for(int ass=0;ass<assignment.size();ass++){
		
		for(int i=0;i<this.constantiz.size();i++){
			for(int j=0;j<this.constantiz.get(i).getElement0().getElement0().size();j++){
				if(assignment.get(ass).getElement0().interpolantString(env).equals(
						this.constantiz.get(i).getElement0().getElement0().get(j).getElement0().interpolantString(env)) &&
						assignment.get(ass).getElement1().satString().equals(
								this.constantiz.get(i).getElement0().getElement0().get(j).getElement1().satString())){
					ret=this.constantiz.get(i).getElement1().satString();
					return ret;
				}
			}
		}
		
		}
		
		if(ret.equals("")){
			System.out.println("(AgentFunction->interpolantString(env,assignment) - in this object there's no assignment like the one you're searching for");
			System.exit(0);
		}
			
		
		return ret;
	}
	
	public String toString(){
		return this.messageString;
	}
}
