package se;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

public class AlgorithmState {
	
	private HashSet<Query> querySet;
	private HashSet<Annotation> annotationSet;
	private HashSet<GoalLocation> goalLocationSet;
	
		// constructor (corresponds to the rule INIT)
	public AlgorithmState(Program program) {
		
		HashSet<Parameter> ps = new HashSet<Parameter>();
		ConstraintSet cs = new ConstraintSet();
		Environment e = new Environment();
		e.addDataState(new DataState(IKVariable.getInstance(),new ParametricMessageSet()));
		Iterator<InputVariable> iter = program.getDomain().getDomainInputVars().iterator();
		while (iter.hasNext()){
			InputVariable iV=iter.next();
			e.addDataState(new DataState(iV,iV.eval(null)));			
		}

		SymbolicDataState sds = new SymbolicDataState(ps,cs,e,program.getRPD());//,program.getDomain());
		Query q = new Query(program.getRoot(),sds,null,null,null,false,false,false);
		
		querySet = new HashSet<Query>();
		querySet.add(q);
		
		annotationSet = new HashSet<Annotation>();
		
		goalLocationSet = program.getGoalLocationSet();

	}
	
	public HashSet<Query> getQuerySet(){
		return querySet;
	}

		// not being used at the moment
	public HashSet<Annotation> getAnnotationSet(){
		return annotationSet;
	}

	public HashSet<GoalLocation> getGoalLocationSet(){
		return goalLocationSet;
	}
	
	public void addQuery(Query query){
		querySet.add(query);
	}
	
	public void print() {
		
		System.out.println("------------ AlgorithmState "+this.hashCode()+" ------------");
		System.out.println();
		System.out.println("QuerySet:");
		Iterator<Query> iter=querySet.iterator();
		int i=1;
		while(iter.hasNext()){
			Query current=iter.next();
			System.out.println(i+". ");
			current.print();
			i++;
		}
		System.out.println();
		System.out.println("goalLocationSet:");
		Iterator<GoalLocation> iterG=goalLocationSet.iterator();
		int iG=1;
		while(iterG.hasNext()){
			GoalLocation current=iterG.next();
			System.out.println("("+iG+") "+current.print());
			iG++;
		}
		System.out.println();
		System.out.println("--------- end of AlgorithmState "+this.hashCode()+" ---------");
	}
	
}
