package se;

import java.util.*;

public class Assignment extends Action {

	Variable lhs;
	Evaluable rhs;

	public Assignment(Variable var, Evaluable ev){
		try{
			if(var instanceof Function){
				throw new Exception("You can not have a function on the LHS of an assignment");
			}else{
				this.lhs = var;
				this.rhs = ev;
			}
		}catch(Exception e){
			e.printStackTrace();
			System.exit(0);
		}
	}
	
	public Assignment(Variable var, Evaluable ev, String actionString){
		this(var, ev);
		this.actionString = actionString;
	}

	public void add2Spec(Specification spec){
		spec.addMessage((Variable)this.lhs);
		spec.addMessage(this.rhs);
		spec.addDataState(this.lhs,this.rhs);
	}

	/**
	 * the se (smbolic execution) method creates a datastate ds.
	 * ds is a pair (variable, evaluation of the RHS) and this means that
	 * in order to create an assignment the RHS has to be part of the environment set 
	 * or evaluable. 
	 * An assignment between two variable is possible iff the RHS var is already present in the Environment
	 */
	public SymbolicDataState se(SymbolicDataState sds){
		DataState ds = new DataState(lhs.getDataState(sds.getEnvironment()),rhs.eval(sds.getEnvironment()));
		SymbolicDataState newSds;
		newSds = sds.getCopy();
		newSds.addDataState2Environment(ds);
		return newSds;
	}

	public String print() {
		return "Assignment "+this.hashCode()+" : "+this.lhs.print()+ " := " + this.rhs.print();
	}
	
	public String getSemantics(Domain dom,Environment env, int RPD){
		String ret="(and ";

		//loop on assignment
		ArrayList<ArrayList<Pair<Message, ParametricMessage>>> assignments=dom.getFunAss();

		if(assignments.size()==0){
			//no functions in the spec
			//v'=e
			String msgInterpolantString=this.rhs.interpolantString(env);
			ret+="(= "+this.lhs.prime()+" "+msgInterpolantString+")";
			
		}else{
			//function support
			ret+="(or";
			
			for(int ass=0;ass<assignments.size();ass++){

				//get the actual assignment
				ArrayList<Pair<Message, ParametricMessage>> currentAss=assignments.get(ass);
				ret+="(and";
				for(int var=0;var<currentAss.size();var++){
					ret+=" (= "+currentAss.get(var).getElement0().interpolantString(env, currentAss)+" "+currentAss.get(var).getElement1().satString()+")";
				}
				
				//v'=e
				String msgInterpolantString=this.rhs.interpolantString(env,currentAss);
				ret+="(= "+this.lhs.prime()+" "+msgInterpolantString+"))";

			}
			ret+=")";
		}
		
		
		//forall x in Var s.t. x!=v x'=x
		Iterator<Variable> iter=dom.getDomainVars().iterator();
		while(iter.hasNext()){
			Variable current=iter.next();
			if(!(current.equals(this.lhs))){
				ret+="(= "+current.prime()+" "+current.interpolantString()+" )";
			}
		}
		Iterator<InputVariable> iterInput=dom.getDomainInputVars().iterator();
		while(iterInput.hasNext()){
			InputVariable current=iterInput.next();
			if(!(current.equals(this.lhs))){
				ret+="(= "+current.prime()+" "+current.interpolantString()+" )";
			}
		}

		//forall m in groundMsgs IK'(m)=IK(m)
		Iterator<ParametricMessage> iterGR=dom.getDomain().iterator();
		while(iterGR.hasNext()){
			ParametricMessage currentGR=iterGR.next();
			ret+="(= (IK' "+currentGR.satString()+") (IK "+currentGR.satString()+") )\n";			
		}

		//forall f:facts forall t:tuple f(t)=f'(t)
		ret+=dom.f_eq_fp(env);

		ret+=")"; //and
		
		return ret;
	}

	public String toString(){
		return this.actionString;
	}
	
	public Assignment substitute(ArrayList<Pair<Message, Message>> mex_newMex){
//		System.out.println("DEBUG subs1 "+this.toString());
		
		Variable tmp_lhs=this.lhs;
		Evaluable tmp_rhs=this.rhs;
		
		boolean eq=false;


		for(int i=0;i<mex_newMex.size();i++){
			if(this.lhs.equals(mex_newMex.get(i).getElement0())){
				tmp_lhs=(Variable)mex_newMex.get(i).getElement1();
//				System.out.println(mex_newMex.get(i).getElement1().toString());
				eq=true;
			}
			if(this.rhs.equals(mex_newMex.get(i).getElement0())){
				tmp_rhs=mex_newMex.get(i).getElement1();
				eq=true;
//				System.out.println("trovato!");
			}
		}
		
		String tmp_actionString=tmp_lhs.toString()+" := "+tmp_rhs.toString();
		
		
		
		eq=false;
		
		Assignment a=new Assignment(tmp_lhs,tmp_rhs,tmp_actionString);
//		System.out.println(a.toString());
//		System.out.println();
		return a;
	}
}
