package se;

import java.util.ArrayList;
import java.util.Iterator;

public class ConjunctiveConditional extends StdConditional{
	
	private StdConditional conditional1;
	private StdConditional conditional2;
	
	public void add2Spec(Specification spec){
		this.conditional1.add2Spec(spec);
		this.conditional2.add2Spec(spec);
	}
	
	public ConjunctiveConditional(StdConditional conditional1,StdConditional conditional2){
		this.conditional1 = conditional1;
		this.conditional2 = conditional2;
	}
	
	public ConjunctiveConditional(StdConditional conditional1, StdConditional conditional2, String actionString){
		this(conditional1, conditional2);
		this.actionString = actionString;
	}
	
	public ConjunctiveConditional substitute(ArrayList<Pair<Message, Message>> mex_newMex){
		StdConditional tmp_conditional1=(StdConditional)this.conditional1.substitute(mex_newMex);
		StdConditional tmp_conditional2=(StdConditional)this.conditional2.substitute(mex_newMex);
		return new ConjunctiveConditional(tmp_conditional1,tmp_conditional2,tmp_conditional1.toString()+" AND "+tmp_conditional2.toString());
	}
	
	public StdConditional getLhs(){return this.conditional1;}
	public StdConditional getRhs(){return this.conditional2;}
	
	public SymbolicDataState se(SymbolicDataState sds){
		SymbolicDataState newSds;
		newSds = sds.getCopy();
		
		ConjunctiveConstraint conjunctiveConstraint=this.recursiveSe(newSds);
		newSds.addConstraint2ConstraintSet(conjunctiveConstraint);
		
		return newSds;
	}
	
	public ConjunctiveConstraint recursiveSe(SymbolicDataState sds){
		ConjunctiveConstraint conjunctiveConstraint = new ConjunctiveConstraint(this.getLhs().recursiveSe(sds),this.getRhs().recursiveSe(sds));
		return conjunctiveConstraint;
	}
	
	public String print() {
		return "ConjunctiveConditional "+this.hashCode()+" : ("+this.conditional1.print()+ ") AND (" + this.conditional2.print()+")";
	}
	
	public String getSemantics(Domain dom, Environment env, int RPD){
		String ret="(and ";
		
		ret+=this.conditional1.getSemantics(dom, env, RPD);
		ret+=this.conditional2.getSemantics(dom, env, RPD);
		
		ret+=")"; //and
		return ret;
	}
		
	public String toString(){
		String ret="( "+this.conditional1.toString()+" AND "+this.conditional2.toString()+" )";
		return ret;
	}

}
