package se;

import java.util.ArrayList;

public class DisjunctiveConditional extends StdConditional{

	private StdConditional conditional1;
	private StdConditional conditional2;

	public void add2Spec(Specification spec){
		this.conditional1.add2Spec(spec);
		this.conditional2.add2Spec(spec);
	}

	public DisjunctiveConditional(StdConditional conditional1,StdConditional conditional2){
		this.conditional1 = conditional1;
		this.conditional2 = conditional2;
	}

	public DisjunctiveConditional(StdConditional conditional1,StdConditional conditional2, String actionString){
		this(conditional1,conditional2);
		this.actionString = actionString;
	}
	public DisjunctiveConditional substitute(ArrayList<Pair<Message, Message>> mex_newMex){
		StdConditional tmp_conditional1=(StdConditional)this.conditional1.substitute(mex_newMex);
		StdConditional tmp_conditional2=(StdConditional)this.conditional2.substitute(mex_newMex);
		return new DisjunctiveConditional(tmp_conditional1,tmp_conditional2,tmp_conditional1.toString()+" OR "+tmp_conditional2.toString());
	}
	public StdConditional getLhs(){return this.conditional1;}
	public StdConditional getRhs(){return this.conditional2;}

	public SymbolicDataState se(SymbolicDataState sds){
		SymbolicDataState newSds;
		newSds = sds.getCopy();

		DisjunctiveConstraint conjunctiveConstraint=this.recursiveSe(newSds);
		newSds.addConstraint2ConstraintSet(conjunctiveConstraint);

		return newSds;
	}

	public DisjunctiveConstraint recursiveSe(SymbolicDataState sds){
		DisjunctiveConstraint disjunctiveConstraint = new DisjunctiveConstraint(this.getLhs().recursiveSe(sds),this.getRhs().recursiveSe(sds));
		return disjunctiveConstraint;
	}

	public String print() {
		return "DisjunctiveConditional "+this.hashCode()+" : "+this.conditional1.print()+ " OR " + this.conditional2.print();
	}	

	public String toString(){
		return this.actionString;
	}

	public String getSemantics(Domain dom, Environment env, int RPD){
		String ret="(or ";

		ret+=this.conditional1.getSemantics(dom, env, RPD);
		ret+=this.conditional2.getSemantics(dom, env, RPD);

		ret+=")"; //and
		return ret;
	}
}
