package se;

import java.util.ArrayList;
import java.util.Iterator;

public class EqOPConditional extends OPConditional{

	public EqOPConditional(Evaluable lhs,Evaluable rhs) {
		super(lhs,rhs);
	}
	
	public EqOPConditional(Evaluable lhs,Evaluable rhs, String actionString) {
		this(lhs, rhs);
		this.actionString = actionString;
	}
	
	public EqOPConditional substitute(ArrayList<Pair<Message, Message>> mex_newMex){
		Evaluable tmp_lhs=super.getLhs().substitute(mex_newMex);
		Evaluable tmp_rhs=super.getRhs().substitute(mex_newMex);
		return new  EqOPConditional(tmp_lhs,tmp_rhs,tmp_lhs.toString()+" == "+tmp_rhs.toString());
		
	}
	
//	public void modifyActionString(Evaluable lhs,Evaluable rhs){
//		this.actionString=lhs.toString()+" == "+rhs.toString();
//	}
	
	/**
	 * Warning!
	 * with se() when we consider functions, we create a constraint between instances of the two functions.
	 * The check should be done on the pair (functionName, arguments) but it should be handled in ConstraintSet method satCheck (foci print). 
	 */
	public SymbolicDataState se(SymbolicDataState sds){	
		
		SymbolicDataState newSds;
		newSds = sds.getCopy();
	
//		if(this.getLhs() instanceof FactVariable){
//			EqOPConstraint eqOPConstraint = new EqOPConstraint(((FactVariable)this.getLhs()).getConditional(newSds.getEnvironment()),this.getRhs().eval(newSds.getEnvironment()));
//			newSds.addConstraint2ConstraintSet(eqOPConstraint);
//		}else{
			EqOPConstraint eqOPConstraint = new EqOPConstraint(this.getLhs().eval(newSds.getEnvironment()),this.getRhs().eval(newSds.getEnvironment()));
			newSds.addConstraint2ConstraintSet(eqOPConstraint);
//		}
		
		return newSds;
	}
	
	public EqOPConstraint recursiveSe(SymbolicDataState sds){
		EqOPConstraint eqOPConstraint = new EqOPConstraint(this.getLhs().eval(sds.getEnvironment()),this.getRhs().eval(sds.getEnvironment()));
		return eqOPConstraint;
	}

	public String print() {
		return "EqOPConditional "+this.hashCode()+ " : (" + this.getLhs().print() + ") == (" + this.getRhs().print() + ")";
	}
	
	public String getSemantics(Domain dom, Environment env, int RPD){
		String ret="";

		//for all messages in messageSet of IK+{messageSet}
		ret+="(and ";

		//forall vars assignment (to handle functions)
		ArrayList<ArrayList<Pair<Message, ParametricMessage>>> assignments=dom.getFunAss();
		boolean funAss=false;
		if(assignments.size()!=0){
			funAss=true;
		}

		if(funAss){
			ret+="(or ";

			//loop on assignment
			for(int ass=0;ass<assignments.size();ass++){

				//get the actual assignment
				ArrayList<Pair<Message, ParametricMessage>> currentAss=assignments.get(ass);
				
				ret+="(and";
				for(int var=0;var<currentAss.size();var++){
					ret+=" (= "+currentAss.get(var).getElement0().interpolantString(env, currentAss)+" "+currentAss.get(var).getElement1().satString()+")";
				}

				this.getSemantics_procedure(dom, env, RPD, currentAss);
				
				ret+=")";
			}
			ret+=")";
		}else{
			ret+=this.getSemantics_procedure(dom, env, RPD, null);
		}

		ret+=")";
		return ret;
	}

	private String getSemantics_procedure(Domain dom, Environment env, int RPD,ArrayList<Pair<Message, ParametricMessage>> assignment){
		String ret="";
		
		//e1=e2
		ret+="(= "+((Evaluable)this.getLhs()).interpolantString(env,assignment)+" "+((Evaluable)this.getRhs()).interpolantString(env,assignment)+")";
		
		
		//forall x in Var x'=x
		Iterator<Variable> iter=dom.getDomainVars().iterator();
		while(iter.hasNext()){
			Variable current=iter.next();
			ret+="(= "+current.prime()+" "+current.interpolantString()+" )";			
		}
		Iterator<InputVariable> iterInput=dom.getDomainInputVars().iterator();
		while(iterInput.hasNext()){
			InputVariable current=iterInput.next();
			ret+="(= "+current.prime()+" "+current.interpolantString()+" )\n";
		}
		//forall m in groundMsgs IK'(m)=IK(m)
		Iterator<ParametricMessage> iterGR=dom.getDomain().iterator();
		while(iterGR.hasNext()){
			ParametricMessage currentGR=iterGR.next();
				ret+="(= (IK' "+currentGR.satString()+") (IK "+currentGR.satString()+") )\n";			
		}
	
		//forall f:facts forall t:tuple f(t)=f'(t)
		ret+=dom.f_eq_fp(env);
		
//		ret+=")"; //and
		return ret;
	}

	public String toString(){
		return this.actionString;
	}
}