package se;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

//import Tests.Combinations;

public class IKConditional extends Conditional{
	
	private Message message;

	public IKConditional(Message message){
		this.message=message;
	}

	public IKConditional(Message message, String actionString){
		this(message);
		this.actionString = actionString;
	}
	
	public IKConditional substitute(ArrayList<Pair<Message, Message>> mex_newMex){
//		System.out.println("before IK |- m "+this.toString());
		Message tmp_message=(Message)this.message.substitute(mex_newMex);
		String tmp_actionString="IK |- "+tmp_message.toString();
		return new IKConditional(tmp_message,tmp_actionString);
//		System.out.println("after IK |- m "+this.toString());
	}
	
	public SymbolicDataState se(SymbolicDataState sds){
		
		SymbolicDataState newSds;
		newSds = sds.getCopy();
		
		System.out.println("BEFORE");
		sds.getConstraintSet().print();
		
		IKConstraint IKcon = new IKConstraint(IKVariable.getInstance().eval(newSds.getEnvironment()),message.eval(newSds.getEnvironment()));
		newSds.addConstraint2ConstraintSet(IKcon);
		
		System.out.println("\n\nAFTER");
		newSds.getConstraintSet().print();
		
		return newSds;
	}
	
	public String print() {
		return "IKConditional "+this.hashCode()+" : message= "+this.message.print();
	}
	
	public String getSemantics(Domain dom, Environment env, int RPD){
		String ret="";

		//for all messages in messageSet of IK+{messageSet}
		ret+="(and ";

		//forall vars assignment (to handle functions)
		ArrayList<ArrayList<Pair<Message, ParametricMessage>>> assignments=dom.getFunAss();
		boolean funAss=false;
		if(assignments.size()!=0){
			funAss=true;
		}

		if(funAss){
			ret+="(or ";

			//loop on assignment
			for(int ass=0;ass<assignments.size();ass++){

				//get the actual assignment
				ArrayList<Pair<Message, ParametricMessage>> currentAss=assignments.get(ass);
				
				ret+="(and";
				for(int var=0;var<currentAss.size();var++){
					ret+=" (= "+currentAss.get(var).getElement0().interpolantString(env, currentAss)+" "+currentAss.get(var).getElement1().satString()+")";
				}

				this.getSemantics_procedure(dom, env, RPD, currentAss);
				
				ret+=")";
			}
			ret+=")";
		}else{
			ret+=this.getSemantics_procedure(dom, env, RPD, new ArrayList<Pair<Message, ParametricMessage>>());
		}

		ret+=")";
		return ret;
	}

	private String getSemantics_procedure(Domain dom, Environment env, int RPD,ArrayList<Pair<Message, ParametricMessage>> assignment){
		
		String ret="";
		
		ret+=phi(dom, IKVariable.getInstance().eval(env), RPD, env);
		
		//IK_n(m)
		String temp="";
		
		//We store in temp the result of interpolantString method. 
		//In temp we now have the string for iz3 but
		//with variables for pk and pvt keys
		//(i.e. (pk Vari)...(pvt Varj))
		temp+=this.message.interpolantString(env,assignment);
//		System.out.println("DEBUG1 "+temp);
		
//		ret+=";TEST "+temp;
		
		//if(function)
		// scrivere IK|-f(m) constantizzato rispetto ai vari assegnamenti
		//
		//else{ se invece non e' una funzione non serve parlare di assegnamenti e quindi 
		//      vado avanti con tutto quello che segue
		//now we instantiate all the keys 
		if(temp.contains("(pk") || temp.contains("(pvt")){
			
			ArrayList<String> groups=new ArrayList<String>();
			//substitute (pk AV) and/or (pvt AV) with all the 
			//possible AgentConstants from Domain
			Pattern pattern=Pattern.compile("\\(pk (AV\\d+)\\s*\\)|\\(pvt (AV\\d+)\\s*\\)");
			Matcher matcher=pattern.matcher(temp);

			//create groups that is the arraylist of all 
			//the variable of pk and pvt key
			while (matcher.find()) {
				
				if(!matcher.group().isEmpty()){
					if(!groups.contains(matcher.group().replaceAll("\\(pk", "").replaceAll("\\(pvt", "").replaceAll("\\)", "")))
						groups.add(matcher.group().replaceAll("\\(pk", "").replaceAll("\\(pvt", "").replaceAll("\\)", ""));
				}
			}
			
			//calculate size of the assignments matrix
			//add all the assignments to the matrix
			
			//search constant row in constant
			//we save the row of agentConstant,publickeyconstant and privatekeyconstant
			int constRow=0,constPKRow=0,constPVTRow=0;
			boolean constPKRowExist=false,constPVTRowExist=false;
			for(int k=0;k<dom.getPMConstants().size();k++){
//				ret+="AH! "+dom.getPMConstants().get(k).get(0).getClass().toString();
				if(dom.getPMConstants().get(k).get(0).getClass().equals(ParametricAgentConstant.class)){
					constRow=k;
//					ret+="AH1!";
				}
				if(dom.getPMConstants().get(k).get(0).getClass().equals(ParametricPublicKeyConstant.class)){
					constPKRow=k;
//					ret+="AH2!";
					constPKRowExist=true;
				}
				if(dom.getPMConstants().get(k).get(0).getClass().equals(ParametricPrivateKeyConstant.class)){
//					ret+="AH3!";
					constPVTRow=k;
					constPVTRowExist=true;
				}
			}
			
			//convert groups Arraylist to Array to call combine
			//groupsArray will be the Variable array
			groups.trimToSize();
			String[] varsArray=groups.toArray(new String[groups.size()]);
//			System.out.println();
//			System.out.println("VARS");
//			for(int a=0;a<varsArray.length;a++)
//					System.out.println("["+a+"] "+varsArray[a]);
//			//sanitize
//			for(int vars=0;vars<varsArray.length;vars++)
//				varsArray[vars]=varsArray[vars].trim();
			
			//convert getPMConstant Arraylist to Array to call combine
			//constArray will be the Variable array			
			ParametricMessage[] constArrayP=dom.getPMConstants().get(constRow).toArray(new ParametricMessage[dom.getPMConstants().get(constRow).size()]);
			String[] constArray=new String[constArrayP.length];
			for(int c=0;c<constArrayP.length;c++)
				constArray[c]=constArrayP[c].satString();
				
//			System.out.println();
//			System.out.println("CONST");
//			for(int a=0;a<constArray.length;a++)
//					System.out.println("["+a+"] "+constArray[a]);
			
			
			String[][] combinations=this.combine(constArray, varsArray);
//			//sanitize
//			for(int combs=0;combs<combinations.length;combs++)
//				for(int combs2=0;combs2<combinations[combs].length;combs2++)
//				combinations[combs][combs2]=combinations[combs][combs2].trim();
			
			//here we write
			// (or
			//  (and (assignments) (temp wrt assgnments))
			//  ...
			// )

			String currentemp="";
			//we copy temp so we don't directly modify it
			currentemp="(IK "+temp+")";

			
//			System.out.println("MSG="+currentemp);
			String toModify=currentemp;
			
//			System.out.println();
//			System.out.println("MATRIX");
//			for(int a=0;a<combinations.length;a++)
//				for(int b=0;b<combinations[a].length;b++)
//					System.out.println("["+a+" "+b+"] "+combinations[a][b]);
//			System.out.println();
//			System.out.println("PKS");
//			for(int a=0;a<dom.getPMConstants().get(constPKRow).size();a++)
//					System.out.println("["+constPKRow+" "+a+"] "+((ParametricPublicKeyConstant)dom.getPMConstants().get(constPKRow).get(a)).satString());			
//			System.out.println();
			
			
//			boolean stopCond=false;
			ret+="(or\n";
			for(int row=0;row<combinations.length;row++){
				ret+="(and";
				for(int col=0;col<combinations[row].length;col++){
					ret+="(=\n"+varsArray[col]+"\n "+combinations[row][col]+"\n)\n";
					//we search for the public key of the agent (constant)
					if(constPKRowExist)
					for(int pk=0;pk<dom.getPMConstants().get(constPKRow).size();pk++){
						
//						System.out.println("TEST: "+((ParametricPublicKeyConstant)dom.getPMConstants().get(constPKRow).get(pk)).getParametricAgent().satString()+
//						" =? "+combinations[row][col]);
							if(((ParametricPublicKeyConstant)dom.getPMConstants().get(constPKRow).get(pk)).getParametricAgent().satString().
									equals(combinations[row][col])){
//								currentemp="(IK"+RPD+" "+temp+")";
								
//								DEBUG
//								System.out.println("\n\n**********REGEX**********\n");
//								System.out.println("VAR:"+varsArray[col]+"f");
//								System.out.println("CONS:"+combinations[row][col]+"f");								
//								System.out.println("STRING:"+currentemp+"f");
//								System.out.println("NEW STRING:"+currentemp.replaceAll("\\(pk\\s*"+varsArray[col]+"\\s*\\)",combinations[row][col])+"f");
//								System.out.println("\n\n**********REGEX**********");
								
//								currentemp=currentemp.replaceAll("\\(pk\\s*"+varsArray[col]+"\\s*\\)",combinations[row][col]);
								currentemp=currentemp.replaceAll("\\(pk\\s*"+varsArray[col]+"\\s*\\)",
										((ParametricPublicKeyConstant)dom.getPMConstants().get(constPKRow).get(pk)).satString());
//								stopCond=true;
								break;
							}
//							if(stopCond){
//								break;
//							}
					}
					//we search for the private key of the agent (constant)
					if(constPVTRowExist)
					for(int pvt=0;pvt<dom.getPMConstants().get(constPVTRow).size();pvt++)
							if(((ParametricPrivateKeyConstant)dom.getPMConstants().get(constPVTRow).get(pvt)).getParametricAgent().satString().
									equals(combinations[row][col])){
//								currentemp=currentemp.replaceAll("\\(\\s*pvt\\s*"+varsArray[col]+"\\s*\\)",combinations[row][col]);	
								currentemp=currentemp.replaceAll("\\(\\s*pvt\\s*"+varsArray[col]+"\\s*\\)",
										((ParametricPrivateKeyConstant)dom.getPMConstants().get(constPVTRow).get(pvt)).satString());}
				}
				ret+=currentemp+" )\n";
				currentemp=toModify;
			}
			ret+=")\n";
		}
		else{
			
			ret+="\n(IK "+temp+")\n";
//			System.out.println("DEBUG (IK"+RPD+" "+temp+")\n");
		}
		
		//forall x in Var x'=x
		Iterator<Variable> iter=dom.getDomainVars().iterator();
		while(iter.hasNext()){
			Variable current=iter.next();
			ret+="(= "+current.prime()+"  "+current.interpolantString()+" )\n";
		}
		Iterator<InputVariable> iterInput=dom.getDomainInputVars().iterator();
		while(iterInput.hasNext()){
			InputVariable current=iterInput.next();
			ret+="(= "+current.prime()+" "+current.interpolantString()+" )\n";
		}
		//forall m in groundMsgs IK'(m)=IK(m)
		Iterator<ParametricMessage> iterGR=dom.getDomain().iterator();
		while(iterGR.hasNext()){
			ParametricMessage currentGR=iterGR.next();
				ret+="(= (IK' "+currentGR.satString()+") (IK "+currentGR.satString()+") )\n";			
		}
		
		//forall f:facts forall t:tuple f(t)=f'(t)
		ret+=dom.f_eq_fp(env);
		
//		ret+=")"; //and
		return ret;
	}
	
	
	private String phi(Domain dom,ParametricMessageSet ikEval, int RPD, Environment env){
		String ret="";
		
		
		//add all contants to the domain
		dom=dom.addConst2Dom();
		Iterator<ParametricMessage> domIter=dom.getDomainParametricMessage().iterator();
		String dy_current;
		String dy_previous;
		//System.out.println(dom.print());
		while(domIter.hasNext()){
			ParametricMessage currentM=domIter.next();
			for(int dyn=RPD;dyn>0;dyn--){
//				if(dyn==1){
//					dy_current="IK1";
				dy_current="IK";
//					dy_previous="IK";
//				}else{
//					dy_current="IK"+(dyn);
					dy_previous="IK";
//				}
				
				//function support
				String m=currentM.satString();
				
				String otherSideOfImplicationElimination="";
				String otherSideOfImplicationIntroduction="";
				
				ret+="\n (and";
				ret+="\n (implies";
				ret+="\n("+dy_current+" "+m+")";
				//monotonicity		
				ret+="\n(or";
				ret+="\n("+dy_previous+" "+m+")";
				//elimination phase
				Iterator<ParametricMessage> domIter2=dom.getDomainParametricMessage().iterator();
				//forall messege m2 in domain
				while(domIter2.hasNext()){
					ParametricMessage currentM2=domIter2.next();
					//if m2=pair
					if(currentM2.getClass().equals(ParametricConcatenatedMessage.class)){
						//if m2=m.* || m2=*.m
						if(((ParametricConcatenatedMessage)currentM2).getFirst().equals(currentM) ||
								((ParametricConcatenatedMessage)currentM2).getSecond().equals(currentM)){
							otherSideOfImplicationElimination+="\n ("+dy_previous+" "+currentM2.satString()+")";
							ret+="\n ("+dy_previous+" "+currentM2.satString()+")";
						}

					//if m2=crypt
					}else if(currentM2.getClass().equals(ParametricEncryptedMessage.class)){
						
						//if m2=m_k
						if(((ParametricEncryptedMessage)currentM2).getFirst().equals(currentM)){
//							otherSideOfImplicationElimination+="\n(and ("+dy_previous+" "+currentM2.satString()+") ("+dy_previous+" "+((ParametricEncryptedMessage)currentM2).getSecond().getParametricAgent().getPrivateKey().satString()+") )";
//							satFile.write("\n(and ("+dy_previous+" "+currentM2.satString()+") ("+dy_previous+" "+((ParametricEncryptedMessage)currentM2).getSecond().getParametricAgent().getPrivateKey().satString()+") )");										
//							System.out.println("\n(and ("+dy_previous+" "+currentM2.satString()+") "+((ParametricEncryptedMessage)currentM2).getSecond().getParametricAgent().getPrivateKey().satString()+" )");
							Iterator<ParametricMessage> domIter3=dom.getDomainParametricMessage().iterator();
							//forall m3 in domain
							while(domIter3.hasNext()){
								ParametricMessage currentM3=domIter3.next();
								//if m3=inv(k)
								if(currentM3.equals(((ParametricEncryptedMessage)currentM2).getSecond().getParametricAgent().getPrivateKey())){
									otherSideOfImplicationElimination+="\n(and ("+dy_previous+" "+currentM2.satString()+") ("+dy_previous+" "+((ParametricEncryptedMessage)currentM2).getSecond().getParametricAgent().getPrivateKey().satString()+") )";
									ret+="\n(and ("+dy_previous+" "+currentM2.satString()+") ("+dy_previous+" "+((ParametricEncryptedMessage)currentM2).getSecond().getParametricAgent().getPrivateKey().satString()+") )";					
								}
								
							}
				
						}
						
					}
					//if m2=scrypt
					else if(currentM2.getClass().equals(ParametricSymEncryptedMessage.class)){

						//if m2=m_sk
						if(((ParametricSymEncryptedMessage)currentM2).getFirst().equals(currentM)){
							Iterator<ParametricMessage> domIter4=dom.getDomainParametricMessage().iterator();
							//forall m3 in domain
							while(domIter4.hasNext()){
								ParametricMessage currentM4=domIter4.next();
								if(currentM4.equals(((ParametricSymEncryptedMessage)currentM2).getSecond())){
									otherSideOfImplicationElimination+="\n(and ("+dy_previous+" "+currentM2.satString()+") ("+dy_previous+" "+((ParametricSymEncryptedMessage)currentM2).getSecond().satString()+") )";
									ret+="\n(and ("+dy_previous+" "+currentM2.satString()+") ("+dy_previous+" "+((ParametricSymEncryptedMessage)currentM2).getSecond().satString()+") )";
								}
							}

						}

					}
				}
				
				//introduction phase
				if(currentM.getClass().equals(ParametricConcatenatedMessage.class)){
					otherSideOfImplicationIntroduction+="\n(and ("+dy_previous+" "+((ParametricConcatenatedMessage)currentM).getFirst().satString()+")";
					otherSideOfImplicationIntroduction+="\n("+dy_previous+" "+((ParametricConcatenatedMessage)currentM).getSecond().satString()+"))";
					ret+=otherSideOfImplicationIntroduction;
				}else if(currentM.getClass().equals(ParametricEncryptedMessage.class)){
					otherSideOfImplicationIntroduction+="\n(and ("+dy_previous+" "+((ParametricEncryptedMessage)currentM).getFirst().satString()+")";
					otherSideOfImplicationIntroduction+="\n("+dy_previous+" "+((ParametricEncryptedMessage)currentM).getSecond().satString()+"))";
					ret+=otherSideOfImplicationIntroduction;
				}


				ret+="\n ) \n"; //end of or
				ret+=") \n"; //end of implies
				
				//other direction of the implication
				ret+="\n (implies";
				
				//monotonicity		
				ret+="\n(or";
				ret+="\n("+dy_previous+" "+m+")";
				//elimination phase
				ret+=otherSideOfImplicationElimination;
				
				//introduction phase
				ret+=otherSideOfImplicationIntroduction;
				
				ret+="\n ) \n"; //end of or
				
				ret+="\n("+dy_current+" "+m+")";
				
				ret+=") \n"; //end of implies

				ret+=")\n"; //end of and
				
				otherSideOfImplicationElimination="";
				otherSideOfImplicationIntroduction="";
			}
		}
		
		return ret;
	}

	private int pow(int value, int power){
		if(power == 0){
			return 1;
		}

		int result = value;
		for(int i = 0; i < power -1; i++){
			result = result * value; 
		}
		return result;
	}

	public void add2Spec(Specification spec){
		spec.addMessage(this.message);
	}
	
	
//	public String[][] combine(String[] constants,String[] variables){
//		//we want a combination matrix that represents every possible combination
//		//with repetitions
//		//that has variables on columns
//		//so every rows is a combination
//		String[][] combinations=new String[this.pow(variables.length,constants.length)][variables.length];
//
//		int z=this.pow(constants.length,variables.length);
//		int index=0;
//		int k=0;
//
//		for(int j=0;j<variables.length;j++){
//			z=z/variables.length;
//
//			index=0;
//			k=1;
//			for(int i=0;i<this.pow(variables.length,constants.length);i++){
//				System.out.print("i="+i+"\tj="+j+"\tz="+z+"\tindex="+index+"\tk="+k);
//				if(k<=z){
//					k++;
//				}else{ 
//					if(index<constants.length-1)
//						index++;
//					else
//						index=0;
//					k=2;
//				}
//				System.out.println("\t["+i+"]["+j+"]="+constants[index]);
//				combinations[i][j]=constants[index];	
//
//
//			}
//		}
//
//
//		return combinations;
//	}
	public String[][] combine(String[] constants,String[] variables){
		//we want a combination matrix that represents every possible combination
		//with repetitions
		//that has variables on columns
		//so every rows is a combination
		
		String[][] combinations;
		
		if(variables.length==1){
			combinations=new String[constants.length][1];
//			System.out.println("length="+constants.length);
			for(int i=0;i<constants.length;i++)
				combinations[i][0]=constants[i];
		}else{
			combinations=new String[this.pow(variables.length,constants.length)][variables.length];
			int z=this.pow(constants.length,variables.length);
			int index=0;
			int k=0;

			for(int j=0;j<variables.length;j++){
				z=z/variables.length;

				index=0;
				k=1;
				for(int i=0;i<this.pow(variables.length,constants.length);i++){
//					System.out.print("i="+i+"\tj="+j+"\tz="+z+"\tindex="+index+"\tk="+k);
					if(k<=z){
						k++;
					}else{ 
						if(index<constants.length-1)
							index++;
						else
							index=0;
						k=2;
					}
//					System.out.println("\t["+i+"]["+j+"]="+constants[index]);
					combinations[i][j]=constants[index];	


				}
			}
		}

		return combinations;
	}
	
		
	public String toString(){
		return this.actionString;
	}	


	//    private static int factorial(int n) {
	//        int fact = 1; // this  will be the result
	//        for (int i = 1; i <= n; i++) {
	//            fact *= i;
	//        }
	//        return fact;
	//    }

}
