package se;

import java.util.*;


public class IKUpdate extends Action{

	IKVariable ikvar=IKVariable.getInstance();
	Evaluable rhs;
	String actionString="";
	
	public IKUpdate(Evaluable ev){
		this.rhs = ev;
	}
	
	public IKUpdate(Evaluable ev, String actionString){
		this(ev);
		this.actionString = actionString;
	}
	

	public void add2Spec(Specification spec){
		spec.addMessage(this.rhs);
	}

	public IKUpdate substitute(ArrayList<Pair<Message, Message>> mex_newMex){
//		System.out.println("IKUPDATE before "+this.toString());
		Evaluable tmp_rhs=this.rhs.substitute(mex_newMex);
		
		return new IKUpdate(tmp_rhs,"IK := IK + "+tmp_rhs.toString());
//		System.out.println("IKUPDATE after "+this.toString());
//		System.out.println("\n");
	}
	
	/**
	 * the se (smbolic execution) method creates a datastate ds.
	 * ds is a pair (variable, evaluation of the RHS) and this means that
	 * in order to create an assignment the RHS has to be part of the environment set 
	 * or evaluable. 
	 * An assignment between two variable is possible iff the RHS var is already present in the Environment
	 */
	public SymbolicDataState se(SymbolicDataState sds){
		System.out.println("BEFORE");
		sds.getEnvironment().print();
		//		System.out.println("RHS="+this.rhs.print());
		//		System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");
		//		System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");
		//		System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");
		//		System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");
		SymbolicDataState newSds;
		newSds = sds.getCopy();

		DataState ds = new DataState(ikvar,rhs.eval(newSds.getEnvironment()));
		newSds.updateIKinEnvironment(ds);
		System.out.println("AFTER");
		newSds.getEnvironment().print();
		return newSds;
	}

	public String print() {
		return "IKUpdate";
	}

	public String getSemantics(Domain dom, Environment env, int RPD){
		String ret="";

		//for all messages in messageSet of IK+{messageSet}
		ret+="(and ";

		//forall vars assignment (to handle functions)
		ArrayList<ArrayList<Pair<Message, ParametricMessage>>> assignments=dom.getFunAss();
		boolean funAss=false;
		if(assignments.size()!=0){
			funAss=true;
		}

		if(funAss){
			ret+="(or ;debiku1\n";

			//loop on assignment
			for(int ass=0;ass<assignments.size();ass++){

				//get the actual assignment
				ArrayList<Pair<Message, ParametricMessage>> currentAss=assignments.get(ass);
				
				ret+="(and";
				for(int var=0;var<currentAss.size();var++){
					ret+=" (= "+currentAss.get(var).getElement0().interpolantString(env, currentAss)+" "+currentAss.get(var).getElement1().satString()+")";
				}

				this.getSemantics_procedure(dom, env, RPD, currentAss);
				
				ret+=")";
			}
			ret+=")";
		}else{
			ret+=this.getSemantics_procedure(dom, env, RPD, new ArrayList<Pair<Message, ParametricMessage>>());
		}

		ret+=")";
		return ret;
	}

	private String getSemantics_procedure(Domain dom, Environment env, int RPD,ArrayList<Pair<Message, ParametricMessage>> assignment){
//		System.out.println("DEBUG getasspermtrans semantics ikupdate");
//		dom.printAssPermTrans();
		String ret="";
		ArrayList<ArrayList<Message>> vars=new ArrayList<ArrayList<Message>>();
		ArrayList<Message> unsortedVars=new ArrayList<Message>();

		ret+="(or \n";
		Iterator<Message> iter=((MessageSet)this.rhs).iterator();
		while(iter.hasNext()){
			Message current=iter.next();
			//1. get all the variables
			vars=current.getVars(vars);
			for(int o=0;o<vars.size();o++)
				for(int p=0;p<vars.get(o).size();p++)
					unsortedVars.add(vars.get(o).get(p));

			//			System.out.println("VARS of "+current.interpolantString(env));
			//			for(int i=0;i<vars.size();i++)
			//				for(int j=0;j<vars.get(i).size();j++)
			//					System.out.println("["+i+" "+j+"] "+vars.get(i).get(j).interpolantString(env));


			//2. create all assignments
			ArrayList<String[][]> assignmentMatrix=new ArrayList<String[][]>();
			String[] varsArray;
			//forall variables
			for(int varType=0;varType<vars.size();varType++){
				varsArray=new String[vars.get(varType).size()];
				for(int d=0;d<vars.get(varType).size();d++){
					varsArray[d]=vars.get(varType).get(d).interpolantString(env,assignment);
				}
				//				System.out.println("VARSARRAY");
				//				for(int test=0;test<varsArray.length;test++)
				//					System.out.println("["+test+"] "+varsArray[test]);


				//search the constant row in domain constant list 
				for(int conType=0;conType<dom.getDomainPMConstans().size();conType++){
					//check if type are right
					if(dom.getPMConstants().get(conType).get(0).getClass().equals(ParametricAgentConstant.class) && vars.get(varType).get(0).getClass().equals(AgentVariable.class) ||
							dom.getPMConstants().get(conType).get(0).getClass().equals(ParametricAgentConstant.class) && vars.get(varType).get(0).getClass().equals(InputAgentVariable.class) ||
							dom.getPMConstants().get(conType).get(0).getClass().equals(ParametricNonceConstant.class) && vars.get(varType).get(0).getClass().equals(NonceVariable.class) ||
							dom.getPMConstants().get(conType).get(0).getClass().equals(ParametricNonceConstant.class) && vars.get(varType).get(0).getClass().equals(InputNonceVariable.class) ||
							dom.getPMConstants().get(conType).get(0).getClass().equals(ParametricSymKeyConstant.class) && vars.get(varType).get(0).getClass().equals(InputSymKeyVariable.class) ||
							dom.getPMConstants().get(conType).get(0).getClass().equals(ParametricGenericConstant.class) && vars.get(varType).get(0).getClass().equals(GenericVariable.class) ||
							dom.getPMConstants().get(conType).get(0).getClass().equals(ParametricPublicKeyConstant.class) && vars.get(varType).get(0).getClass().equals(PublicKey.class) ||
							dom.getPMConstants().get(conType).get(0).getClass().equals(ParametricPrivateKeyConstant.class) && vars.get(varType).get(0).getClass().equals(PrivateKey.class)||
							dom.getPMConstants().get(conType).get(0).getClass().equals(ParametricSymKeyConstant.class) && vars.get(varType).get(0).getClass().equals(SymKeyVariable.class)
							){

						//						System.out.println("PMCONST");
						//						for(int test=0;test<dom.getPMConstants().get(conType).size();test++)
						//							System.out.println(dom.getPMConstants().get(conType).get(test).satString()+", ");



						//convert getPMConstant Arraylist to Array to call combine
						ParametricMessage[] constArrayP=dom.getPMConstants().get(conType).toArray(new ParametricMessage[dom.getPMConstants().get(conType).size()]);
						String[] constArray=new String[constArrayP.length];
						for(int c=0;c<constArrayP.length;c++)
							constArray[c]=constArrayP[c].satString();
						//						System.out.println("CONSTARRAY");
						//						for(int test=0;test<constArray.length;test++)
						//							System.out.println("["+test+"] "+constArray[test]);

						//						String[][] a=this.createAssignments(constArray,varsArray);
						assignmentMatrix.add(this.createAssignments(constArray,varsArray));

						//						System.out.println("SUPERMATRIX");
						//						for(int test=0;test<a.length;test++)
						//							for(int test2=0;test2<a[test].length;test2++)
						//							System.out.println("["+test+" "+test2+"] "+a[test][test2]);
					}
				}
			}



			//			for(int m=0;m<assignmentMatrix.size();m++){
			//				System.out.println("\n\nMATRIX_"+m+":\n");
			//			for(int i=0;i<assignmentMatrix.get(m).length;i++)
			//				for(int j=0;j<assignmentMatrix.get(m)[i].length;j++)
			//					System.out.println("["+i+" "+j+"] "+assignmentMatrix.get(m)[i][j]);
			//			}
			//we now have assignmentMatrix that contains in each field a matrix with all the combination of a type
			//we need a matrix instead of a list of matrix and we have to combine matrixes
			//
			//forall Matrix
			// forall row1 in Matrix1
			//  forall row2 in Matrix2
			//   new Matrix3 <- row1.row2
			//   substitute Matrix1 and Matrix2 with Matrix3

			String[][] tempMatrix;
			while(assignmentMatrix.size()>1){
				String[][] toRemove0=assignmentMatrix.get(0);
				String[][] toRemove1=assignmentMatrix.get(1);
				tempMatrix=matrixCombine(toRemove0,toRemove1);
				assignmentMatrix.remove(toRemove0);
				assignmentMatrix.remove(toRemove1);

				assignmentMatrix.add(0,tempMatrix);
			}


			if(assignmentMatrix.size()!=0){
				String[][] finalAssignmentMatrix=assignmentMatrix.get(0);


				//			System.out.println("FINAL");
				//			for(int i=0;i<finalAssignmentMatrix.length;i++)
				//				for(int j=0;j<finalAssignmentMatrix[i].length;j++)
				//					System.out.println("["+i+" "+j+"] "+finalAssignmentMatrix[i][j]);


				//3. constantize wrt assignment (forall assignment)
				//4. write 
				//(or
				// (and ass1 (IK' constantized(m,ass1)) 
				//     forall m' in Ground
				//      if(m' != m) (=(IK' m')(IK m)) 
				// )
				// ...
				// )

				//search constant row in constant
				//we save the row of agentConstant,publickeyconstant and privatekeyconstant
				int constPKRow=0,constPVTRow=0;
				boolean constPKRowExist=false,constPVTRowExist=false;
				//			System.out.println("SIZE="+dom.getPMConstants().size());
				for(int k=0;k<dom.getPMConstants().size();k++){
					//				System.out.println("AH! "+dom.getPMConstants().get(k).get(0).getClass().toString());
					if(dom.getPMConstants().get(k).get(0).getClass().equals(ParametricPublicKeyConstant.class)){
						constPKRow=k;
						//					System.out.println("AH2!");
						constPKRowExist=true;
					}
					if(dom.getPMConstants().get(k).get(0).getClass().equals(ParametricPrivateKeyConstant.class)){
						//					System.out.println("AH3!");
						constPVTRow=k;
						constPVTRowExist=true;
					}
				}


				//			System.out.println("\n\n");
				String tempCurrent=current.interpolantString(env,assignment);

				for(int asRow=0;asRow<finalAssignmentMatrix.length;asRow++){
					tempCurrent=current.interpolantString(env,assignment);
					//				ret+="AAA"+tempCurrent;
					ret+="(and";
					for(int asCol=0;asCol<finalAssignmentMatrix[asRow].length;asCol++){
						//					System.out.println("colM ="+asCol);
						//					System.out.println("rowM ="+asRow);
						//					System.out.println("(= "+unsortedVars.get(asCol).interpolantString(env)+" "+finalAssignmentMatrix[asRow][asCol]+")");
						ret+="(= "+unsortedVars.get(asCol).interpolantString(env,assignment)+" "+finalAssignmentMatrix[asRow][asCol]+")";

						//					System.out.println("from:"+tempCurrent);
						tempCurrent=tempCurrent.replaceAll(unsortedVars.get(asCol).interpolantString(env,assignment),finalAssignmentMatrix[asRow][asCol]);

						//we search for the public key of the agent (constant)
						if(constPKRowExist)
							for(int pvt=0;pvt<dom.getPMConstants().get(constPKRow).size();pvt++)
								if(((ParametricPublicKeyConstant)dom.getPMConstants().get(constPKRow).get(pvt)).getParametricAgent().satString().
										equals(finalAssignmentMatrix[asRow][asCol]))
									tempCurrent=tempCurrent.replaceAll("\\(\\s*pk\\s*"+finalAssignmentMatrix[asRow][asCol]+"\\s*\\)",((ParametricPublicKeyConstant)dom.getPMConstants().get(constPKRow).get(pvt)).satString());	
						//we search for the private key of the agent (constant)
						if(constPVTRowExist)
							for(int pvt=0;pvt<dom.getPMConstants().get(constPVTRow).size();pvt++)
								if(((ParametricPrivateKeyConstant)dom.getPMConstants().get(constPVTRow).get(pvt)).getParametricAgent().satString().
										equals(finalAssignmentMatrix[asRow][asCol]))
									tempCurrent=tempCurrent.replaceAll("\\(\\s*pvt\\s*"+finalAssignmentMatrix[asRow][asCol]+"\\s*\\)",((ParametricPrivateKeyConstant)dom.getPMConstants().get(constPVTRow).get(pvt)).satString());	
					}

					ret+="(IK' "+tempCurrent+")";

					//we now add (= (IK' m')(IK m)) for any m'!=current considered message (i.e. tempCurrent)
					//forall m in groundMsgs IK'(m)=IK(m)
					Iterator<ParametricMessage> iterGR=dom.getDomain().iterator();
					while(iterGR.hasNext()){
						ParametricMessage currentGR=iterGR.next();
						if(!currentGR.satString().equals(tempCurrent))
							ret+="(= (IK' "+currentGR.satString()+") (IK "+currentGR.satString()+") )\n";			
					}

					ret+=")\n";
				}


				ArrayList<ArrayList<ParametricMessage>> C=dom.getDomainPMConstans();
				ArrayList<ParametricMessage> agentConstants;
				for(int find=0;find<C.size();find++)
					if(C.get(find).getClass().equals(ParametricAgentConstant.class)){
						for(int col=0;col<C.get(find).size();col++){
							ret.replaceAll("\\(\\s*pk\\s"+C.get(find).get(col).satString()+"\\s*\\)",((ParametricAgentConstant)C.get(find).get(col)).getPublicKey().satString());
							ret.replaceAll("\\(\\s*pvt\\s"+C.get(find).get(col).satString()+"\\s*\\)",((ParametricAgentConstant)C.get(find).get(col)).getPrivateKey().satString());
						}
						break;
					}

			}else{
				ret+="(and";
				ret+="(IK' "+current.interpolantString(env,assignment)+")";
				//we now add (= (IK' m')(IK m)) for any m'!=current considered message (i.e. tempCurrent)
				//forall m in groundMsgs IK'(m)=IK(m)
				Iterator<ParametricMessage> iterGR=dom.getDomain().iterator();
				while(iterGR.hasNext()){
					ParametricMessage currentGR=iterGR.next();
					if(!currentGR.satString().equals(current.interpolantString(env,assignment)))
						ret+="(= (IK' "+currentGR.satString()+") (IK "+currentGR.satString()+") )\n";			
				}
				ret+=")";
			}


		}
		ret+=")";


		//forall f:facts forall t:tuple f(t)=f'(t)
		ret+=dom.f_eq_fp(env);

		//forall x in Var x'=x
		Iterator<Variable> iterV=dom.getDomainVars().iterator();
		while(iterV.hasNext()){
			Variable currentV=iterV.next();			
			ret+="(= "+currentV.prime()+"  "+currentV.interpolantString()+" )\n";			
		}
		Iterator<InputVariable> iterInput=dom.getDomainInputVars().iterator();
		while(iterInput.hasNext()){
			InputVariable current=iterInput.next();
			ret+="(= "+current.prime()+" "+current.interpolantString()+" )\n";
		}
		System.out.println("DEBUG getasspermtrans semantics ikupdate");
		dom.getAssPermTrans();
		return ret;
	}

	public String[][] createAssignments(String[] constants,String[] variables){
		//we want a combination matrix that represents every possible combination
		//with repetitions
		//that has variables on columns
		//so every rows is a combination

		String[][] combinations;

		if(variables.length==1){
			combinations=new String[constants.length][1];
			//			System.out.println("length="+constants.length);
			for(int i=0;i<constants.length;i++)
				combinations[i][0]=constants[i];
		}else{
			combinations=new String[this.pow(variables.length,constants.length)][variables.length];
			int z=this.pow(constants.length,variables.length);
			int index=0;
			int k=0;

			for(int j=0;j<variables.length;j++){
				z=z/variables.length;

				index=0;
				k=1;
				for(int i=0;i<this.pow(variables.length,constants.length);i++){
					//					System.out.print("i="+i+"\tj="+j+"\tz="+z+"\tindex="+index+"\tk="+k);
					if(k<=z){
						k++;
					}else{ 
						if(index<constants.length-1)
							index++;
						else
							index=0;
						k=2;
					}
					//					System.out.println("\t["+i+"]["+j+"]="+constants[index]);
					combinations[i][j]=constants[index];	


				}
			}
		}

		return combinations;
	}

	private int pow(int value, int power){
		if(power == 0){
			return 1;
		}

		int result = value;
		for(int i = 0; i < power -1; i++){
			result = result * value; 
		}
		return result;
	}

	private String[][] matrixCombine(String[][] m1,String[][] m2){
		String[][] m3=new String[(m1.length)*(m2.length)][(m1[0].length)+(m2[0].length)];

		int i3=0,j3=0;

		for(int i1=0;i1<m1.length;i1++){

			for(int i2=0;i2<m2.length;i2++){

				for(int j1=0;j1<m1[i1].length;j1++){

					m3[i3][j3]=m1[i1][j1];
					j3++;
				}
				for(int j2=0;j2<m2[i2].length;j2++){

					m3[i3][j3]=m2[i2][j2];
					j3++;
				}				
				i3++;
				j3=0;
			}
		}

		return m3;
	}

	public String toString(){
		return this.actionString;
	}

}
