package se;

import java.util.HashSet;
import java.util.Iterator;
import java.util.ArrayList;

public class NotMInIKStdConstraint extends StdConstraint{

	private ParametricMessageSet ik;
	private ParametricMessage message;
	
	public NotMInIKStdConstraint(ParametricMessageSet ik,ParametricMessage message){
		this.setIK(ik);
		this.setMessage(message);
		StdConstraint.adddy2List(this.ik.satString());
	}
	
	public ParametricMessageSet getIK(){return this.ik;};
	public ParametricMessage getMessage(){return this.message;};
	public void setIK(ParametricMessageSet ik){this.ik=ik;}
	public void setMessage(ParametricMessage message){this.message=message;}
	
	public ArrayList<ParametricMessage> getParameters(ArrayList<ParametricMessage> params){
		return ik.getParameters(this.message.getParameters(params));
	}

	//satprint
	public String satPrint(Domain dom){
		String ret="";
		boolean inik=false;
		ret+=("\n(and true\n");
		ArrayList<ParametricMessage> mexL=new ArrayList<ParametricMessage>();
		//forall messages in IK
		Iterator<ParametricMessage> iterIK=this.ik.iterator();
		while(iterIK.hasNext()){
			ParametricMessage currentIter=iterIK.next();
			mexL.add(currentIter);
			ret+="("+this.ik.satString()+" "+currentIter.satString()+")\n";
		}

		//forall messages in domain (and forall constants)
		//if it is not a message of IK then not(DY0 m)
		Iterator<ParametricMessage> iterDom=dom.getDomain().iterator();
		while(iterDom.hasNext()){
			ParametricMessage current=iterDom.next();
			for(int ikit=0;ikit<mexL.size();ikit++){
				if(current.satString().equals(mexL.get(ikit).satString())){
					inik=true;
					break;
				}
			}
			if(!inik){
				ret+="(not ("+this.ik.satString()+" "+current.satString()+"))\n";
			}
			inik=false;
		}

		for(int itC=0;itC<dom.getDomainPMConstans().size();itC++)
			for(int itC2=0;itC2<dom.getDomainPMConstans().get(itC).size();itC2++){
				ParametricMessage current=dom.getDomainPMConstans().get(itC).get(itC2);
				for(int ikit=0;ikit<mexL.size();ikit++)
					if(current.satString().equals(mexL.get(ikit).satString())){
						inik=true;
						break;
					}
				if(!inik)
					ret+="(not ("+this.ik.satString()+" "+current.satString()+"))\n";
				inik=false;
			}

		//here we create the constantiziation of m (of IK|-m)
		ret+="(not ("+this.ik.satString()+" "+this.message.satString()+"))) ;minik\n";

		return ret;
	}
	
	
	public String satPrint(Pair<ParametricMessage, ParametricMessage> assignment, Domain dom){
		String ret="(and";
		Pair<ArrayList<ParametricMessage[]>,ParametricMessage> mex;
		ArrayList<ParametricMessage> temp_ikMsgs=new ArrayList<ParametricMessage>();
		Pair<ArrayList<ParametricMessage>, Pair<HashSet<ParametricMessage>,ArrayList<ParametricMessage>>> params_messages=this.ik.getParametersAndMessages();
		HashSet<ParametricMessage> messagesWithParams=params_messages.getElement1().getElement0();
		ArrayList<ParametricMessage> messagesWithoutParams=params_messages.getElement1().getElement1();
		boolean inik=false;

		//forall messages in IK that contain parameters										
		Iterator<ParametricMessage> iterIK_mexWithParams=messagesWithParams.iterator();
		ArrayList<Pair<ParametricMessage,ParametricMessage>> temp=new ArrayList<Pair<ParametricMessage,ParametricMessage>>();
		temp.add(assignment);
		
		while(iterIK_mexWithParams.hasNext()){			
			ParametricMessage currentIter=iterIK_mexWithParams.next();
			mex=dom.findConstantizedWrtAss(currentIter, temp);
			temp_ikMsgs.add(mex.getElement1());
			ret+="("+this.ik.satString(assignment)+" "+mex.getElement1().satString(assignment)+")\n";
		}

		//forall messages in domain (and forall constants)
		//if it is not a message of IK then not(DY0 m)
		Iterator<ParametricMessage> iterDom=dom.getDomainParametricMessage().iterator();
		while(iterDom.hasNext()){
			ParametricMessage current=iterDom.next();
			for(int ikit=0;ikit<temp_ikMsgs.size();ikit++){
				if(current.satString().equals(temp_ikMsgs.get(ikit).satString())){
					inik=true;
					break;
				}
			}
			if(!inik){
				for(int ikit=0;ikit<messagesWithoutParams.size();ikit++){
					if(current.satString().equals(messagesWithoutParams.get(ikit).satString())){
						inik=true;
						break;
					}
				}
				if(!inik)
					ret+="(not ("+this.ik.satString(assignment)+" "+current.satString(assignment)+"))\n";
			}
			inik=false;
		}

		for(int itC=0;itC<dom.getDomainPMConstans().size();itC++)
			for(int itC2=0;itC2<dom.getDomainPMConstans().get(itC).size();itC2++){
				ParametricMessage current=dom.getDomainPMConstans().get(itC).get(itC2);
				for(int ikit=0;ikit<temp_ikMsgs.size();ikit++){
					if(current.satString().equals(temp_ikMsgs.get(ikit).satString())){
						inik=true;
						break;
					}
				}
				if(!inik){
					for(int ikit=0;ikit<messagesWithoutParams.size();ikit++){
						if(current.satString().equals(messagesWithoutParams.get(ikit).satString())){
							inik=true;
							break;
						}
					}
					if(!inik)
						ret+="(not ("+this.ik.satString(assignment)+" "+current.satString()+"))\n";
				}
				inik=false;
			}

		//here we create the constantiziation of m
		ret+="(not ("+this.ik.satString(assignment)+" "+this.message.satString(assignment)+")));minik\n";
		temp_ikMsgs.clear();
		
		return ret;
	}
	public String satPrint(ArrayList<Pair<ParametricMessage, ParametricMessage>> assignment, Domain dom){
		String ret="(and ";
		Pair<ArrayList<ParametricMessage[]>,ParametricMessage> mex;
		ArrayList<ParametricMessage> temp_ikMsgs=new ArrayList<ParametricMessage>();
		Pair<ArrayList<ParametricMessage>, Pair<HashSet<ParametricMessage>,ArrayList<ParametricMessage>>> params_messages=
				this.ik.getParametersAndMessages();
		HashSet<ParametricMessage> messagesWithParams=params_messages.getElement1().getElement0();
		ArrayList<ParametricMessage> messagesWithoutParams=params_messages.getElement1().getElement1();
		boolean inik=false;

		//forall messages in IK that contain parameters										
		Iterator<ParametricMessage> iterIK_mexWithParams=messagesWithParams.iterator();

		while(iterIK_mexWithParams.hasNext()){			
			ParametricMessage currentIter=iterIK_mexWithParams.next();
			mex=dom.findConstantizedWrtAss(currentIter, assignment);
			temp_ikMsgs.add(mex.getElement1());
			ret+="("+this.ik.satString(assignment)+" "+mex.getElement1().satString(assignment)+")\n";
		}

		//forall messages in domain (and forall constants)
		//if it is not a message of IK then not(DY0 m)
		Iterator<ParametricMessage> iterDom=dom.getDomainParametricMessage().iterator();
		while(iterDom.hasNext()){
			ParametricMessage current=iterDom.next();
			for(int ikit=0;ikit<temp_ikMsgs.size();ikit++){
				if(current.satString().equals(temp_ikMsgs.get(ikit).satString())){
					inik=true;
					break;
				}
			}
			if(!inik){
				for(int ikit=0;ikit<messagesWithoutParams.size();ikit++){
					if(current.satString().equals(messagesWithoutParams.get(ikit).satString())){
						inik=true;
						break;
					}
				}
				if(!inik)
					ret+="(not ("+this.ik.satString(assignment)+" "+current.satString(assignment)+"))\n";
			}
			inik=false;
		}

		for(int itC=0;itC<dom.getDomainPMConstans().size();itC++)
			for(int itC2=0;itC2<dom.getDomainPMConstans().get(itC).size();itC2++){
				ParametricMessage current=dom.getDomainPMConstans().get(itC).get(itC2);
				for(int ikit=0;ikit<temp_ikMsgs.size();ikit++){
					if(current.satString().equals(temp_ikMsgs.get(ikit).satString())){
						inik=true;
						break;
					}
				}
				if(!inik){
					for(int ikit=0;ikit<messagesWithoutParams.size();ikit++){
						if(current.satString().equals(messagesWithoutParams.get(ikit).satString())){
							inik=true;
							break;
						}
					}
					if(!inik)
						ret+="(not ("+this.ik.satString(assignment)+" "+current.satString()+"))\n";
				}
				inik=false;
			}

		//here we create the constantiziation of m
		ret+="(not ("+this.ik.satString(assignment)+" "+this.message.satString(assignment)+")));minik\n";
		temp_ikMsgs.clear();
		
		return ret;
	}
	
	public String print(){return "NotMinIKstdConstraint";}
	public String print(int i){return "NotMinIKstdConstraint";}
	public String getFunctionNameFromConstraint(){
		return "";
	}
	public String getConstantsFromConstraint(){
		String ret="";
		
		//reading constants from IK of IK|-m
		Iterator<ParametricMessage> iterator=this.ik.iterator();	
		while (iterator.hasNext()){
			ParametricMessage current=iterator.next();
			
				ret+=current.getConstant()+" ";
			
		}
		//reading constants from m of IK|-m
		ret+=this.message.getConstant();
		
		return ret;
	}
	
	//old method, to remove
	public void createDomain(Domain domain){
//		this.ik.add2Domain(domain,this);
//		this.message.add2Domain(domain,this);	
	}
	
	
}