package se;

import java.util.ArrayList;
import java.util.Iterator;
import java.lang.Object;

public class ParametricEncryptedMessage extends ParametricMessage{
	
	private ParametricMessage message;
	private ParametricKey key;
	private int uid=getNewUID();
	private String name="ParametricEncryptedMessage"+this.getUID();
	
	
	public int getUID(){return this.uid;}
	public String getName(){return this.name;}
	

	
	public String print() {
		return this.getName()+" : {"+this.message.print()+ "}_{" + this.key.print()+"}";
	}
	
	public String print(int i) {
		return "error: called print(int) in ParametricEncryptedMessage";
	}
	
	public ParametricEncryptedMessage(){} // method used only for Domain construction
	public ParametricEncryptedMessage(ParametricMessage message, ParametricKey key){
		this.setFirst(message);
		this.setSecond(key);
	}
	
	public String getConstant(){
		return this.getFirst().getConstant()+" "+this.getSecond().getConstant();
	}
	
	public ArrayList<ParametricMessage> getPvtKey(ArrayList<ParametricMessage> pm_list){
		ArrayList<ParametricMessage> new_list=new ArrayList<ParametricMessage>();
		pm_list.addAll(message.getPvtKey(new_list));
		pm_list.addAll(key.getPvtKey(new_list));
		return pm_list;		
	}
	

	public ParametricMessage getFirst(){return this.message;}
	public ParametricKey getSecond(){return this.key;}
	public void setFirst(ParametricMessage message){this.message=message;}
	public void setSecond(ParametricKey key){this.key=key;}
	
	public void add2Domain(Domain domain){
		domain.add(this);
		this.message.add2Domain(domain);
		this.key.add2Domain(domain);
	}
	public void add2Domain(Domain domain,IKStdConstraint ik){
		domain.add(this,ik);
//		System.out.println("DEBUG Enc msg "+this.satString());
		this.message.add2Domain(domain,ik);
		this.key.add2Domain(domain,ik);
	}
	
	public ArrayList<ParametricMessage> Msg2List(ArrayList<ParametricMessage> list,ParametricEncryptedMessage enc,ParametricConcatenatedMessage pair,ParametricSymEncryptedMessage scrypt){
		list.add(enc);
		this.getFirst().Msg2List(list, enc, pair, scrypt);
		this.getSecond().Msg2List(list, enc, pair, scrypt);				
		return list;
	}

	public Pair<Boolean,ArrayList<ParametricMessage[]>> compare(ParametricMessage m,ArrayList<ParametricMessage[]> l){
//		System.out.println(this.satString()+" ?= "+m.satString());
		if(m.getClass().equals(ParametricEncryptedMessage.class)){
//			System.out.println("both enc");
			//they're similar
			//check its parts
			Pair<Boolean, ArrayList<ParametricMessage[]>> res=this.getFirst().compare(((ParametricEncryptedMessage) m).getFirst(), l);
//			System.out.println("its parts are:");
//			System.out.println(this.getFirst().satString()+" ?= "+((ParametricEncryptedMessage) m).getFirst().satString());
//			System.out.println(res.getElement0());
			if(res.getElement0())
				res=this.getSecond().compare(((ParametricEncryptedMessage) m).getSecond(), res.getElement1());
//			System.out.println(this.getSecond().satString()+" ?= "+((ParametricEncryptedMessage) m).getSecond().satString());
//			System.out.println(res.getElement0());
			return res;			
		}else{
			//they're not equal
//			System.out.println("NO!");
			return new Pair<Boolean, ArrayList<ParametricMessage[]>>(false,null);
		}
	}
	
	public boolean equals(Object pm){
		boolean ret=false;
		if(pm==null){return false;}
		if(pm==this){return true;}
		
		if(pm.getClass().equals(ParametricEncryptedMessage.class)){
//			System.out.println("in enc "+this.satString()+" ?= "+((ParametricEncryptedMessage)pm).satString());
//			System.out.println("SAME type");
			if(this.getFirst().equals(((ParametricEncryptedMessage)pm).getFirst()) &&
					this.getSecond().equals(((ParametricEncryptedMessage)pm).getSecond())){
//				System.out.println("in enc "+this.satString()+" ?= "+((ParametricEncryptedMessage)pm).satString());
//				System.out.println("same parts");
				ret=true;
				
			}
//			else{
//				System.out.println("uno:"+this.getFirst().hashCode()+" != "+((ParametricEncryptedMessage)pm).getFirst().satString());
//				System.out.println("due:"+this.getSecond().hashCode()+" != "+((ParametricEncryptedMessage)pm).getSecond().satString());
//				System.out.println("different parts");
//			}
//			System.out.println("RES:"+ret);
		}
		
		return ret;
	}
	
	public boolean sameEquivalenceClass(ParametricMessage pm){
		
		if(this.getClass().equals(pm.getClass()) &&
				this.getFirst().sameEquivalenceClass(((ParametricEncryptedMessage)pm).getFirst()) &&
				this.getSecond().sameEquivalenceClass(((ParametricEncryptedMessage)pm).getSecond())){
			return true;
		}
		
		return false;
			
	}
	
	public ArrayList<ParametricMessage> getParameters(ArrayList<ParametricMessage> params){
		return this.message.getParameters(this.key.getParameters(params));
	}
	
	//sat strings
	public String satString(){
		return "(crypt "+this.getFirst().satString()+" "+this.getSecond().satString()+")";
	}
	public String satString(Pair<ParametricMessage, ParametricMessage> assignment){
		return "(crypt "+this.getFirst().satString(assignment)+" "+this.getSecond().satString(assignment)+")";
	}
	public String satString(ArrayList<Pair<ParametricMessage, ParametricMessage>> assignment){
		return "(crypt "+this.getFirst().satString(assignment)+" "+this.getSecond().satString(assignment)+")";
	}
	
}
