package se;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;

public class Specification{

	//environment variables
	IKVariable IK = IKVariable.getInstance();	
	Location l1 = new Location("1");
	Program program = new Program(l1);
	ArrayList<Pair<Integer,Location>> locations=new ArrayList<Pair<Integer,Location>>();
	String name="";
	boolean goalFlag=false;
	Domain dom;
	ArrayList<Pair<FactVariable,Tuple>> facts=new ArrayList<Pair<FactVariable,Tuple>>();
	ArrayList<Evaluable> messages=new ArrayList<Evaluable>();
	Environment env=new Environment();
	ArrayList<ParametricMessage> params=new ArrayList<ParametricMessage>();
	ArrayList<Function> functions=new ArrayList<Function>();
	private boolean alreadyRemoved;
	public ArrayList<ActionEdge> edgesRemoved=new ArrayList<ActionEdge>();
	
	//	// matrix with ID that represents a function
	//	// cons stores which constants has been used (for closure)
	//	//	
	//	// _ID___cons___
	//	// | 1| c1,c2  |
	//	// | 2| c1,c2  |
	//	// | 3| c1     |
	//	// ------------
	//	ArrayList<Pair<Integer,ArrayList<ParametricConstant>>> functionClosure=new ArrayList<Pair<Integer,ArrayList<ParametricConstant>>>();


	//constructors
	public Specification(String name){
		this.locations.add(0, new Pair<Integer,Location>(new Integer(1),l1));
		this.name=name;
		this.alreadyRemoved=false;
	}
	
	public Specification(){
		new Specification("");
	}

	public void addDataState(Variable v,Evaluable e){
		this.env.addDataState(new DataState(v,e.eval(env)));
	}
	
	public ArrayList<ActionEdge> getEdgesRemoved(){return this.edgesRemoved;}
	public void addRemovedEdges(ArrayList<ActionEdge> l){
		for(int i=0;i<l.size();i++)
			this.getEdgesRemoved().add(l.get(i));
	}
	
	public ArrayList<ParametricMessage> getAgentsConstants(){
		return this.getDomain().getAgentsConstants();
	}
	
	public void removeDeadEdges(){
		//alreadyRemoved is used because the removeDeadEdges can be called by the translator.parser
		//if the spec is written in ASLan++ (so it can do the same modifications within the dot file)
		//or here if we have a SiL spec, and we don't want to do this process twice
		
		if(!this.alreadyRemoved){
			ArrayList<Location> deadLocations=new ArrayList<Location>();
			for(int i=0;i<this.locations.size();i++){
				if(!this.locations.get(i).getElement1().isGoal() && 
						this.program.outDegreeOf(this.locations.get(i).getElement1())==0){
					deadLocations.add(this.locations.get(i).getElement1());
				}
			}
			removeDeadEdges(deadLocations);
			this.alreadyRemoved=true;
		}
	}
	
	public void removeDeadEdges(ArrayList<Location> deadLocations){
		if(deadLocations.isEmpty()) return;
//		else if(deadLocations.get(0).isGoal()){deadLocations.remove(deadLocations.get(0));removeDeadEdges(deadLocations);}
		else if(this.program.outDegreeOf(deadLocations.get(0))==0){
			Location locationToRemove=deadLocations.get(0);
			ArrayList<ActionEdge> edgesToRemove=new ArrayList<ActionEdge>();
			Iterator<ActionEdge> incomingEdges=this.program.incomingEdgesOf(deadLocations.get(0)).iterator();
			while(incomingEdges.hasNext()){
				ActionEdge current=incomingEdges.next();
				if(!current.getSource().isGoal() && this.program.outDegreeOf(current.getSource())==1)
					deadLocations.add(current.getSource());
				edgesToRemove.add(current);
			}
			
			deadLocations.remove(locationToRemove);
			for(int edge=0;edge<edgesToRemove.size();edge++)
				this.program.removeEdge(edgesToRemove.get(edge));
			this.addRemovedEdges(edgesToRemove);
			edgesToRemove.clear();
			removeDeadEdges(deadLocations);
		}

	}

	//to perform a step from one node (from) to another (to) executing an action (action) 
	public boolean step(int from,int to, Action action){
		
//		System.out.println("DEBUG insert step("+from+", "+to+", "+action.print()+")");
		Location source=null;
		Location dest;
//		System.out.println("DEBUG add an action of type"+action.getClass());
		action.add2Spec(this);

		if(this.goalFlag)
			dest=new GoalLocation("l"+to);
		else
			dest=new Location("l"+to);

		ActionEdge ae = new ActionEdge(action);
		boolean ret=true;

		//find location source/from in the location set 
		if(this.locations.get(from-1).getElement0().intValue() == new Integer(from).intValue()){
			source=this.locations.get(from-1).getElement1();
		}else{
			System.out.println("ERROR in Program.java->step()");
			return false;
		}
		
		//find location dest/to in the location set, otherwise we create it
		if(this.locations.size()>=to && this.locations.get(to-1).getElement0().intValue() == new Integer(to).intValue()){
			dest=this.locations.get(to-1).getElement1();
		}else{
			this.locations.add(to-1,new Pair<Integer,Location>(new Integer(to),dest));
			program.addVertex(dest);
		}
		
		
		
		if(source!=null && dest!=null){
//			this.locations.add(to,new Pair<Integer,Location>(new Integer(to),dest));
//			program.addVertex(dest);
			if(!(program.containsEdge(ae)))
				program.addEdge(source,dest,ae);
			else
				System.out.println("Edge alreay inserted (Program->Step())!");
			
			ret=true;
		}else{
			System.out.println("From or Dest field error in Program->!");
			ret=false;
		}
		
		return ret;

	}

	//to perform a step from one node (from) to a goal location (to) executing an action (action)
	public boolean step(int from,int to, Action action,String goal){
		this.goalFlag=true;
		boolean ret=this.step(from,to,action);
		this.goalFlag=false;
		return ret;
	}

	//to run an algorithm on the spec: intraLA or fullExplore 
	//default is intraLA
	public void run(){
		this.initDomain();
		this.program.intraLA(this.dom,env);
	}
	public void run(String algorithm){
		if(!this.alreadyRemoved){removeDeadEdges();}
		this.initDomain();
		//DEBUG print
		this.program.print();
		if(algorithm.equals("intrala"))
			this.program.intraLA(this.dom,env);
		else if(algorithm.equals("explore"))
			this.program.fullExplore(this.dom,env);
		else{
			System.out.println("wrong algorithm: "+algorithm+"\n you can only choose: \n intrala \n explore");
			System.exit(0);
		}
	}

	//create the domain
	public void initDomain(){
//		Environment env=new Environment();
		ArrayList<Pair<FactVariable,ParametricTuple>> parametricFacts=
				new ArrayList<Pair<FactVariable,ParametricTuple>>(this.facts.size());
		ParametricMessageSet parametricMessages=new ParametricMessageSet();
		ArrayList<InputVariable> inputVars=new ArrayList<InputVariable>();
		ArrayList<Variable> vars=new ArrayList<Variable>();

		//System.out.println("\n facts SIZE pre:"+this.facts.size());
		//from messages to parametric messages
		for(int i=0;i<this.facts.size();i++){
			parametricFacts.add(i,new Pair<FactVariable,ParametricTuple>(this.facts.get(i).getElement0(),
					this.facts.get(i).getElement1().eval(env)));
		}
		//System.out.println("\n facts SIZE post:"+parametricFacts.size()+"\n");

		
//		System.out.println("\nECCOLI Message prima :");
//		for(int deb0=0;deb0<this.messages.size();deb0++){
//			System.out.println(this.messages.get(deb0).print());
//		}
//		System.out.println("FINITI Message prima\n");
		
		
//		System.out.println("TO ADD SIZE:"+this.messages.size());
		for(int j=0;j<this.messages.size();j++){
//			System.out.println("TO ADD J="+j);
			
//			System.out.println("DEBUG TO ADD:"+this.messages.get(j).interpolantString(env));
			
			//PARAMETER
			if(this.messages.get(j) instanceof Parameter){
				this.params.add((ParametricMessage)this.messages.get(j).eval(env));
//				System.out.println("parameter");
			}
			
			//MESSAGE SET
			else if(this.messages.get(j) instanceof MessageSet){
//				System.out.println("message set");
				Iterator<Message> iter=((MessageSet)this.messages.get(j)).iterator();
				while(iter.hasNext()){
					Message current=iter.next();
					this.messages.add(current);
//					System.out.println("DEBUG TO ADD MSGset:"+current.interpolantString(env));
//					if(current instanceof Variable){
//						parametricMessages.add(((Variable)current).getFakeParam());
//						
//					}else if(current instanceof Constant){
//						if(((Constant) current).eval4Domain()!=null){
//							parametricMessages.add(((Constant)current).eval4Domain());
////							System.out.println("DEBUG "+this.messages.get(j).eval(env).print()+" = "+this.messages.get(j).eval(env).satString());
//						}
//					}
//					else
//						parametricMessages.add((ParametricMessage)(current.eval4Domain(env)));
				}
			}
			else if(this.messages.get(j) instanceof Variable &&
					!(this.messages.get(j) instanceof InputVariable) &&
					!(this.messages.get(j) instanceof IKVariable) &&
					!(this.messages.get(j) instanceof FactVariable) &&
					!(this.messages.get(j) instanceof InputSwitchIntruderVariable)){
				vars.add(((Variable)this.messages.get(j)));
//				System.out.println("variable");
			}else if(this.messages.get(j) instanceof InputSwitchIntruderVariable ||
					this.messages.get(j) instanceof SwitchIntruderConstant || 
					this.messages.get(j) instanceof IKVariable ||
					this.messages.get(j) instanceof FactVariable ||
					this.messages.get(j) instanceof InputSwitchIntruderVariable
					){
//				this.messages.remove(j);
//				System.out.println("remove");
				//do nothing
			}
			else if(this.messages.get(j) instanceof AgentConstant){
//				System.out.println("DEBUG agent found!");
//				System.out.println("DEBUG agent="+((AgentConstant)this.messages.get(j)).eval(env).satString());
//				System.out.println("DEBUG agent pk="+((AgentConstant)(this.messages.get(j))).eval(env).getPrivateKey().satString());
//				System.out.println("DEBUG agent pvt="+((AgentConstant)(this.messages.get(j))).eval(env).getPublicKey().satString());
				parametricMessages.add(((AgentConstant)this.messages.get(j)).eval(env));
				parametricMessages.add(((AgentConstant)(this.messages.get(j))).eval(env).getPrivateKey());
				parametricMessages.add(((AgentConstant)(this.messages.get(j))).eval(env).getPublicKey());
			}
			else if(this.messages.get(j) instanceof InputVariable &&
					!(this.messages.get(j) instanceof InputIntruderVariable)){
//				System.out.println("inputvariable");
				inputVars.add((InputVariable)this.messages.get(j));
			}
			else if(this.messages.get(j) instanceof Function){
//				System.out.println("function");
				this.functions.add((Function)this.messages.get(j));
				//add to the domain all functions to create 				
			}
			else {
				parametricMessages.add((ParametricMessage)((Message)this.messages.get(j)).eval4Domain(env));
//				System.out.println("else "+this.messages.get(j).getClass());
			}
		}

		
//		System.out.println("\nECCOLI Message dopo:");
//		for(int deb0=0;deb0<this.messages.size();deb0++){
//			System.out.println(this.messages.get(deb0).print());
//		}
//		System.out.println("FINITI Message dopo\n");
//		
////		DEBUG
//		Iterator<ParametricMessage> iterpm=parametricMessages.iterator();
//		System.out.println("ECCOLI ParametricMessage:");
//		while(iterpm.hasNext()){
//			System.out.println(iterpm.next().satString());
//		}
//		System.out.println("FINITI ParametricMessage\n");


		//add params corresponding to variables
		boolean param_exists=false;
		for(int var=0;var<inputVars.size();var++){
			for(int par=0;par<this.params.size();par++){
				if(inputVars.get(var).getFakeParam().satString().equals(this.params.get(par).satString())){
					param_exists=true;
					break;
				}
			}
			if(!param_exists)
				this.params.add(inputVars.get(var).getFakeParam());

			param_exists=false;
		}

//		for(int a=0;a<this.params.size();a++)
//			System.out.println("DEBUG "+this.params.get(a).satString());

		//TEST SANITIZATION
		//slower than non-sanitized so it's better not to use it... it was...now seems to be better
		ArrayList<ParametricMessage> todelete=new ArrayList<ParametricMessage>();
		Iterator<ParametricMessage> iterMessageTest=parametricMessages.iterator();
		int howmany=0;//,n=0,m=0;
		while(iterMessageTest.hasNext()){
			//n++;
			ParametricMessage currentTest=iterMessageTest.next();
			Iterator<ParametricMessage> iterMessageTest2=parametricMessages.iterator();
			while(iterMessageTest2.hasNext()){
				//m++;
				ParametricMessage currentTest2=iterMessageTest2.next();	
				if(!todelete.contains(currentTest2)){
					if(currentTest2.sameEquivalenceClass(currentTest)){
						howmany++;
					}
					if(howmany==2){
						//System.out.println(currentTest2.getClass()+" in ["+n+"] = "+currentTest.getClass()+" in ["+m+"]");
						//System.out.println(currentTest2.print()+" in ["+n+"] = \n"+currentTest.print()+" in ["+m+"]\n");
						todelete.add(currentTest2);
						//parametricMessages.remove(currentTest);
						howmany=0;

						break;
					}
				}
			}
			howmany=0;
			//m=0;
		}
		for(int a=0;a<todelete.size();a++){
			parametricMessages.remove(todelete.get(a));
		}
		//END TEST SANITIZATION

		//		Iterator<ParametricMessage> debIter=parametricMessages.iterator();
		//		int deb=0;
		//		while(debIter.hasNext()){
		//			System.out.println("DEBUG ["+deb+"] "+debIter.next().print());
		//			deb++;
		//		}
		//		deb=0;	

//		System.out.println("DEBUG before new domain");
		this.dom=new Domain(parametricMessages,this.params);
//		System.out.println("DEBUG after new domain");

		//add tuples
		for(int tuple=0;tuple<parametricFacts.size();tuple++){
			dom.addPTuple(parametricFacts.get(tuple).getElement1());
		}

		//add fact variables
//		System.out.println("DEBUG pmF size="+parametricFacts.size());
		for(int fact=0;fact<parametricFacts.size();fact++){
			dom.addFactVar(parametricFacts.get(fact).getElement0(),parametricFacts.get(fact).getElement1().getArgs().length);
		}

		//add facts
		for(int fact=0;fact<parametricFacts.size();fact++){
			dom.addFact(new Pair<FactVariable,ParametricTuple>(parametricFacts.get(fact).getElement0(),parametricFacts.get(fact).getElement1()));
		}

		//add input vars
		for(int inputvar=0;inputvar<inputVars.size();inputvar++)
			this.dom.getDomainInputVars().add(inputVars.get(inputvar));

		//add vars
		for(int var=0;var<vars.size();var++)
			this.dom.getDomainVars().add(vars.get(var));
		
		
//		errore qui i seguito non dovrei avere eval ma gia' le funzini parametriche
//		o magari e' giusto e devo gestire il null in Domain->constantize_PTupleSet
		//add funs
		for(int fun=0;fun<this.functions.size();fun++){
			this.dom.addPfunList(((Function)this.functions.get(fun)).eval(env));
			this.dom.addPTuple(((Function)this.functions.get(fun)).getArgs().eval(env));
//			dom.addPTuple(((ParametricFunction)((Function)this.functions.get(fun)).eval(env)).getArgs(env));
//			l'errore e' qui! la satstring della tupla ritorna null
			System.out.println((((Function)this.functions.get(fun)).getArgs().eval(env).satString()));
		}
		
				
//		System.out.println("DEBUG before constantize tuple set");
		this.dom.constantize_PTupleSet();
//		//debug
//		Iterator<ParametricTuple> iter=this.dom.getPTuple().iterator();
//		while(iter.hasNext()){System.out.println("DEBUG_PT "+iter.next().satString());}
//		System.out.println("DEBUG after constantize tuple set");

		//create satString for every function
		//IMPORTANT after tuple constantization in domain
		//with closure because each uninterpreted function creates a new constant that is added to the domain
		if(this.functions.size()!=0){
			int constantDomainSize=0;
			boolean[] closed=new boolean[this.functions.size()];
			for(int i=0;i<closed.length;i++)
				closed[i]=false;
			boolean anotherRound=true;

			while(anotherRound){

				constantDomainSize=this.dom.howManyConstants();
//				System.out.println("const="+constantDomainSize);
//				for(int i=0;i<this.dom.getPMConstants().size();i++){
//					for(int j=0;j<this.dom.getPMConstants().get(i).size();j++){
//						System.out.println(this.dom.getPMConstants().get(i).get(j).getClass()+" "+this.dom.getPMConstants().get(i).get(j).satString());
//					}
//				}

				for(int f=0;f<this.functions.size();f++){
					this.createSatString(this.functions.get(f),env);

					if(constantDomainSize==this.dom.howManyConstants()){
						closed[f]=true;
					}
				}

				for(int i=0;i<closed.length;i++){
					anotherRound=false;
					if(closed[i]==false){
						anotherRound=true;
						break;
					}
				}

			}
		}
		
		
//		System.out.println("DEBUG before set domain");
		this.setDomain(dom);
//		System.out.println("DEBUG after set domain");

		//DEBUG purpose
//		this.printDomain();

	}

	//for domain purpose
	//actions use this method to add a message to the domain
	public boolean addMessage(Evaluable message){
		
		boolean ret=true;
		//questo non funziona! 
		for(int i=0;i<this.messages.size();i++){
			if(message.equals(this.messages.get(i))){
//				System.out.println("DEBUGEquals \n"+message.print()+" ?= \n"+this.messages.get(i).print()+"\n");
				ret=false;
				break;
			}
		}

		if(ret){
			this.messages.add(message);
//			System.out.println("DEBUG ADDED:"+message.print());

		}
		
		
		//if there is a function we've to add also all the messages in the tuple of the fun
		if(message instanceof Function){
//			System.out.println("DEBUG there's a function!");
			for(int mex=0;mex<((Function)message).getArgs().getArgs().length;mex++){
//				System.out.println("DEBUG elem:"+((Function)message).getArgs().getArgs()[mex].print());
				this.messages.add(((Function)message).getArgs().getArgs()[mex]);
			}
		}
		
		return ret;
	}

	//for domain purpose
	//an action uses this method to add a fact/tuple to the domain
	public boolean addFact(FactVariable fact,Tuple tuple){
		boolean ret=true;
		for(int i=0;i<this.facts.size();i++){
			if(this.facts.get(i).getElement0().equals(fact) &&
					this.facts.get(i).getElement1().equals(tuple)){
				ret=false;
				break;
			}
		}

		if(ret){
			//System.out.println("spec ADDED:"+fact.getName()+" ("+tuple.interpolantString()+")");
			this.facts.add(new Pair<FactVariable,Tuple>(fact,tuple));
			for(int mex=0;mex<tuple.getArgs().length;mex++){
				this.messages.add(tuple.getArgs()[mex]);
			}
		}
		return ret;
	}

	//this method creates sat strings for a function fun
	private void createSatString(Function fun,Environment env){
		ParametricFunction pfun=fun.eval(env);
		ParametricTuple args=fun.getArgs().eval(env);
//		System.out.println("DEBUG_Spec args: "+args.satString());
		HashSet<Pair<ArrayList<ParametricMessage[]>,ParametricTuple>> constArg=this.dom.getPTupleConstantized(args);
		
//		//debug
//		Iterator<Pair<ArrayList<ParametricMessage[]>,ParametricTuple>> iterA=constArg.iterator();
//		while(iterA.hasNext()){System.out.println("DEBUG_SpecA "+iterA.next().getElement1().satString());}
		
		//delete assignment already checked
		Iterator<Pair<ArrayList<ParametricMessage[]>,ParametricTuple>> iter=constArg.iterator();
		ArrayList<Pair<ArrayList<ParametricMessage[]>,ParametricTuple>> toDel=new ArrayList<Pair<ArrayList<ParametricMessage[]>,ParametricTuple>>();

//		System.out.println("DEBUG_Spec before, elements? "+iter.hasNext());
		//here we remove elements already checked
		while(iter.hasNext()){
			
			if(pfun.getContantiz()==null)
				break;

			Pair<ArrayList<ParametricMessage[]>,ParametricTuple> current=iter.next();
//			System.out.println("DEBUG_Spec "+current.getElement1().satString());
			
			
			for(int j=0;j<pfun.getContantiz().size();j++){
				if(pfun.getContantiz().get(j).getElement0().getElement1().satString().equals(current.getElement1().satString()) &&
						pfun.getContantiz().get(j).getElement0().getElement0().size() == current.getElement0().size()
						){
					for(int counter=0;counter<current.getElement0().size();counter++){
						if((!current.getElement0().get(counter)[0].satString().equals(pfun.getContantiz().get(j).getElement0().getElement0().get(counter)[0].satString())) &&
								(!current.getElement0().get(counter)[1].satString().equals(pfun.getContantiz().get(j).getElement0().getElement0().get(counter)[1].satString()))){
							toDel.add(current);
						}
					}
				}

			}

		}
		
//		//debug
//		Iterator<Pair<ArrayList<ParametricMessage[]>,ParametricTuple>> iterB=constArg.iterator();
//		while(iterB.hasNext()){System.out.println("DEBUG_SpecB "+iterB.next().getElement1().satString());}

		for(int c=0;c<toDel.size();c++){
			constArg.remove(toDel.get(c));
		}

//		//debug
//		Iterator<Pair<ArrayList<ParametricMessage[]>,ParametricTuple>> iterC=constArg.iterator();
//		while(iterC.hasNext()){System.out.println("DEBUG_SpecC "+iterC.next().getElement1().satString());}
		
		//structure to remember all new constants introduced
		ArrayList<ParametricConstant> newConst4Dom=new ArrayList<ParametricConstant>();
		
//		System.out.println("DEBUG_Spec before const args");
		if(constArg.size()!=0){
			//System.out.println("DEBUG_Spec after const args");
			//here we append a constant to each entry of constArg above
			//[ ((assignment1,constantizedTuple1),constant1) ;...; ((assignmentN,constantizedTupleN),constantN) ]
			ArrayList<Pair<Pair<ArrayList<ParametricMessage[]>,ParametricTuple>,ParametricConstant>> funStruct=
					new ArrayList<Pair<Pair<ArrayList<ParametricMessage[]>,ParametricTuple>,ParametricConstant>>();

			Iterator<Pair<ArrayList<ParametricMessage[]>,ParametricTuple>> iterConstArg=constArg.iterator();
			while(iterConstArg.hasNext()){
				
				Pair<ArrayList<ParametricMessage[]>,ParametricTuple> currentConst=iterConstArg.next();
				
				ParametricConstant constant=pfun.getCons(currentConst,this.dom); 
				//System.out.println("DEBUG_Spec introduced: "+constant.satString());
				Pair<Pair<ArrayList<ParametricMessage[]>,ParametricTuple>,ParametricConstant> temp=
						new Pair<Pair<ArrayList<ParametricMessage[]>,ParametricTuple>,ParametricConstant>(currentConst, constant);
				funStruct.add(temp);	

				newConst4Dom.add(constant);
				
				//add the list: assignment,pTuple(assignment),functionName,constant to the domain
				this.dom.addAss_ptuple_fname_const(currentConst, pfun.getFunctionName(), constant);
			}
			pfun.setConstantiz(funStruct);

//			//DEBUG
//			for(int d=0;d<funStruct.size();d++){
//				System.out.println("DEBUG_Spec ["+d+"] constant: "+funStruct.get(d).getElement1().satString());
//				System.out.println("DEBUG_Spec ["+d+"] tuple: "+funStruct.get(d).getElement0().getElement1().satString());
//				for(int d2=0;d2<funStruct.get(d).getElement0().getElement0().size();d2++){
//					System.out.println("DEBUG_Spec ["+d+"] elements: ("+funStruct.get(d).getElement0().getElement0().get(d2)[0]+", "+funStruct.get(d).getElement0().getElement0().get(d2)[1]+")");
//				}
//			}
//			
//			//END DEBUG
			
			//add const to dom
			for(int a=0;a<newConst4Dom.size();a++)
				this.dom.add(newConst4Dom.get(a));

			
			
			this.createInterpolantString(fun,pfun,funStruct,env);
		}

	}

	private void createInterpolantString(Function fun,ParametricFunction pfun,ArrayList<Pair<Pair<ArrayList<ParametricMessage[]>,ParametricTuple>,ParametricConstant>> funStruct,Environment env){

		Message[] fTuple=fun.getArgs().getArgs();
		ParametricMessage[] pfTuple=pfun.getArgs(env).getArgs();
		ArrayList<ArrayList<Message>> varlist=new ArrayList<ArrayList<Message>>();
		ArrayList<ArrayList<ParametricMessage>> parlist=new ArrayList<ArrayList<ParametricMessage>>();

		for(int i=0;i<fTuple.length;i++){
			varlist=fTuple[i].getVars(new ArrayList<ArrayList<Message>>());
		}
		for(int j=0;j<pfTuple.length;j++){
			parlist=pfTuple[j].getParams(new ArrayList<ArrayList<ParametricMessage>>());
		}

		//structure with [var, param],...,[var,param]
		ArrayList<Pair<Message,ParametricMessage>> var_par=new ArrayList<Pair<Message,ParametricMessage>>();
		for(int i1=0;i1<varlist.size();i1++){
			for(int i2=0;i2<varlist.size();i2++){
				for(int j1=0;j1<parlist.size();j1++){
					for(int j2=0;j2<parlist.size();j2++){

						if(((Variable)varlist.get(i1).get(i2)).getFakeParam().getClass().equals(parlist.get(j1).get(j2).getClass())){
							var_par.add(new Pair<Message,ParametricMessage>(((Message)varlist.get(i1).get(i2)),((ParametricMessage)parlist.get(j1).get(j2))));
						}
					}
				}
			}
		}


		//we substitute params with vars and ad the struct to the function
		ArrayList<Pair<Pair<ArrayList<Pair<Message,ParametricMessage>>,ParametricTuple>,ParametricConstant>> funStruct4Functions=
				new ArrayList<Pair<Pair<ArrayList<Pair<Message,ParametricMessage>>,ParametricTuple>,ParametricConstant>>();
		for(int a=0;a<funStruct.size();a++){
			ArrayList<Pair<Message,ParametricMessage>> temp=new ArrayList<Pair<Message,ParametricMessage>>();
			for(int pm=0;pm<funStruct.get(a).getElement0().getElement0().size();pm++){
				for(int b=0;b<var_par.size();b++){
					if(var_par.get(b).getElement1().satString().equals(funStruct.get(a).getElement0().getElement0().get(pm)[0].satString())){
						temp.add(new Pair<Message,ParametricMessage>(var_par.get(b).getElement0(),funStruct.get(a).getElement0().getElement0().get(pm)[1]));
					}
				}
			}

			Pair<ArrayList<Pair<Message,ParametricMessage>>,ParametricTuple> temp2=new Pair<ArrayList<Pair<Message,ParametricMessage>>,ParametricTuple>(temp,funStruct.get(a).getElement0().getElement1());
			Pair<Pair<ArrayList<Pair<Message,ParametricMessage>>,ParametricTuple>,ParametricConstant> temp3=new Pair<Pair<ArrayList<Pair<Message,ParametricMessage>>,ParametricTuple>,ParametricConstant>(temp2,funStruct.get(a).getElement1());
			funStruct4Functions.add(temp3);

			//add function assignments to dom 
//			for(int k=0;k<temp.size();k++){
				dom.addFunAss(temp);//.get(k));
//			}
		}



		fun.setConstantiz(funStruct4Functions);
	}

	//get and set methods
	public Domain getDomain(){return this.dom;}
	public Program getProgram(){return this.program;}
	private void setDomain(Domain dom){this.dom=dom;}
	public ArrayList<Pair<FactVariable,Tuple>> getFacts(){return this.facts;}

	public void printMessages(){

		for(int a=0; a<this.messages.size();a++)
			System.out.println("DEBUG ["+a+"] "+this.messages.get(a).print());
	}

	public void printDomain(){
		System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");
		//print inputVars
		System.out.println("************ Input Variables ************");
		Iterator<InputVariable> iterInputVar=this.dom.getDomainInputVars().iterator();
		while(iterInputVar.hasNext()){
			System.out.println(iterInputVar.next().interpolantString());
		}
		//print Facts + Tuple
		System.out.println("************ Fact Variables ************");
		Iterator<Pair<FactVariable,ParametricTuple>> iterFactVar=this.dom.getFacts().iterator();
		while(iterFactVar.hasNext()){
			Pair<FactVariable,ParametricTuple> current=iterFactVar.next();
			System.out.println(current.getElement0().interpolantString()+" ("+current.getElement1().satString()+")");
		}
		//print messages
		System.out.println("************ Messages ************");
		Iterator<ParametricMessage> iterMessage=this.dom.getDomainParametricMessage().iterator();
		while(iterMessage.hasNext()){
			System.out.println(iterMessage.next().satString());
		}
		//print constants
		System.out.println("************ Constants ************");
		Iterator<ArrayList<ParametricMessage>> iterConst=this.dom.getDomainPMConstans().iterator();
		while(iterConst.hasNext()){
			ArrayList<ParametricMessage> current=iterConst.next();
			for(int i=0;i<current.size();i++)
				System.out.println(current.get(i).satString());
		}
		//print params
		System.out.println("************ Parameters ************");
		for(int p=0;p<this.params.size();p++)
			System.out.println(this.params.get(p).satString());
		System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");
	}

}