package se;

import java.util.HashSet;
import java.util.Iterator;
import java.io.File;
import java.io.FileWriter;
import java.io.BufferedWriter;
import java.util.ArrayList;

public class StdConstraintSet extends HashSet<StdConstraint> {
	//list of [String new constant][String function.satString()][String constant.satString()]
	//new constant = f(constant)
	private ArrayList<String[]> functionMatrix=new ArrayList<String[]>(0);
	private String[] constantsSet;
	private int maxConstHashCode;

	public boolean addConstraint(StdConstraint constraint){
		boolean exitStatus;				
		exitStatus=this.add(constraint);
		return exitStatus;
	}

	/**
	 * The satCheck method translate the constraint set into a input (temporary) file for Z3.
	 * Then Z3 is called on this file to check if it (the file/the constraint set) is satisfiable or not.
	 * 
	 * INPUT
	 * boolean interpolation to activate interpolation procedure
	 * Environment for evaluation
	 * last action performed to create the B partition of the interpolant
	 * 
	 * OUTPUT
	 * true if StdConstraintSet is SAT
	 * false if StdConstraintSet is UNSAT
	 */

	public File translate4Z3(SymbolicDataState sds, Domain dom,int RPD, Environment env){

		try{
//			System.out.println("DEBUG translate4Z3 START");
			File constraintSetFile = File.createTempFile("Z3_ConstraintSet_",".tmp");

			//temp.deleteOnExit();
			BufferedWriter satFile = new BufferedWriter(new FileWriter(constraintSetFile));
			String z3Init="(benchmark test\n :status unknown \n :logic AUFLIA \n\n";
			z3Init+="\n \n:extrafuns((pair Int Int Int)(crypt Int Int Int)(scrypt Int Int Int) \n";
			z3Init+=ParametricMessage.getVarsList()+")";
			z3Init+="\n:extrapreds (";
			//add DYn to extrapreds
			z3Init+=StdConstraint.getdyList(); //DY0
			for(int dyn=1;dyn<=RPD;dyn++){
				z3Init+=StdConstraint.getdyList().replaceAll("DY0_", "DY"+dyn+"_");
			}	        	     
			z3Init+=dom.getFactVars4benchmark();
			//z3Init+=FactDataState.getFactsList()+")";
			z3Init+=")";
			satFile.write(z3Init);

			satFile.write("\n:formula (\n ( and ");
			satFile.write("\ntrue\n");
			satFile.write("\n");
			//here we print every constraint
			//if OPconstraint then we print an (in)equality
			//if IKconstraint we print DY0 of each message (with params) in it

			ArrayList<ArrayList<Pair<ParametricMessage,ParametricMessage>>> assignments=dom.getAssPerm();

			if(assignments.size()==0){
				//create an empty assignment otherwise the code below would not be executed
				assignments.add(new ArrayList<Pair<ParametricMessage,ParametricMessage>>());
			}

			//2. forall c constant,m (composed)message. c!=m
			satFile.write(dom.constdiffMsg());
			//3. forall m pair,crypt,scrypt in Dom. injectivity(m) 
			satFile.write(dom.printInj());
			//5. forall func(args). 
			// if(isPublic(func)) DY(arg1,...,argn)=>DY(f(arg1,...,argn)
			// if(isInvertible(func)) DY(f(arg1,...,argn)=>DY(arg1)/\.../\DY(argn)
			if(dom.getAss_ptuple_fname_const().size()!=0 && StdConstraint.getdyList()!=""){
				satFile.write(dom.publicInvertibleFunctions(RPD,env));
			}

			if(dom.getAss_ptuple_fname_const().size()!=0 || !this.isEmpty()){

				//%%%%%%%%%%%%%%%%%%%%%%%
				//A) domain OP|Fact-constraints
				//%%%%%%%%%%%%%%%%%%%%%%%
				Iterator<StdConstraint> iter=this.iterator();
				while(iter.hasNext()){
					StdConstraint currentConstraint=iter.next();
					//if there's an OPconstraint then we print it
					//if it is an IK then we save it because we want to instantiate it before printing

					//not IK constraint
				if(!(currentConstraint instanceof IKStdConstraint)){
//					System.out.println("DEBUG "+currentConstraint.toString());
						ArrayList<ParametricMessage> params=new ArrayList<ParametricMessage>();
						params=currentConstraint.getParameters(params);
						if(!params.isEmpty()){
							//if there are parameters
							Pair<Boolean[],Integer[]> assignments_bool_pair=dom.getAssPerm(params);
							Boolean[] assignments_bool=assignments_bool_pair.getElement0();
							Integer[] assignments_cols=assignments_bool_pair.getElement1();
							satFile.write("\n(or \n");
							for(int row=0;row<assignments_bool.length;row++){
								if(assignments_bool[row]){
									satFile.write("\n(and true \n");
									ArrayList<Pair<ParametricMessage,ParametricMessage>> currentAss=new ArrayList<Pair<ParametricMessage,ParametricMessage>>(assignments_cols.length);
									for(int ass_i=0;ass_i<assignments_cols.length;ass_i++){
										currentAss.add(assignments.get(row).get(assignments_cols[ass_i]));

										satFile.write("(= "+assignments.get(row).get(assignments_cols[ass_i]).getElement0().satString()+" "+
												assignments.get(row).get(assignments_cols[ass_i]).getElement1().satString()+")");
									}
									satFile.write(currentConstraint.satPrint(currentAss,dom));
									satFile.write(")");
								}
							}
							satFile.write(")");
						}else{
							//if no parameter found
							satFile.write(currentConstraint.satPrint(dom));
						}
					
					}else{
						//IK constraint
						ArrayList<ParametricMessage> params=new ArrayList<ParametricMessage>();
						Pair<ArrayList<ParametricMessage>, Pair<HashSet<ParametricMessage>,ArrayList<ParametricMessage>>> params_messages=((IKStdConstraint)currentConstraint).getikconstraint().getIK().getParametersAndMessages();
						params=params_messages.getElement0();
						HashSet<ParametricMessage> messagesWithParams=params_messages.getElement1().getElement0();
						ArrayList<ParametricMessage> messagesWithoutParams=params_messages.getElement1().getElement1();
						params=((IKStdConstraint)currentConstraint).getikconstraint().getMessage().getParameters(params);
						
						
						if(!params.isEmpty()){
							//if there are parameters
							satFile.write("(and \n");
							//write all messages without parameters
							for(int i=0;i<messagesWithoutParams.size();i++){
								satFile.write("("+((IKStdConstraint)currentConstraint).getikconstraint().getIK().satString()+" "+messagesWithoutParams.get(i).satString()+")\n");
							}

							//consider messages with parameters
//							System.out.println("DEBUG IK constraint, IK |- "+((IKStdConstraint)currentConstraint).getikconstraint().getMessage().satString());
//							for(int deb=0; deb<params.size();deb++)
//								System.out.println("DEBUG "+params.get(deb).satString());
							
							Pair<Boolean[],Integer[]> assignments_bool_pair=dom.getAssPerm(params);
							Boolean[] assignments_bool=assignments_bool_pair.getElement0();
							Integer[] assignments_cols=assignments_bool_pair.getElement1();
							satFile.write("\n(or \n");
							
//							System.out.println("DEBUG assignments size="+assignments.size());
//							dom.printAssPerm_trans();
							
							
							for(int row=0;row<assignments_bool.length;row++){
								if(assignments_bool[row]){
									satFile.write("\n(and true \n");
									ArrayList<Pair<ParametricMessage,ParametricMessage>> currentAss=new ArrayList<Pair<ParametricMessage,ParametricMessage>>(assignments_cols.length);
//									System.out.println("DEBUG currentass "+currentAss.toString());
									for(int ass_i=0;ass_i<assignments_cols.length;ass_i++){
//										System.out.println("DEBUG row="+row+" , ass_i="+ass_i+" , assignments_cols length="+assignments_cols.length+" assignment_cols[ass_i]="+assignments_cols[ass_i]);// = "+assignments.get(row).get(assignments_cols[ass_i]).toString());

										currentAss.add(assignments.get(row).get(assignments_cols[ass_i]));

										satFile.write("(= "+assignments.get(row).get(assignments_cols[ass_i]).getElement0().satString()+" "+
												assignments.get(row).get(assignments_cols[ass_i]).getElement1().satString()+")");
									}
									//begin - write IK constraint for the constraintSetFile 

									//%%%%%%%%%%%%%%%%%%%%%%%
									//B) write the constraints IKConstriants
									//%%%%%%%%%%%%%%%%%%%%%%%
									//1) forall IKn IKConstraint we define DY0n
									//in which every message is instantiated 
									//but we explicitly add the constraints that permits the instantiation.			
									//e.g 
									//(A=c & B=c & DY0(c.c) & \not(DY0(c.d) & \not(DY0(d.c)) &
									//(\not(DY0(d.d))) & (\not(DY0(k))) ... ) & DYn({c.d}k))
									//	OR
									//(A=c & B=d & DY0(c.d) & \not(DY0(c.c) & \not(DY0(d.c)) &
									//(\not(DY0(d.d))) & (\not(DY0(k))) ... ) & DYn({c.d}k))
									//	OR
									//(A=d & B=c & DY0(c.d) & \not(DY0(c.c) & \not(DY0(d.c)) &
									//(\not(DY0(d.d))) & (\not(DY0(k))) ... ) & DYn({c.d}k))
									//	OR
									//(A=d & B=d & DY0(d.d) & \not(DY0(c.c) & \not(DY0(d.c)) &
									//(\not(DY0(c.d))) & (\not(DY0(k))) ... ) & DYn({c.d}k))
									//search for IKStdConstraint

									//vars
									Pair<ArrayList<ParametricMessage[]>,ParametricMessage> mex;
									Pair<ArrayList<ParametricMessage[]>,ParametricMessage> em;
									ArrayList<ParametricMessage> temp_ikMsgs=new ArrayList<ParametricMessage>();
									boolean inik=false;

									//forall assignments
									satFile.write("\n");
//									satFile.write("(and true \n");

									//forall messages in IK that contain parameters
									//Iterator<ParametricMessage> iterIK=((IKStdConstraint)currentConstraint).getikconstraint().getIK().iterator();										
									Iterator<ParametricMessage> iterIK_mexWithParams=messagesWithParams.iterator();

									while(iterIK_mexWithParams.hasNext()){			
										ParametricMessage currentIter=iterIK_mexWithParams.next();
										mex=dom.findConstantizedWrtAss(currentIter, currentAss);
										temp_ikMsgs.add(mex.getElement1());
										satFile.write("("+((IKStdConstraint)currentConstraint).getikconstraint().getIK().satString(currentAss)+" "+mex.getElement1().satString(currentAss)+")\n");
									}

									//forall messages in domain (and forall constants)
									//if it is not a message of IK then not(DY0 m)
									Iterator<ParametricMessage> iterDom=dom.getDomainParametricMessage().iterator();
									while(iterDom.hasNext()){
										ParametricMessage current=iterDom.next();
										for(int ikit=0;ikit<temp_ikMsgs.size();ikit++){
											if(current.satString().equals(temp_ikMsgs.get(ikit).satString())){
												inik=true;
												break;
											}
										}
										if(!inik){
											for(int ikit=0;ikit<messagesWithoutParams.size();ikit++){
												if(current.satString().equals(messagesWithoutParams.get(ikit).satString())){
													inik=true;
													break;
												}
											}
											if(!inik)
												satFile.write("(not ("+((IKStdConstraint)currentConstraint).getikconstraint().getIK().satString(currentAss)+" "+current.satString(currentAss)+"))\n");
										}
										inik=false;
									}

									for(int itC=0;itC<dom.getDomainPMConstans().size();itC++)
										for(int itC2=0;itC2<dom.getDomainPMConstans().get(itC).size();itC2++){
											ParametricMessage current=dom.getDomainPMConstans().get(itC).get(itC2);
											for(int ikit=0;ikit<temp_ikMsgs.size();ikit++){
												if(current.satString().equals(temp_ikMsgs.get(ikit).satString())){
													inik=true;
													break;
												}
											}
											if(!inik){
												for(int ikit=0;ikit<messagesWithoutParams.size();ikit++){
													if(current.satString().equals(messagesWithoutParams.get(ikit).satString())){
														inik=true;
														break;
													}
												}
												if(!inik)
													satFile.write("(not ("+((IKStdConstraint)currentConstraint).getikconstraint().getIK().satString(currentAss)+" "+current.satString()+"))\n");
											}
											inik=false;
										}

									//here we create the constantiziation of m (of IK|-m)
									em=dom.findConstantizedWrtAss(((IKStdConstraint)currentConstraint).getikconstraint().getMessage(), currentAss);
									satFile.write("("+((IKStdConstraint)currentConstraint).getikconstraint().getIK().satString(currentAss).replaceAll("DY0_", "DY"+RPD+"_")+" "
											+em.getElement1().satString(currentAss)+") \n");


//									satFile.write(")\n");
									temp_ikMsgs.clear();
									

									//2)
									//- scrivere phi (senza parametri, solo costanti). restringendo il dominio
									//per ciascun IK|-m ai soli sottomessggi di m e dei messaggi in IK. Altra
									//ottimizzazione, semplificare phi e fare due fasi di
									//eliminazinoe/introduzione.
									//
									//In generale, per ogni m nel dominio avrò:
									//
									//	Per la fase di eliminazione:
									//
									//DY_i+1(m) <-> DY_i(m) V DY_i(m_1) V ... V DY_i(m_n) V ( DY_i({m}k_1 &
									//DY_i(inv(k_1)) ) V ... V ( DY_i({m}k_p & DY_i(inv(k_p)) )
									//
									//dove m_1, ..., m_n sono tutti i messaggi della forma m.m' o m'.m che
									//appartengono al dominio
									//
									//ed {m}k_1, ..., {m}k_p sono tutti i messaggi criptati con contenuto m
									//che appartengono al dominio t.c. inv(k_1), ..., inv(k_p) appartengono
									//anch'essi al dominio
									//
									//
									//	Per la fase di introduzione:
									//
									//m atomico
									//DY_i+1(m) <-> DY_i(m)
									//
									//m=m1.m2
									//DY_i+1(m) <-> DY_i(m) V ( DY_i(m1) & DY_i(m2) )
									//
									//m={m1}m2
									//DY_i+1(m) <-> DY_i(m) V ( DY_i(m1) & DY_i(m2) )
									//
									//m={m1}m2 /\ m2 symmetricKey
									//DY_i+1(m) <-> DY_i(m) V ( DY_i(m1) & DY_i(m2) )

									//add all contants to the domain
									dom=dom.addConst2Dom();
									Iterator<ParametricMessage> domIter=dom.getDomainParametricMessage().iterator();
									//System.out.println(dom.print());
									while(domIter.hasNext()){
										ParametricMessage currentM=domIter.next();
										for(int dyn=RPD;dyn>0;dyn--){
											String dy_current=((IKStdConstraint)currentConstraint).getikconstraint().getIK().satString(currentAss).replaceAll("DY0_", "DY"+(dyn)+"_");
											String dy_previous=((IKStdConstraint)currentConstraint).getikconstraint().getIK().satString(currentAss).replaceAll("DY0_", "DY"+(dyn-1)+"_");
											String m=currentM.satString(currentAss);
											String otherSideOfImplicationElimination="";
											String otherSideOfImplicationIntroduction="";

											satFile.write("\n (and");
											satFile.write("\n (implies");
											satFile.write("\n("+dy_current+" "+m+")");
											//monotonicity		
											satFile.write("\n(or \n");
											satFile.write("\n("+dy_previous+" "+m+")");
											//elimination phase
											Iterator<ParametricMessage> domIter2=dom.getDomainParametricMessage().iterator();
											//forall messege m2 in domain
											while(domIter2.hasNext()){
												ParametricMessage currentM2=domIter2.next();
												//if m2=pair
												if(currentM2.getClass().equals(ParametricConcatenatedMessage.class)){
													//if m2=m.* || m2=*.m
													if(((ParametricConcatenatedMessage)currentM2).getFirst().equals(currentM) ||
															((ParametricConcatenatedMessage)currentM2).getSecond().equals(currentM)){
														otherSideOfImplicationElimination+="\n ("+dy_previous+" "+currentM2.satString(currentAss)+")";
														satFile.write("\n ("+dy_previous+" "+currentM2.satString(currentAss)+")");
													}

													//if m2=crypt
												}else if(currentM2.getClass().equals(ParametricEncryptedMessage.class)){

													//if m2=m_k
													if(((ParametricEncryptedMessage)currentM2).getFirst().equals(currentM)){
														Iterator<ParametricMessage> domIter3=dom.getDomainParametricMessage().iterator();
														//forall m3 in domain
														while(domIter3.hasNext()){
															ParametricMessage currentM3=domIter3.next();
															//if m3=inv(k)
															if(currentM3.equals(((ParametricEncryptedMessage)currentM2).getSecond().getParametricAgent().getPrivateKey())){
																otherSideOfImplicationElimination+="\n(and ("+dy_previous+" "+currentM2.satString(currentAss)+") ("+dy_previous+" "+((ParametricEncryptedMessage)currentM2).getSecond().getParametricAgent().getPrivateKey().satString(currentAss)+") )";
																satFile.write("\n(and ("+dy_previous+" "+currentM2.satString(currentAss)+") ("+dy_previous+" "+((ParametricEncryptedMessage)currentM2).getSecond().getParametricAgent().getPrivateKey().satString(currentAss)+") )");					
															}

														}

													}

												}
												//if m2=scrypt
												else if(currentM2.getClass().equals(ParametricSymEncryptedMessage.class)){

													//if m2=m_sk
													if(((ParametricSymEncryptedMessage)currentM2).getFirst().equals(currentM)){
														Iterator<ParametricMessage> domIter4=dom.getDomainParametricMessage().iterator();
														//forall m3 in domain
														while(domIter4.hasNext()){
															ParametricMessage currentM4=domIter4.next();
															if(currentM4.equals(((ParametricSymEncryptedMessage)currentM2).getSecond())){
																otherSideOfImplicationElimination+="\n(and ("+dy_previous+" "+currentM2.satString(currentAss)+") ("+dy_previous+" "+((ParametricSymEncryptedMessage)currentM2).getSecond().satString(currentAss)+") )";
																satFile.write("\n(and ("+dy_previous+" "+currentM2.satString(currentAss)+") ("+dy_previous+" "+((ParametricSymEncryptedMessage)currentM2).getSecond().satString(currentAss)+") )");
															}
														}

													}

												}
											}

											//introduction phase
											if(currentM.getClass().equals(ParametricConcatenatedMessage.class)){
												otherSideOfImplicationIntroduction+="\n(and ("+dy_previous+" "+((ParametricConcatenatedMessage)currentM).getFirst().satString(currentAss)+")";
												otherSideOfImplicationIntroduction+="\n("+dy_previous+" "+((ParametricConcatenatedMessage)currentM).getSecond().satString(currentAss)+"))";
												satFile.write(otherSideOfImplicationIntroduction);
											}else if(currentM.getClass().equals(ParametricEncryptedMessage.class)){
												otherSideOfImplicationIntroduction+="\n(and ("+dy_previous+" "+((ParametricEncryptedMessage)currentM).getFirst().satString(currentAss)+")";
												otherSideOfImplicationIntroduction+="\n("+dy_previous+" "+((ParametricEncryptedMessage)currentM).getSecond().satString(currentAss)+"))";
												satFile.write(otherSideOfImplicationIntroduction);
											}else if(currentM.getClass().equals(ParametricSymEncryptedMessage.class)){
												otherSideOfImplicationIntroduction+="\n(and ("+dy_previous+" "+((ParametricSymEncryptedMessage)currentM).getFirst().satString(currentAss)+")";
												otherSideOfImplicationIntroduction+="\n("+dy_previous+" "+((ParametricSymEncryptedMessage)currentM).getSecond().satString(currentAss)+"))";
												satFile.write(otherSideOfImplicationIntroduction);
											}


											satFile.write("\n ) \n"); //end of or
											satFile.write(") \n"); //end of implies

											//other direction of the implication
											satFile.write("\n (implies");

											//monotonicity		
											satFile.write("\n(or ;deb4\n");
											satFile.write("\n("+dy_previous+" "+m+")");
											//elimination phase
											satFile.write(otherSideOfImplicationElimination);

											//introduction phase
											satFile.write(otherSideOfImplicationIntroduction);

											satFile.write("\n ) \n"); //end of or

											satFile.write("\n("+dy_current+" "+m+")");

											satFile.write(") \n"); //end of implies

											satFile.write(")\n"); //end of and

											otherSideOfImplicationElimination="";
											otherSideOfImplicationIntroduction="";
										}
									}


									//end - write IK constraint for the constraintSetFile
									satFile.write(")");
								}
							}
							satFile.write(")"); //or on assignments
							satFile.write(")"); //and between assignments and (IK mex) of mexs without params
						}else{
							//if no parameters found
							//%%%%%%%%%%%%%%%%%%%%%%%
							//B) write the constraints IKConstriants
							//%%%%%%%%%%%%%%%%%%%%%%%
							//1) forall IKn IKConstraint we define DY0n
							//in which every message is instantiated 
							//but we explicitly add the constraints that permits the instantiation.			
							//e.g 
							//(A=c & B=c & DY0(c.c) & \not(DY0(c.d) & \not(DY0(d.c)) &
							//(\not(DY0(d.d))) & (\not(DY0(k))) ... ) & DYn({c.d}k))
							//	OR
							//(A=c & B=d & DY0(c.d) & \not(DY0(c.c) & \not(DY0(d.c)) &
							//(\not(DY0(d.d))) & (\not(DY0(k))) ... ) & DYn({c.d}k))
							//	OR
							//(A=d & B=c & DY0(c.d) & \not(DY0(c.c) & \not(DY0(d.c)) &
							//(\not(DY0(d.d))) & (\not(DY0(k))) ... ) & DYn({c.d}k))
							//	OR
							//(A=d & B=d & DY0(d.d) & \not(DY0(c.c) & \not(DY0(d.c)) &
							//(\not(DY0(c.d))) & (\not(DY0(k))) ... ) & DYn({c.d}k))

							boolean inik=false;
							satFile.write("\n(and true\n");
							ArrayList<ParametricMessage> mexL=new ArrayList<ParametricMessage>();
							//forall messages in IK
							Iterator<ParametricMessage> iterIK=((IKStdConstraint)currentConstraint).getikconstraint().getIK().iterator();
							while(iterIK.hasNext()){
								ParametricMessage currentIter=iterIK.next();
								mexL.add(currentIter);
								satFile.write("("+((IKStdConstraint)currentConstraint).getikconstraint().getIK().satString()+" "+currentIter.satString()+")\n");
							}

							//forall messages in domain (and forall constants)
							//if it is not a message of IK then not(DY0 m)
							Iterator<ParametricMessage> iterDom=dom.getDomain().iterator();
							while(iterDom.hasNext()){
								ParametricMessage current=iterDom.next();
								for(int ikit=0;ikit<mexL.size();ikit++){
									if(current.satString().equals(mexL.get(ikit).satString())){
										inik=true;
										break;
									}
								}
								if(!inik){
									satFile.write("(not ("+((IKStdConstraint)currentConstraint).getikconstraint().getIK().satString()+" "+current.satString()+"))\n");
								}
//								else{
									inik=false;
//									satFile.write("("+((IKStdConstraint)currentConstraint).getikconstraint().getIK().satString()+" "+current.satString()+")\n");
//								}
							}

							for(int itC=0;itC<dom.getDomainPMConstans().size();itC++)
								for(int itC2=0;itC2<dom.getDomainPMConstans().get(itC).size();itC2++){
									ParametricMessage current=dom.getDomainPMConstans().get(itC).get(itC2);
									for(int ikit=0;ikit<mexL.size();ikit++)
										if(current.satString().equals(mexL.get(ikit).satString())){
											inik=true;
											break;
										}
									if(!inik)
										satFile.write("(not ("+((IKStdConstraint)currentConstraint).getikconstraint().getIK().satString()+" "+current.satString()+"))\n");
//									else {
										inik=false;
//										satFile.write("("+((IKStdConstraint)currentConstraint).getikconstraint().getIK().satString()+" "+current.satString()+")\n");
//									}
								}

							//here we create the constantiziation of m (of IK|-m)
							satFile.write("("+((IKStdConstraint)currentConstraint).getikconstraint().getIK().satString().replaceAll("DY0_", "DY"+RPD+"_")+" "+((IKStdConstraint)currentConstraint).getikconstraint().getMessage().satString()+")");

							//2)
							//- scrivere phi (senza parametri, solo costanti). restringendo il dominio
							//per ciascun IK|-m ai soli sottomessggi di m e dei messaggi in IK. Altra
							//ottimizzazione, semplificare phi e fare due fasi di
							//eliminazinoe/introduzione.
							//
							//In generale, per ogni m nel dominio avrò:
							//
							//	Per la fase di eliminazione:
							//
							//DY_i+1(m) <-> DY_i(m) V DY_i(m_1) V ... V DY_i(m_n) V ( DY_i({m}k_1 &
							//DY_i(inv(k_1)) ) V ... V ( DY_i({m}k_p & DY_i(inv(k_p)) )
							//
							//dove m_1, ..., m_n sono tutti i messaggi della forma m.m' o m'.m che
							//appartengono al dominio
							//
							//ed {m}k_1, ..., {m}k_p sono tutti i messaggi criptati con contenuto m
							//che appartengono al dominio t.c. inv(k_1), ..., inv(k_p) appartengono
							//anch'essi al dominio
							//
							//
							//	Per la fase di introduzione:
							//
							//m atomico
							//DY_i+1(m) <-> DY_i(m)
							//
							//m=m1.m2
							//DY_i+1(m) <-> DY_i(m) V ( DY_i(m1) & DY_i(m2) )
							//
							//m={m1}m2
							//DY_i+1(m) <-> DY_i(m) V ( DY_i(m1) & DY_i(m2) )
							//
							//m={m1}m2 /\ m2 symmetricKey
							//DY_i+1(m) <-> DY_i(m) V ( DY_i(m1) & DY_i(m2) )

							//add all contants to the domain
							dom=dom.addConst2Dom();
							Iterator<ParametricMessage> domIter=dom.getDomainParametricMessage().iterator();
							//System.out.println(dom.print());
							while(domIter.hasNext()){
								ParametricMessage currentM=domIter.next();
								for(int dyn=RPD;dyn>0;dyn--){
									String dy_current=((IKStdConstraint)currentConstraint).getikconstraint().getIK().satString().replaceAll("DY0_", "DY"+(dyn)+"_");
									String dy_previous=((IKStdConstraint)currentConstraint).getikconstraint().getIK().satString().replaceAll("DY0_", "DY"+(dyn-1)+"_");
									String m=currentM.satString();
									String otherSideOfImplicationElimination="";
									String otherSideOfImplicationIntroduction="";

									satFile.write("\n (and");
									satFile.write("\n (implies");
									satFile.write("\n("+dy_current+" "+m+")");
									//monotonicity		
									satFile.write("\n(or \n");
									satFile.write("\n("+dy_previous+" "+m+")");
									//elimination phase
									Iterator<ParametricMessage> domIter2=dom.getDomainParametricMessage().iterator();
									//forall messege m2 in domain
									while(domIter2.hasNext()){
										ParametricMessage currentM2=domIter2.next();
										//if m2=pair
										if(currentM2.getClass().equals(ParametricConcatenatedMessage.class)){
											//if m2=m.* || m2=*.m
											if(((ParametricConcatenatedMessage)currentM2).getFirst().equals(currentM) ||
													((ParametricConcatenatedMessage)currentM2).getSecond().equals(currentM)){
												otherSideOfImplicationElimination+="\n ("+dy_previous+" "+currentM2.satString()+")";
												satFile.write("\n ("+dy_previous+" "+currentM2.satString()+")");
											}

											//if m2=crypt
										}else if(currentM2.getClass().equals(ParametricEncryptedMessage.class)){

											//if m2=m_k
											if(((ParametricEncryptedMessage)currentM2).getFirst().equals(currentM)){
												Iterator<ParametricMessage> domIter3=dom.getDomainParametricMessage().iterator();
												//forall m3 in domain
												while(domIter3.hasNext()){
													ParametricMessage currentM3=domIter3.next();
													//if m3=inv(k)
													if(currentM3.equals(((ParametricEncryptedMessage)currentM2).getSecond().getParametricAgent().getPrivateKey())){
														otherSideOfImplicationElimination+="\n(and ("+dy_previous+" "+currentM2.satString()+") ("+dy_previous+" "+((ParametricEncryptedMessage)currentM2).getSecond().getParametricAgent().getPrivateKey().satString()+") )";
														satFile.write("\n(and ("+dy_previous+" "+currentM2.satString()+") ("+dy_previous+" "+((ParametricEncryptedMessage)currentM2).getSecond().getParametricAgent().getPrivateKey().satString()+") )");					
													}

												}

											}

										}
										//if m2=scrypt
										else if(currentM2.getClass().equals(ParametricSymEncryptedMessage.class)){

											//if m2=m_sk
											if(((ParametricSymEncryptedMessage)currentM2).getFirst().equals(currentM)){
												Iterator<ParametricMessage> domIter4=dom.getDomainParametricMessage().iterator();
												//forall m3 in domain
												while(domIter4.hasNext()){
													ParametricMessage currentM4=domIter4.next();
													if(currentM4.equals(((ParametricSymEncryptedMessage)currentM2).getSecond())){
														otherSideOfImplicationElimination+="\n(and ("+dy_previous+" "+currentM2.satString()+") ("+dy_previous+" "+((ParametricSymEncryptedMessage)currentM2).getSecond().satString()+") )";
														satFile.write("\n(and ("+dy_previous+" "+currentM2.satString()+") ("+dy_previous+" "+((ParametricSymEncryptedMessage)currentM2).getSecond().satString()+") )");
													}
												}

											}

										}
									}

									//introduction phase
									if(currentM.getClass().equals(ParametricConcatenatedMessage.class)){
										otherSideOfImplicationIntroduction+="\n(and ("+dy_previous+" "+((ParametricConcatenatedMessage)currentM).getFirst().satString()+")";
										otherSideOfImplicationIntroduction+="\n("+dy_previous+" "+((ParametricConcatenatedMessage)currentM).getSecond().satString()+"))";
										satFile.write(otherSideOfImplicationIntroduction);
									}else if(currentM.getClass().equals(ParametricEncryptedMessage.class)){
										otherSideOfImplicationIntroduction+="\n(and ("+dy_previous+" "+((ParametricEncryptedMessage)currentM).getFirst().satString()+")";
										otherSideOfImplicationIntroduction+="\n("+dy_previous+" "+((ParametricEncryptedMessage)currentM).getSecond().satString()+"))";
										satFile.write(otherSideOfImplicationIntroduction);
									}else if(currentM.getClass().equals(ParametricSymEncryptedMessage.class)){
										otherSideOfImplicationIntroduction+="\n(and ("+dy_previous+" "+((ParametricSymEncryptedMessage)currentM).getFirst().satString()+")";
										otherSideOfImplicationIntroduction+="\n("+dy_previous+" "+((ParametricSymEncryptedMessage)currentM).getSecond().satString()+"))";
										satFile.write(otherSideOfImplicationIntroduction);
									}


									satFile.write("\n ) \n"); //end of or
									satFile.write(") \n"); //end of implies

									//other direction of the implication
									satFile.write("\n (implies");

									//monotonicity		
									satFile.write("\n(or \n");
									satFile.write("\n("+dy_previous+" "+m+")");
									//elimination phase
									satFile.write(otherSideOfImplicationElimination);

									//introduction phase
									satFile.write(otherSideOfImplicationIntroduction);

									satFile.write("\n ) \n"); //end of or

									satFile.write("\n("+dy_current+" "+m+")");

									satFile.write(") \n"); //end of implies

									satFile.write(")\n"); //end of and

									otherSideOfImplicationElimination="";
									otherSideOfImplicationIntroduction="";
								}
							}
							satFile.write(")\n");	
						}

					}

				}
			}
			satFile.write("\n)));EOF");
			satFile.flush();
			satFile.close();
//			System.out.println("DEBUG translate4Z3 END");
			return constraintSetFile;
		}catch (Exception e){ 
			e.printStackTrace();
		}

		return null;
	}

	//				//collect all the parameter in the constraint set
	//				ArrayList<ParametricMessage> params=new ArrayList<ParametricMessage>();
	//				Iterator<StdConstraint> iterConstraint=this.iterator();
	//				while(iterConstraint.hasNext()){
	//					StdConstraint currentConstraint=iterConstraint.next();
	//					params=currentConstraint.getParameters(params);
	//				}
	//
	//				Pair<Boolean[],Integer[]> assignments_bool_pair=dom.getAssPerm(params);
	//				Boolean[] assignments_bool=assignments_bool_pair.getElement0();
	//				Integer[] assignments_cols=assignments_bool_pair.getElement1();
	//
	//				satFile.write("(or ;assignments\n");
	//				for(int row=0;row<assignments_bool.length;row++){
	//					if(assignments_bool[row]){
	//
	//						satFile.write("(and true \n");
	//
	//						ArrayList<Pair<ParametricMessage,ParametricMessage>> currentAss=new ArrayList<Pair<ParametricMessage,ParametricMessage>>(assignments_cols.length);
	//						for(int ass_i=0;ass_i<assignments_cols.length;ass_i++){
	//							currentAss.add(assignments.get(row).get(assignments_cols[ass_i]));
	//
	//							satFile.write("(= "+assignments.get(row).get(assignments_cols[ass_i]).getElement0().satString()+" "+
	//									assignments.get(row).get(assignments_cols[ass_i]).getElement1().satString()+")");
	//						}
	//
	//
	//
	//						//%%%%%%%%%%%%%%%%%%%%%%%
	//						//B) write the constraints IKConstriants
	//						//%%%%%%%%%%%%%%%%%%%%%%%
	//						//1) forall IKn IKConstraint we define DY0n
	//						//in which every message is instantiated 
	//						//but we explicitly add the constraints that permits the instantiation.			
	//						//e.g 
	//						//(A=c & B=c & DY0(c.c) & \not(DY0(c.d) & \not(DY0(d.c)) &
	//						//(\not(DY0(d.d))) & (\not(DY0(k))) ... ) & DYn({c.d}k))
	//						//	OR
	//						//(A=c & B=d & DY0(c.d) & \not(DY0(c.c) & \not(DY0(d.c)) &
	//						//(\not(DY0(d.d))) & (\not(DY0(k))) ... ) & DYn({c.d}k))
	//						//	OR
	//						//(A=d & B=c & DY0(c.d) & \not(DY0(c.c) & \not(DY0(d.c)) &
	//						//(\not(DY0(d.d))) & (\not(DY0(k))) ... ) & DYn({c.d}k))
	//						//	OR
	//						//(A=d & B=d & DY0(d.d) & \not(DY0(c.c) & \not(DY0(d.c)) &
	//						//(\not(DY0(c.d))) & (\not(DY0(k))) ... ) & DYn({c.d}k))
	//						//search for IKStdConstraint
	//
	//
	//						Iterator<StdConstraint> iksearch=this.iterator();
	//						while(iksearch.hasNext()){
	//							StdConstraint currentik=iksearch.next();
	//							if(currentik instanceof IKStdConstraint){
	//								//vars
	//								Pair<ArrayList<ParametricMessage[]>,ParametricMessage> mex;
	//								Pair<ArrayList<ParametricMessage[]>,ParametricMessage> em;
	//								ArrayList<ParametricMessage> temp_ikMsgs=new ArrayList<ParametricMessage>();
	//								boolean inik=false;
	//
	//								//forall assignments
	//
	//								if(assignments_bool.length==0){
	//									// NO assignments here
	//									satFile.write("(and \n");
	//									ArrayList<ParametricMessage> mexL=new ArrayList<ParametricMessage>();
	//									//forall messages in IK
	//									Iterator<ParametricMessage> iterIK=((IKStdConstraint)currentik).getikconstraint().getIK().iterator();
	//									while(iterIK.hasNext()){	
	//										ParametricMessage currentIter=iterIK.next();
	//										mexL.add(currentIter);
	//										satFile.write("("+((IKStdConstraint)currentik).getikconstraint().getIK().satString()+" "+currentIter.satString()+")\n");
	//									}
	////										for(int mexC=0;mexC<dom.findConstantized(currentIter).size();mexC++){
	////											mexL.add(dom.findConstantized(currentIter).get(mexC).getElement1());
	////											satFile.write("("+((IKStdConstraint)currentik).getikconstraint().getIK().satString()+" "+mexL.get(mexC).satString()+")\n");
	////										}
	//
	//											
	//										//forall messages in domain (and forall constants)
	//										//if it is not a message of IK then not(DY0 m)
	//										Iterator<ParametricMessage> iterDom=dom.getDomain().iterator();
	//										while(iterDom.hasNext()){
	//											ParametricMessage current=iterDom.next();
	//											for(int ikit=0;ikit<mexL.size();ikit++){
	//												if(current.equals(mexL.get(ikit))){
	//													inik=true;
	//													break;
	//												}
	//											}
	//											if(!inik){
	//												satFile.write("(not ("+((IKStdConstraint)currentik).getikconstraint().getIK().satString()+" "+current.satString()+"))\n");
	//											}else{
	//												inik=false;
	//												satFile.write("("+((IKStdConstraint)currentik).getikconstraint().getIK().satString()+" "+current.satString()+")\n");
	//											}
	//										}
	//
	//										for(int itC=0;itC<dom.getDomainPMConstans().size();itC++)
	//											for(int itC2=0;itC2<dom.getDomainPMConstans().get(itC).size();itC2++){
	//												ParametricMessage current=dom.getDomainPMConstans().get(itC).get(itC2);
	//												for(int ikit=0;ikit<mexL.size();ikit++)
	//													if(current.equals(mexL.get(ikit))){
	//														inik=true;
	//														break;
	//													}
	//												if(!inik)
	//													satFile.write("(not ("+((IKStdConstraint)currentik).getikconstraint().getIK().satString()+" "+current.satString()+"))\n");
	//												else {
	//													inik=false;
	//													satFile.write("("+((IKStdConstraint)currentik).getikconstraint().getIK().satString()+" "+current.satString()+")\n");
	//												}
	//											}
	//
	//										//here we create the constantiziation of m (of IK|-m)
	////										ArrayList<ParametricMessage> emL=new ArrayList<ParametricMessage>();
	////										for(int emC=0;emC<dom.findConstantized(((IKStdConstraint)currentik).getikconstraint().getMessage()).size();emC++){
	////											emL.add(dom.findConstantized(((IKStdConstraint)currentik).getikconstraint().getMessage()).get(emC).getElement1());
	////											satFile.write("("+((IKStdConstraint)currentik).getikconstraint().getIK().satString().replaceAll("DY0_", "DY"+RPD+"_")+" "+emL.get(emC).satString()+")");	
	////										}
	//										satFile.write("("+((IKStdConstraint)currentik).getikconstraint().getIK().satString().replaceAll("DY0_", "DY"+RPD+"_")+" "+((IKStdConstraint)currentik).getikconstraint().getMessage()+")");
	//										satFile.write(")\n");	
	////										emL.clear();	
	////									}
	//
	//									mexL.clear();
	//
	//								}else{
	//									//if there are assignments
	//									satFile.write("\n");
	//									satFile.write("(and true \n");
	//
	//									//forall messages in IK
	//									Iterator<ParametricMessage> iterIK=((IKStdConstraint)currentik).getikconstraint().getIK().iterator();
	//									while(iterIK.hasNext()){			
	//										ParametricMessage currentIter=iterIK.next();
	//										mex=dom.findConstantizedWrtAss(currentIter, currentAss);
	//										temp_ikMsgs.add(mex.getElement1());
	//										satFile.write("("+((IKStdConstraint)currentik).getikconstraint().getIK().satString(currentAss)+" "+mex.getElement1().satString(currentAss)+")\n");
	//									}
	//
	//									//forall messages in domain (and forall constants)
	//									//if it is not a message of IK then not(DY0 m)
	//									Iterator<ParametricMessage> iterDom=dom.getDomainParametricMessage().iterator();
	//									while(iterDom.hasNext()){
	//										ParametricMessage current=iterDom.next();
	//										for(int ikit=0;ikit<temp_ikMsgs.size();ikit++){
	//											if(current.equals(temp_ikMsgs.get(ikit))){
	//												inik=true;
	//												break;
	//											}
	//										}
	//										if(!inik){
	//											satFile.write("(not ("+((IKStdConstraint)currentik).getikconstraint().getIK().satString(currentAss)+" "+current.satString(currentAss)+"))\n");
	//										}
	//										inik=false;
	//									}
	//
	//									for(int itC=0;itC<dom.getDomainPMConstans().size();itC++)
	//										for(int itC2=0;itC2<dom.getDomainPMConstans().get(itC).size();itC2++){
	//											ParametricMessage current=dom.getDomainPMConstans().get(itC).get(itC2);
	//											for(int ikit=0;ikit<temp_ikMsgs.size();ikit++){
	//												if(current.equals(temp_ikMsgs.get(ikit))){
	//													inik=true;
	//													break;
	//												}
	//											}
	//											if(!inik){
	//												satFile.write("(not ("+((IKStdConstraint)currentik).getikconstraint().getIK().satString(currentAss)+" "+current.satString(currentAss)+"))\n");
	//											}
	//											inik=false;
	//										}
	//
	//									em=dom.findConstantizedWrtAss(((IKStdConstraint)currentik).getikconstraint().getMessage(), currentAss);
	//									satFile.write("("+((IKStdConstraint)currentik).getikconstraint().getIK().satString(currentAss).replaceAll("DY0_", "DY"+RPD+"_")+" "
	//											+em.getElement1().satString(currentAss)+");\n");
	//
	//
	//									satFile.write(")\n");
	//									temp_ikMsgs.clear();
	//								}
	//							}
	//						}


	//2)
	//- scrivere phi (senza parametri, solo costanti). restringendo il dominio
	//per ciascun IK|-m ai soli sottomessggi di m e dei messaggi in IK. Altra
	//ottimizzazione, semplificare phi e fare due fasi di
	//eliminazinoe/introduzione.
	//
	//In generale, per ogni m nel dominio avrò:
	//
	//	Per la fase di eliminazione:
	//
	//DY_i+1(m) <-> DY_i(m) V DY_i(m_1) V ... V DY_i(m_n) V ( DY_i({m}k_1 &
	//DY_i(inv(k_1)) ) V ... V ( DY_i({m}k_p & DY_i(inv(k_p)) )
	//
	//dove m_1, ..., m_n sono tutti i messaggi della forma m.m' o m'.m che
	//appartengono al dominio
	//
	//ed {m}k_1, ..., {m}k_p sono tutti i messaggi criptati con contenuto m
	//che appartengono al dominio t.c. inv(k_1), ..., inv(k_p) appartengono
	//anch'essi al dominio
	//
	//
	//	Per la fase di introduzione:
	//
	//m atomico
	//DY_i+1(m) <-> DY_i(m)
	//
	//m=m1.m2
	//DY_i+1(m) <-> DY_i(m) V ( DY_i(m1) & DY_i(m2) )
	//
	//m={m1}m2
	//DY_i+1(m) <-> DY_i(m) V ( DY_i(m1) & DY_i(m2) )
	//
	//m={m1}m2 /\ m2 symmetricKey
	//DY_i+1(m) <-> DY_i(m) V ( DY_i(m1) & DY_i(m2) )

	//I want to define phi for each IK
	//						Iterator<StdConstraint> iks=this.iterator();
	//						while(iks.hasNext()){
	//							StdConstraint currentik=iks.next();
	//							if(currentik instanceof IKStdConstraint){
	//								//add all contants to the domain
	//								dom=dom.addConst2Dom();
	//								Iterator<ParametricMessage> domIter=dom.getDomainParametricMessage().iterator();
	//								//System.out.println(dom.print());
	//								while(domIter.hasNext()){
	//									ParametricMessage currentM=domIter.next();
	//									for(int dyn=RPD;dyn>0;dyn--){
	//										String dy_current=((IKStdConstraint)currentik).getikconstraint().getIK().satString(currentAss).replaceAll("DY0_", "DY"+(dyn)+"_");
	//										String dy_previous=((IKStdConstraint)currentik).getikconstraint().getIK().satString(currentAss).replaceAll("DY0_", "DY"+(dyn-1)+"_");
	//										String m=currentM.satString(currentAss);
	//										String otherSideOfImplicationElimination="";
	//										String otherSideOfImplicationIntroduction="";
	//
	//										satFile.write("\n (and");
	//										satFile.write("\n (implies");
	//										satFile.write("\n("+dy_current+" "+m+")");
	//										//monotonicity		
	//										satFile.write("\n(or");
	//										satFile.write("\n("+dy_previous+" "+m+")");
	//										//elimination phase
	//										Iterator<ParametricMessage> domIter2=dom.getDomainParametricMessage().iterator();
	//										//forall messege m2 in domain
	//										while(domIter2.hasNext()){
	//											ParametricMessage currentM2=domIter2.next();
	//											//if m2=pair
	//											if(currentM2.getClass().equals(ParametricConcatenatedMessage.class)){
	//												//if m2=m.* || m2=*.m
	//												if(((ParametricConcatenatedMessage)currentM2).getFirst().equals(currentM) ||
	//														((ParametricConcatenatedMessage)currentM2).getSecond().equals(currentM)){
	//													otherSideOfImplicationElimination+="\n ("+dy_previous+" "+currentM2.satString(currentAss)+")";
	//													satFile.write("\n ("+dy_previous+" "+currentM2.satString(currentAss)+")");
	//												}
	//
	//												//if m2=crypt
	//											}else if(currentM2.getClass().equals(ParametricEncryptedMessage.class)){
	//
	//												//if m2=m_k
	//												if(((ParametricEncryptedMessage)currentM2).getFirst().equals(currentM)){
	//													Iterator<ParametricMessage> domIter3=dom.getDomainParametricMessage().iterator();
	//													//forall m3 in domain
	//													while(domIter3.hasNext()){
	//														ParametricMessage currentM3=domIter3.next();
	//														//if m3=inv(k)
	//														if(currentM3.equals(((ParametricEncryptedMessage)currentM2).getSecond().getParametricAgent().getPrivateKey())){
	//															otherSideOfImplicationElimination+="\n(and ("+dy_previous+" "+currentM2.satString(currentAss)+") ("+dy_previous+" "+((ParametricEncryptedMessage)currentM2).getSecond().getParametricAgent().getPrivateKey().satString(currentAss)+") )";
	//															satFile.write("\n(and ("+dy_previous+" "+currentM2.satString(currentAss)+") ("+dy_previous+" "+((ParametricEncryptedMessage)currentM2).getSecond().getParametricAgent().getPrivateKey().satString(currentAss)+") )");					
	//														}
	//
	//													}
	//
	//												}
	//
	//											}
	//											//if m2=scrypt
	//											else if(currentM2.getClass().equals(ParametricSymEncryptedMessage.class)){
	//
	//												//if m2=m_sk
	//												if(((ParametricSymEncryptedMessage)currentM2).getFirst().equals(currentM)){
	//													Iterator<ParametricMessage> domIter4=dom.getDomainParametricMessage().iterator();
	//													//forall m3 in domain
	//													while(domIter4.hasNext()){
	//														ParametricMessage currentM4=domIter4.next();
	//														if(currentM4.equals(((ParametricSymEncryptedMessage)currentM2).getSecond())){
	//															otherSideOfImplicationElimination+="\n(and ("+dy_previous+" "+currentM2.satString(currentAss)+") ("+dy_previous+" "+((ParametricSymEncryptedMessage)currentM2).getSecond().satString(currentAss)+") )";
	//															satFile.write("\n(and ("+dy_previous+" "+currentM2.satString(currentAss)+") ("+dy_previous+" "+((ParametricSymEncryptedMessage)currentM2).getSecond().satString(currentAss)+") )");
	//														}
	//													}
	//
	//												}
	//
	//											}
	//										}
	//
	//										//introduction phase
	//										if(currentM.getClass().equals(ParametricConcatenatedMessage.class)){
	//											otherSideOfImplicationIntroduction+="\n(and ("+dy_previous+" "+((ParametricConcatenatedMessage)currentM).getFirst().satString(currentAss)+")";
	//											otherSideOfImplicationIntroduction+="\n("+dy_previous+" "+((ParametricConcatenatedMessage)currentM).getSecond().satString(currentAss)+"))";
	//											satFile.write(otherSideOfImplicationIntroduction);
	//										}else if(currentM.getClass().equals(ParametricEncryptedMessage.class)){
	//											otherSideOfImplicationIntroduction+="\n(and ("+dy_previous+" "+((ParametricEncryptedMessage)currentM).getFirst().satString(currentAss)+")";
	//											otherSideOfImplicationIntroduction+="\n("+dy_previous+" "+((ParametricEncryptedMessage)currentM).getSecond().satString(currentAss)+"))";
	//											satFile.write(otherSideOfImplicationIntroduction);
	//										}else if(currentM.getClass().equals(ParametricSymEncryptedMessage.class)){
	//											otherSideOfImplicationIntroduction+="\n(and ("+dy_previous+" "+((ParametricSymEncryptedMessage)currentM).getFirst().satString(currentAss)+")";
	//											otherSideOfImplicationIntroduction+="\n("+dy_previous+" "+((ParametricSymEncryptedMessage)currentM).getSecond().satString(currentAss)+"))";
	//											satFile.write(otherSideOfImplicationIntroduction);
	//										}
	//
	//
	//										satFile.write("\n ) \n"); //end of or
	//										satFile.write(") \n"); //end of implies
	//
	//										//other direction of the implication
	//										satFile.write("\n (implies");
	//
	//										//monotonicity		
	//										satFile.write("\n(or");
	//										satFile.write("\n("+dy_previous+" "+m+")");
	//										//elimination phase
	//										satFile.write(otherSideOfImplicationElimination);
	//
	//										//introduction phase
	//										satFile.write(otherSideOfImplicationIntroduction);
	//
	//										satFile.write("\n ) \n"); //end of or
	//
	//										satFile.write("\n("+dy_current+" "+m+")");
	//
	//										satFile.write(") \n"); //end of implies
	//
	//										satFile.write(")\n"); //end of and
	//
	//										otherSideOfImplicationElimination="";
	//										otherSideOfImplicationIntroduction="";
	//									}
	//								}
	//							}
	//						}

	//						//END DY support for Z3
	//						satFile.write(")\n");
	//					}//close if(assignment_bool[row])
	//				}//close of assignment loop
	//
	//
	//
	//				satFile.write(")\n");
	//			}
	//
	//			//close benchmarck for z3
	//			satFile.write(")));EOF");
	//			satFile.flush();
	//			satFile.close();
	//			System.out.println("DEBUG translate4Z3 END");
	//			return temp;
	//		}catch (Exception e){ 
	//			e.printStackTrace();
	//		}
	//
	//		return null;
	//	}

	//	public boolean satCheck(boolean interpolation, Environment env,String lastActionSemantics){
	//		int satCheckResult=3; //3 default
	//		boolean satCheckResultbool;
	//		
	//		//we first create a file with the translation of the constraintSet for Z3		
	//		File temp=this.translate4Z3(dom);
	//		
	//		try{
	//			
	//			if(temp.length()==0){
	//				//check if sat input file is empty
	//				//if it is empty than the constraintSet is SAT
	//				//(DEBUG) System.out.println("EMPTY FILE, length: "+temp.length());
	//				System.out.println("WARNING - you have produced no constraints");
	//				satCheckResult=1;
	//			}else{
	//			
	//			String cmd="z3 "+ temp;
	//			Process satProcess=Runtime.getRuntime().exec(cmd);
	//			satProcess.waitFor();
	//			
	//			//BeginOf: evaluation of sat output
	//			BufferedReader b = new BufferedReader(new InputStreamReader(satProcess.getInputStream()));
	//			BufferedReader err = new BufferedReader(new InputStreamReader(satProcess.getErrorStream()));
	//			String line = null;
	//				
	//			
	//				while ( ((line=b.readLine()) != null) || ((line=err.readLine())!=null) ) {
	//					System.out.println(line);
	//					if(line.startsWith("sat")){
	//						
	//						satCheckResult=1;
	//						//break;
	//					}
	//					else if(line.startsWith("unsat")){
	//						satCheckResult=0;
	//						//break;
	//					}
	//						
	//					}
	//				try{
	//					if(satCheckResult==2){
	//						String errorlog="\nsat error: \n";
	//						BufferedReader berr = new BufferedReader(new InputStreamReader(satProcess.getErrorStream()));
	//						while ((line=berr.readLine()) != null) {
	//							errorlog+=line+"\n";	
	//							}
	//						berr.close();
	//						b.close();
	//						throw new Exception(errorlog);
	//					}
	//					else if(satCheckResult==0 && interpolation){
	//						
	//							Interpolant I=new Interpolant(temp,env,1,constantsSet,ParametricMessage.getVarsList(),lastActionSemantics);
	//							err.close();
	//							b.close();
	//					}
	//				}catch(Exception e){
	//					e.printStackTrace();
	//					System.exit(0);
	//				}
	//			err.close();
	//			b.close();
	//			
	//			//endOf: evaluation of sat output
	//			}//endOf check length of file
	//			}	catch (Exception e){ 
	//				  //System.err.println("Error: " + e.getMessage());
	//				  e.printStackTrace();
	//			}
	//		
	//		
	//		
	//		if(satCheckResult==1)satCheckResultbool=true;
	//		else satCheckResultbool=false;
	//				
	//		return satCheckResultbool;
	//	}

	/**
	 * populate function matrix with
	 * [new_value][functName][constant] that represents newValue=functName(constant)
	 * in each field
	 * 
	 * and sanitize FunctionMatrix eliminating all the field with empty value/constant
	 */

	private ArrayList<String[]> createFunctionMatrix(){
		String[] cs=this.getConstantsSet();	
		String[] functNames= new String[this.getFunctionMatrix().size()];

		for(int j=0;j<this.getFunctionMatrix().size();j++){
			functNames[j]=this.getFunctionMatrix().get(j)[1];
		}
		this.functionMatrix=new ArrayList<String[]>(0); //reset the matrix

		for(int j=0;j<functNames.length;j++)
			for(int i=0;i<cs.length;i++){

				String[] temp=new String[3];
				temp[0]=""+this.maxConstHashCode+1;
				temp[1]=functNames[j];
				temp[2]=cs[i];
				//(DEBUG) System.out.println("fm: "+this.getFunctionMatrix().get(i));
				this.maxConstHashCode++;
				this.getFunctionMatrix().add(temp);
			}
		//(DEBUG) this.printFunctionMatrix();
		return getFunctionMatrix();
	}

	/**
	 * add a new element to FunctionMatrix
	 * if the pair function_name,constant is not already present
	 * Note: this method does not add a new value to function_name(constant)
	 */
	private void addFunctionToMatrix(String[] functionMatrixElement){
		Iterator<String[]> iter=this.getFunctionMatrix().iterator();
		while(iter.hasNext()){
			String[] current=iter.next();
			if(current[1].equals(functionMatrixElement[1]) && current[2].equals(functionMatrixElement[2]))
				return;
		}
		this.getFunctionMatrix().add(functionMatrixElement);
	}

	/**
	 * reuturns the functionMatrix
	 */
	public ArrayList<String[]> getFunctionMatrix(){
		return this.functionMatrix;
	}

	//	/**
	//	 * this method has multiple functions:
	//	 * 1. craete the constantsSet
	//	 * 2. set the maximum hashcode number 
	//	 * 3. add all the functions to functionMatrix (FunctionMatrix initialization)
	//	 * 	
	//	 * WARNING 
	//	 * the constants set containts a lot of repetitions. It should be sanitized
	//	 */
	//	private void setConstantsSet(){
	//		StdConstraint current;
	//		Iterator<StdConstraint> iter=this.iterator();
	//		String[] constants;
	//		String[] functName;
	//		String temp="";
	//		
	//		while (iter.hasNext()){	
	//			current = iter.next();			
	//			functName=current.getFunctionNameFromConstraint().split(" ");
	//			//(DEBUG)System.out.println("\n\nDEBUG\nlength nomi funzioni: "+functName.length+" size funcMatrix: "+this.getFunctionMatrix().size());
	//
	//			if(functName.length>=0)
	//			for(int i=0;i<functName.length;i++){
	//				boolean add=true;
	//				//(DEBUG)System.out.println("nome funzione "+i+" : "+functName[i]);
	//				for(int j=0;j<this.getFunctionMatrix().size();j++)
	//					if(this.getFunctionMatrix().get(j)[1].equals(functName[i]))
	//						add=false;
	//				if(add){
	//					String[] matrixElement={"",functName[i],""};
	//					this.addFunctionToMatrix(matrixElement);
	//				}
	//			}
	//			//(DEBUG)System.out.println("length nomi funzioni: "+functName.length+" size funcMatrix: "+this.getFunctionMatrix().size());
	//			//(DEBUG)this.printFunctionMatrix();
	//			//(DEBUG)System.out.println("DEBUG END\n");
	//			
	//			constants=current.getConstantsFromConstraint().split(" ");
	//			for(int i=0;i<constants.length;i++){	
	//				//(DEBUG)System.out.println("current: "+constants[i]);
	//				//(WW)(MARCOR)here some field of the array constants are empty...why?
	//				if(!(constants[i].equals("")) && constants[i].charAt(0)<65 ){
	//					String[] tempArray=temp.split(" ");
	//					boolean repetition=false;
	//					for(int k=0;k<tempArray.length;k++)
	//						if(tempArray[k].equals(constants[i])){repetition=true;break;}
	//					if(!repetition)
	//						temp+=constants[i]+" ";
	//					if(this.maxConstHashCode<Integer.parseInt(constants[i]))
	//						this.maxConstHashCode=Integer.parseInt(constants[i]);
	//					
	//				}else if(!(constants[i].equals("")) && ( constants[i].equals("true") || constants[i].equals("false"))){
	//					String[] tempArray=temp.split(" ");
	//					boolean repetition=false;
	//					for(int k=0;k<tempArray.length;k++)
	//						if(tempArray[k].equals(constants[i])){repetition=true;break;}
	//					
	//					if(!repetition)
	//						temp+=constants[i]+" ";						
	//				}
	//			}
	//			
	//			
	//		}
	//		this.constantsSet=temp.split(" ");
	//	}

	public String[] getConstantsSet(){
		return this.constantsSet;
	}

	public String functionMatrixEval(String functionName,String constant){
		Iterator<String[]> iter=this.getFunctionMatrix().iterator();
		while(iter.hasNext()){
			String[] current=iter.next();
			if(current[1].equals(functionName)&& current[2].equals(constant))
				return current[0];
		}
		//error
		return null;
	}


	//a method for testing
	public void printConstantsSet(){
		System.out.println("print constants set: "+this.getConstantsSet().length);
		for(int i=0;i<this.getConstantsSet().length;i++){
			String temp=this.getConstantsSet()[i];
			System.out.println(temp+" , ");
		}

	}

	//a method for testing
	public void printFunctionMatrix(){
		System.out.println("print function_matrix of size: "+this.getFunctionMatrix().size());
		for(int i=0;i<this.getFunctionMatrix().size();i++){
			System.out.println("constant: "+this.getFunctionMatrix().get(i)[0]+" ,functionName: "+this.getFunctionMatrix().get(i)[1]+" ,newvalue: "+this.getFunctionMatrix().get(i)[2]);
		}

	}



	// methods used just for testing

	public void print(){
		int i=0;
		StdConstraint current;
		Iterator<StdConstraint> iterator=this.iterator();			
		System.out.println("Elements of the StdConstraintSet " + this.hashCode());
		while (iterator.hasNext()){
			i++;
			current = iterator.next();
			System.out.println(i+"- "+current.print());
		}
		System.out.println("------------end of StdConstraintSet "+ this.hashCode()+" -------------");
		System.out.println();
	}

	public void printAsTree(){
		int i=0;
		StdConstraint current;
		Iterator<StdConstraint> iter=this.iterator();			
		System.out.println("Elements of the StdConstraintSet " + this.hashCode());
		while (iter.hasNext()){
			i++;
			current = iter.next();
			System.out.println("("+i+") textual:");
			System.out.println(current.print());
			System.out.println();
			System.out.println("("+i+") as a tree:");
			System.out.println(current.print(0));
		}
		System.out.println("------------end of StdConstraintSet "+ this.hashCode()+" -------------");
		System.out.println();
	}	

}
