package se;

import java.util.HashSet;
import java.util.Iterator;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.nio.file.*;
import java.io.InputStreamReader;
import java.nio.file.FileSystems;
import java.nio.charset.Charset;

public class SymbolicDataState {

	private HashSet<Parameter> parameterSet;	// do we really need this ParameterSet?
	private ConstraintSet constraintSet;
	private Environment environment;
	private File constraintSetFile;
	private File sdsFile;
	private String v_eq_EOfv="";
	private String[] constantSet;
	private int RPD;

	public HashSet<Parameter> getParameterSet(){return parameterSet;};
	public ConstraintSet getConstraintSet(){return constraintSet;};
	public Environment getEnvironment(){return environment;};

	public SymbolicDataState(HashSet<Parameter> pS, ConstraintSet cS, Environment e, int RPD){
		parameterSet = pS;
		constraintSet = cS;
		environment = e;
		this.RPD=RPD; 
	}

	public void addDataState2Environment(DataState ds){
		environment.addDataState(ds);	
	}
	public void removeDataStateFromEnvironment(DataState ds){
		environment.remove(ds);	
	}

	public void updateIKinEnvironment(DataState ds){
		environment.updateIK(ds);
	}

	public void addConstraint2ConstraintSet(Constraint constraint){
		constraintSet.addConstraint(constraint);	
	}

	public void setConstraintSetFile(File translatedConstraint){
		this.constraintSetFile=translatedConstraint;
	}
	public File getConstraintSetFile(Domain dom, Environment env){
		if(this.constraintSetFile==null){
			this.translateConstraintSet4Z3(dom, env);
		}
		return this.constraintSetFile;
	}
	public File getSdsFile(Domain dom,Environment env){
		if(this.sdsFile==null){
			this.translateSDS(dom,env);
		}
		return this.sdsFile;
	}
	public String getV_eq_EV(Domain dom){

		if(this.v_eq_EOfv==""){
			this.v_eq_EOfv(dom);

		}		
		return this.v_eq_EOfv;
	}

	public int getRPD(){return this.RPD;}

	public void setConstantSet(String[] constantSet){this.constantSet=constantSet;}
	public String[] getConstantSet(){return this.constantSet;}

	//the following method translate the constraint set into a file (named Z3_input*)
	//in /tmp that Z3 can understand and validate
	//and returns the File	
	//(DY intruder is hard-coded in the obtained file)
	public File translateConstraintSet4Z3(Domain dom, Environment env){
//		System.out.println("DEBUG translateConstraintSet4Z3");
		StdConstraintSet scs=this.getConstraintSet().standardize();
		scs.getConstantsSet();
		File translatedConstraint=scs.translate4Z3(this, dom, this.getRPD(), env);
		this.setConstraintSetFile(translatedConstraint);
		return translatedConstraint;
	}

	//forall v in Environment. v=E(v)
	//TODO:
	//if v=IKVariable
	// forall m in v
	//  (IK m)
	public void v_eq_EOfv(Domain dom){
		//(WW) the part of IKVar should be eliminated because already present from z3 file
//		System.out.println("DEBUG v=E(v) START");
		String ret="(and true\n";

		Iterator<DataState> iterator=this.getEnvironment().iterator();	

		while (iterator.hasNext()){
			DataState currentds=iterator.next(); 
			if(currentds.getVariable() instanceof IKVariable){
//				ret+="\n;DEBUG IKVar=E(IKVar)\n";
				ret+=((ParametricMessageSet)currentds.getValue()).ret_e_eq_EOfv(dom)+" \n";
			}else if(currentds.getVariable() instanceof FactVariable){
//				ret+="\n;DEBUG FactVar=E(FactVar)\n";
				ret+=((ParametricTupleSet)(currentds.getValue())).ret_e_eq_EOfv(dom,(FactVariable)currentds.getVariable())+"\n";
			}else{
//				ret+="\n;DEBUG else=E(else)\n";
				ret+=(" (= "+currentds.getVariable().interpolantString()+"  "+currentds.getValue().satString()+" )  \n");
			}
		}
//		System.out.println("DEBUG v=E(v) END");
		this.v_eq_EOfv=ret+" )\n";
	}


	public void translateSDS(Domain dom, Environment env){
		boolean add=false;
		try{
			File sdsFile = File.createTempFile("Z3_SDS_",".tmp");

			BufferedReader reader = Files.newBufferedReader(FileSystems.getDefault().getPath(this.getConstraintSetFile(dom,env).getAbsolutePath()),Charset.defaultCharset());
			BufferedWriter writer = Files.newBufferedWriter(FileSystems.getDefault().getPath(sdsFile.getAbsolutePath()), Charset.defaultCharset());
			String line = null;

			while ((line=reader.readLine())!=null) {
				if(add){
					writer.write(line+"\n");
					if(!this.getV_eq_EV(dom).equals("(and  )\n")){						
						writer.write(this.getV_eq_EV(dom));
					}
					add=false;
				}else if(line.startsWith(":formula")){
					writer.write(line+"\n");
					add=true;
				}else if(line.startsWith(" :logic AUFLIA")){
					writer.write(line);					
					writer.write("\n:extrafuns("+Message.getVarsList()+")\n");
					writer.write("\n:extrapreds((IK Int))\n");

				}
				else{
					writer.write(line+"\n");
				}
			}
			this.sdsFile=sdsFile;

			reader.close();
			writer.flush();
			writer.close();


		}catch (Exception e){
			e.printStackTrace();
		}

	}

	public boolean isBlocked(String annotation, Domain dom, Environment env,Program program){
		boolean ret=true;

		try{
			File blockedFile = File.createTempFile("Z3_BLOCKED_ANNOTATION_",".tmp");
			BufferedReader reader = Files.newBufferedReader(FileSystems.getDefault().getPath(this.getSdsFile(dom,env).getAbsolutePath()),Charset.defaultCharset());
			BufferedWriter writer = Files.newBufferedWriter(FileSystems.getDefault().getPath(blockedFile.getAbsolutePath()), Charset.defaultCharset());

			String line = null;

			while ((line=reader.readLine())!=null) {
				if(!(line.startsWith(":formula"))&&
						!(line.endsWith(";EOF"))){
					writer.write(line+"\n");
				}else if(line.endsWith(";EOF")){
					writer.write(")"+annotation+"))))\n");
				}else{
					writer.write(line+"\n");
					writer.write("(not(implies \n");
				}
			}			

			reader.close();
			writer.flush();
			writer.close();

			ret=this.satCheck(blockedFile,program);
//			blockedFile.delete();
			
		}catch (Exception e){
			e.printStackTrace();
		}

		return !ret;
	}

	public boolean satCheck(File temp, Program program){
		int satCheckResult=3; //3 default
		double startTime=0,stopTime=0;
		boolean satCheckResultbool;
		System.out.println(temp.getPath());
		try{

			if(temp.length()==0){
				//check if sat input file is empty
				System.out.println("WARNING - you are checking an empty file");
				satCheckResult=1;
			}else{

				String cmd="z3 "+ temp;
								
				startTime=System.nanoTime();
				Process satProcess=Runtime.getRuntime().exec(cmd);
							
				//BeginOf: evaluation of sat output
				BufferedReader b = new BufferedReader(new InputStreamReader(satProcess.getInputStream()));
				BufferedReader err = new BufferedReader(new InputStreamReader(satProcess.getErrorStream()));
				String line = null;

				while ( ((line=b.readLine()) != null) || ((line=err.readLine())!=null) ) {
					if(line.startsWith("sat")){
						stopTime=System.nanoTime();	
						satCheckResult=1;
						break;
					}
					else if(line.startsWith("unsat")){
						stopTime=System.nanoTime();	
						satCheckResult=0;
						break;
					}else 
						satCheckResult=2;

				}
				satProcess.destroy();
				satProcess.waitFor();

				try{
					if(satCheckResult==2){
						String errorlog="\nsat error: "+temp+"\n";
						BufferedReader berr = new BufferedReader(new InputStreamReader(satProcess.getErrorStream()));
						while ((line=berr.readLine()) != null) {
							errorlog+=line+"\n";	
						}
						berr.close();
						b.close();
						throw new Exception(errorlog);
					}
				}catch(Exception e){
					e.printStackTrace();
					System.exit(0);
				}
				
				err.close();
				b.close();
				
				//endOf: evaluation of sat output
			}//endOf check length of file
		}	catch (Exception e){ 
			e.printStackTrace();
		}


		if(satCheckResult==1){
			satCheckResultbool=true;
			program.setZ3satTime((stopTime-startTime)/(double)Math.pow(10,6));
			program.addZ3satExecTime((stopTime-startTime)/(double)Math.pow(10,6));
		}else{ 
			satCheckResultbool=false;
			program.setZ3unsatTime((stopTime-startTime)/(double)Math.pow(10,6));
			program.addZ3unsatExecTime((stopTime-startTime)/(double)Math.pow(10,6));
		}

		
		return satCheckResultbool;
	}

	SymbolicDataState getCopy() {
		SymbolicDataState returnSds;
		HashSet<Parameter> parSet = new HashSet<Parameter>();
		returnSds = new SymbolicDataState(parSet,this.getConstraintSet().getCopy(),this.getEnvironment().getCopy(),this.getRPD());
		// you might need to copy also files, strings, etc.			
		return returnSds;
	}

	public void print() {
		System.out.println("---------  SymbolicDataState "+this.hashCode()+ " with ConstraintSet "+this.constraintSet.hashCode()+ " and Environment "+this.environment.hashCode());
		constraintSet.print();
		environment.print();
	}	

}

