package translator;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
//import java.io.InputStreamReader;
//import java.nio.charset.Charset;
//import java.nio.file.FileSystems;
//import java.nio.file.Files;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.regex.*;
import java.util.ArrayList;

import javax.swing.JFileChooser;

import se.AgentConstant;
import se.Message;
import se.Specification;

/**
 * Main class that can be used to launch SPiM.
 * The launch configuration is:
 * translator.Main [ fileName ] [ -gv (--graphviz) [dir] ] [--intrala || --explore] [--allSessions]
 * If fileName is present it will be interpreted as the ASLan++ specification input file.
 * Otherwise a GUI will allow to choose the ASLan++ specification the input file.
 * If -gv or --graphviz option is present the program graph is generated;
 * in this case it is possible to specify the .dot and .png output directory from command line or,
 * otherwise, a GUI will allow to choose the output directory. 
 *
 * @author Fabio Pettenuzzo
 * @author Marco Palamà
 * @version 1.0, 12/13
 */
public class Main {

	public static void main(String args[]) {

		File file = null;
		File directory = null;
		String dir = "";
		boolean gv=false,explore_algorithm=false,intrala_algorithm=false,allSessions=false;
		
		if( args.length > 0 && !("-".equals(args[0].charAt(0)+"")) ) file = new File(args[0]);
		if( args.length > 0 && Arrays.asList(args).contains("--graphviz") ) args[Arrays.asList(args).indexOf("--graphviz")] = "-gv";		
		if( args.length > 0 && ( Arrays.asList(args).contains("-gv") ) ) {
			gv=true;
			int indexGv = Arrays.asList(args).indexOf("-gv");
			if( indexGv != args.length - 1 && !("-".equals(args[indexGv+1].charAt(0)+"")) ) {
				dir = args[indexGv+1];
			} else {
				// Create file chooser
				JFileChooser folder = new JFileChooser();
				folder.setDialogTitle("Output Graphviz directory");
				folder.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
				int returnVal = folder.showOpenDialog(null);
				if (returnVal == JFileChooser.APPROVE_OPTION) { // Open folder
					directory = folder.getSelectedFile();
					dir = directory.getAbsolutePath()+"/";
				} else { // Exit with exit status 0 (No error).
					System.exit(0);
				}
			}
		}
		if( args.length > 0 && ( Arrays.asList(args).contains("--explore") ) ) {
			explore_algorithm=true;
		}
		if( args.length > 0 && ( Arrays.asList(args).contains("--intrala") ) ) {
			intrala_algorithm=true;
		}
		if(!intrala_algorithm && !explore_algorithm){
			explore_algorithm=false;
			intrala_algorithm=true;
		}
		if( args.length > 0 && ( Arrays.asList(args).contains("--allSessions") ) ) {
			allSessions=true;
		}
		if( file==null ) {
			// Create file chooser
			JFileChooser fc = new JFileChooser();
			// Set the ASLan++ filter
			fc.setFileFilter(new ASLanppFilter());
			// Display file chooser
			int returnVal = fc.showOpenDialog(null);
			if (returnVal == JFileChooser.APPROVE_OPTION) { // Open file
				file = fc.getSelectedFile();
			} else { // Exit with exit status 0 (No error).
				System.exit(0);
			}
		}

		FileInputStream fis=null;
		
		try { // Get FileinputStream from file.
			fis = new FileInputStream(file);
		} catch (FileNotFoundException e) {
			System.out.println("File not found: "+args[0]);
			System.exit(3);
		}
		// Create BufferedInputStream from FileinputStream.
		BufferedInputStream bis = new BufferedInputStream(fis);
		
		// Create a new parser instance that reads from the BufferedInputStream.
		Parser parser = new Parser(bis, gv, dir);
		// Translate the input ASLan++ specification into a SPiM specification.
		Specification spec = parser.translate();

		//MarcoR
		//sessionNumber = how many sessions per specification
		//agentsInASession = how many agents per session
		//allSessions_agents = all the permutations of session instantiation (agentsInASession agents involved)
		
		if(allSessions){
			int sessionNumber=parser.getHowManySessions();
			int agentsInASession=parser.getHowManyAgentsPerSessions();

//			System.out.println("DEBUG There are "+sessionNumber+" sessions");
//			System.out.println("DEBUG Each session has "+agentsInASession+" entities");


//			System.out.println("\nDEBUG agents are:");
			ArrayList<String> agents=new ArrayList<String>();
			boolean already=false;

			Iterator<String> entitiesIterator = parser.getEntities().keySet().iterator();
			while (entitiesIterator.hasNext()) {
				String current_key = entitiesIterator.next();
				Iterator<String> current_entitySymbols_iterator= parser.getEntities().get(current_key).getSymbols().keySet().iterator();
				while(current_entitySymbols_iterator.hasNext()){
					String current_symbolKey=current_entitySymbols_iterator.next();
					Message value = parser.getEntities().get(current_key).getSymbols().get(current_symbolKey);
					if(value.getClass().equals(AgentConstant.class)){
						if(value  instanceof AgentConstant){
							if(((AgentConstant)value).toString().matches("[^\\.]*\\..*")){
								String regex = ".+\\.(.*)\\s*";
								Pattern pattern = Pattern.compile(regex);
								Matcher matcher = pattern.matcher(((AgentConstant)value).toString());

								while (matcher.find()) {
//									System.out.println("DEBUG agent (Entity "+current_key+"): "+matcher.group(1));
									//checks if already present
									for(int i=0;i<agents.size();i++){
										if(agents.get(i).equals(matcher.group(1)))
											already=true;
									}
									if(!already)
										agents.add(matcher.group(1));
									already=false;
								}
							}else{
//								System.out.println("DEBUG agent (Entity "+current_key+"): "+((AgentConstant)value).toString());
								//checks if already present
								for(int i=0;i<agents.size();i++){
									if(agents.get(i).equals(((AgentConstant)value).toString()))
										already=true;
								}
								if(!already)
									agents.add(((AgentConstant)value).toString());
								already=false;

							}
						}
					}
				}

			}

//			System.out.print("DEBUG In the agents struct we now have: ");
//			for(int i=0;i<agents.size();i++)
//				System.out.print(agents.get(i)+" ");
//			System.out.println();

			//in agents I have all the agent constants
//			System.out.println("DEBUG all session instantiations ("+agentsInASession+" agents per session)");
			String[][] allSessions_agents=combinations(agents.toArray(new String[agents.size()]),agentsInASession);

			//print session instantiations
//			for(int i=0;i<allSessions_agents.length;i++){
//				System.out.print("any Session(");
//				for(int j=0;j<allSessions_agents[i].length;j++){
//					System.out.print(allSessions_agents[i][j]);
//					if(j!=(allSessions_agents[i].length)-1)
//						System.out.print(", ");							
//				}
//				System.out.println(");");
//			}

//			System.out.println("DEBUG all sessions per specification ("+sessionNumber+" per spec)");
			String[][][] allSessionCombinations=sessionCombinations(allSessions_agents,sessionNumber);

//			for(int deb=0;deb<allSessionCombinations.length;deb++)
//				for(int deb2=0;deb2<allSessionCombinations[deb].length;deb2++)
//					for(int deb3=0;deb3<allSessionCombinations[deb][deb2].length;deb3++)
//						System.out.println("DEBUG allSC ["+deb+" "+deb2+" "+deb3+"]"+allSessionCombinations[deb][deb2][deb3]);

			String[] sessionsToAdd=new String[allSessionCombinations.length];
//			System.out.println("DEBUG allSessionCombinations - rows: "+allSessionCombinations.length+" cols:"+allSessionCombinations[0].length);


			for(int i=0;i<allSessionCombinations.length;i++){
				sessionsToAdd[i]="";
				for(int j=0;j<allSessionCombinations[i].length;j++){
					sessionsToAdd[i]+="any Session(";
					for(int k=0;k<allSessionCombinations[i][j].length;k++){
						sessionsToAdd[i]+=allSessionCombinations[i][j][k];
						if(k!=allSessionCombinations[i][j].length-1)
							sessionsToAdd[i]+=", ";
					}
					sessionsToAdd[i]+=");\n";
				}

			}

//			System.out.println("\n\nDEBUG String\n\n");
			for(int i=0;i<sessionsToAdd.length;i++){
				System.out.println("["+i+"] ");
				System.out.println(sessionsToAdd[i]);
			}

			allSessions_agents=null;
//			allSessionCombinations=null;
			System.gc();

			try {
				BufferedReader br = new BufferedReader(new FileReader(file));
				File allSpecDir = new File("allSpec_"+file.getName());
				File allSpecDir_graphs = new File("allSpec_"+file.getName()+"_graphs");
				
				// if the directory does not exist, create it
				if(!allSpecDir.exists()) {
					//System.out.println("creating directory: " + allSpecDir);
					allSpecDir.mkdir();
				}
				if(!allSpecDir_graphs.exists())
					allSpecDir_graphs.mkdir();
				
				//read the spec and save all the lines but session instantiations
				String line="";
				String buffer_pre_sessions="";
				String buffer_post_sessions="";
				boolean pre=true;
				while ((line=br.readLine()) != null) {
					if(!line.matches("\\s*any Session.*")){
						if(pre)
							buffer_pre_sessions+=line+"\n";
						else
							buffer_post_sessions+=line+"\n";
					}else
						pre=false;
				}
				br.close();
				
				//write
				short toBeValidated=0;
				boolean flag_toBeValidated=true;
				for(int sessions=0;sessions<sessionsToAdd.length;sessions++){
					try {
						//do not create files for sessions without at least one intruder
						for(int s=0;s<allSessionCombinations[sessions].length;s++){
							for(int agent_s=0;agent_s<allSessionCombinations[sessions][s].length;agent_s++)
								if(allSessionCombinations[sessions][s][agent_s].equals("i"))
									toBeValidated++;
							if(toBeValidated==allSessionCombinations[sessions][s].length || toBeValidated==0){
								flag_toBeValidated=false;
								break;
							}
							toBeValidated=0;
						}
						if(flag_toBeValidated){
							File file_out = new File(allSpecDir+"/"+file.getName().replaceFirst("[.][^.]+$", "")+"_"+sessions+".aslan++");
							BufferedWriter output = new BufferedWriter(new FileWriter(file_out));
							output.write(buffer_pre_sessions);
							output.write(sessionsToAdd[sessions]);
							output.write(buffer_post_sessions);
							output.close();
						}else{
							System.out.println("DEBUG SKIPPED "+allSpecDir+"/"+file.getName().replaceFirst("[.][^.]+$", "")+"_"+sessions+".aslan++ because no intruder or too many found in session instantiation :)");
						}
						toBeValidated=0;
						flag_toBeValidated=true;
					} catch (Exception e) {
						e.printStackTrace();
					}
				}
				
				
				// run spim on all the specs
				//if you want to run spim on the original spec first, uncomment the following lines 
//				if(intrala_algorithm)
//					spec.run("intrala");
//				else
//					spec.run("explore");
				spec=null;
				dir=null;
				
				parser=null;
				System.gc();
				File folder = new File(allSpecDir+"/");
				File[] listOfFiles = folder.listFiles();

				for (File spec2validate : listOfFiles) {
					if (spec2validate.isFile()) {
						System.out.println("I'm validating:  "+spec2validate.getAbsolutePath());
						dir=allSpecDir.getAbsolutePath();
						String prefix=spec2validate.getName().split("/")[spec2validate.getName().split("/").length-1];
						spec = new Parser(new BufferedInputStream(new FileInputStream(spec2validate)), gv, allSpecDir_graphs.getAbsolutePath()+"/"+prefix).translate();
						if(intrala_algorithm){
							System.out.println("**** IntraLA ****");
							spec.run("intrala");
						}
						if(explore_algorithm){
							System.out.println("**** Explore ****");
							if(intrala_algorithm){
								spec=null;
								System.gc();
								spec = new Parser(new BufferedInputStream(new FileInputStream(spec2validate)), false, "").translate();
							}
							spec.run("explore");
						}
					}
				}
				
			}catch (Exception e) {
				e.printStackTrace();
				System.exit(3);
			}
			System.exit(0);



			
		}else{
		//MarcoR
		
			if(intrala_algorithm){
				System.out.println("**** IntraLA ****");
				spec.run("intrala");
			}
			if(explore_algorithm){
				System.out.println("**** Explore ****");
				if(intrala_algorithm){
					spec=null;
					spec=parser.translate();
				}
				spec.run("explore");
			}
		}
		//		buildProgramGraph("...test...");
	}

	//MarcoR
	public static String[][] combinations(String[] C,int k){
		int numColM=k;
		int numRowM=pow(C.length,k);
		String[][] combinations=new String[numRowM][numColM];
		
		int row=0;
		for(int col=0;col<numColM;col++){
			row=0;
			while(row<numRowM)
				for(int c=0;c<C.length;c++)
					for(int i=0;i<pow(C.length,col);i++){
						combinations[row][col]=C[c];
						row++;
					}
		}

		return combinations;
	}
	
	public static String[][][] sessionCombinations(String[][] C, int k){
		int numColM=k;
		int numRowM=pow(C.length,k);
		String[][][] sessionCombinations=new String[numRowM][numColM][C[0].length];
		
		int row=0;
		for(int col=0;col<numColM;col++){
			row=0;
			while(row<numRowM)
				for(int c=0;c<C.length;c++)
					for(int i=0;i<pow(C.length,col);i++){
						sessionCombinations[row][col]=C[c];
						row++;
					}
		}

		return sessionCombinations;
	}

	private static int pow(int value, int power){
		if(power == 0){
			return 1;
		}

		int result = value;
		for(int i = 0; i < power -1; i++){
			result = result * value; 
		}
		return result;
	}
	//MarcoR
	
}

