/* Generated By:JavaCC: Do not edit this line. Parser.java */
        package translator;
       import se.*;
       import org.jgrapht.alg.*;
       import org.jgrapht.*;
       import org.jgrapht.graph.*;
       import java.io.*;
import java.nio.file.*;
import java.nio.charset.Charset;


//       import java.io.BufferedWriter; 
//       import java.io.FileWriter; 
//       import java.io.IOException; 
//       import java.io.InputStreamReader; 
//       import java.io.BufferedReader; 
       import java.util.*;
        import java.lang.*;


  /**
   * The class contains the methods to perform:
   * <ul>
   * 	<li> The parsing of the ASLan++ specification.
   * 	<li> The translation procedure from ASLan++ specifications to SiL sequential programs.
   * 	<li> The algorithm that mixes the sequential programs of the various sessions into a unique program.
   * </ul>
   *
   * @author Fabio Pettenuzzo
   * @author Marco Palamà
   * @version 1.0, 03/14
   */
        public class Parser implements ParserConstants {

                /** Map that associates an entity name to the corresponding object representing it. */
                private HashMap<String, Entity > entities = new HashMap<String, Entity >();
                /** Reference to the root entity object. */
                private Environment environment = null;
                /** Reference to the current session entity object. */
                private Session session = null;
                /**
		 * Contains the sequential programs that are the result of the translation process.
		 * A sequential program is an ArrayList<Action>.
		 * Note: We generate a sequential program for each entity E such that si(E)!=i.
		 */
                private ArrayList<ArrayList<Action>> seqPrograms = new ArrayList<ArrayList<Action>>();
                /** Collection of goal actions objects. */
                private ArrayList<Action> goals = new ArrayList<Action>();

                /** Specifies the number of actions that composes a goal. */
                private HashMap<Action, Integer> goalsActionsNumber = new HashMap<Action, Integer>();

                private ArrayList<SwitchIntruderConstant> switchIntruderConstants = new ArrayList<SwitchIntruderConstant>();
                final int start = 0;
        final int end = 1;
        final String goalString = "goal";
        private Specification spec;
        private int globalCount = 1;
        private HashMap<String, ArrayList<Object[]>> locationsMap = new HashMap<String, ArrayList<Object[]>>();
        private String specName="";

        private int Y_Counter = 1;
        private int c_Counter = 1;
        private int X_counter = 1;

        /** Debug boolean value, if true debug messages are printed. */
        private final static boolean DEBUG = true;
//      private final static boolean DEBUG = false;

        private boolean dot = true;

                private FileWriter fw;
                private BufferedWriter bw;
                private String dirPath;

                public Parser(java.io.InputStream stream, boolean dot, String dirPath) {
                this(stream);
                this.dot = dot;
                this.dirPath = dirPath;
                }

                //MARCO
                private Message tmpSender;
                private Message tmpReceiver;
                private ArrayList<Message[]> who2who=new ArrayList<Message[]>();
                private FactVariable witness_fact=new FactVariable("witness");
                private boolean anotherRound=false;
                private FactAdd witness;
                private boolean itisarealgoal=true;
                private boolean thereisagoal=false;
                private ArrayList<Pair<String,Message>> inlineGoals=new ArrayList<Pair<String,Message>>();
                private int howManySessions=0;
                private int howManyAgentsPerSessions;
                public HashMap<String, Entity> getEntities(){
                        return this.entities;
                }
                public Environment getEnvironment(){
                        return this.environment;
                }
                public int getHowManySessions(){
                  return this.howManySessions;
                }
                public int getHowManyAgentsPerSessions(){
                        return this.howManyAgentsPerSessions;
                }
                //private void setHowManySessions(int howManySessions){
                //  this.howManySessions=howManySessions;	
                //}
                private void setHowManyAgentsPerSessions(int howManyAgentsPerSessions){
                  this.howManyAgentsPerSessions=howManyAgentsPerSessions;
                }
                //MARCO

                /**
		 * Translates an ASLan++ specification into a SiL sequential program.
		 * @return The Specification object representing the SiL sequential program
		 */
                public Specification translate() {
                        try {
                                MainSpec();
                        } catch (ParseException e) {
                                System.out.println("Error - Parsing error:");
                                e.printStackTrace();
                                System.exit(1);
                        }

                        if(DEBUG) {
                                Iterator<String> entitiesIterator = entities.keySet().iterator();
                                while (entitiesIterator.hasNext()) {
                                        String key = entitiesIterator.next().toString();

                                        System.out.println(key + " symbols: "); printHm(entities.get(key).getSymbols());
                                        System.out.println();

                                        System.out.println(key + " FormalParams: "); printStringAl(entities.get(key).getFormalParams());
                                        System.out.println();

                                        System.out.println(key + " actions: "); printActions(entities.get(key).getActions()); //printAl(entities.get(key).getActions());
                                        System.out.println();

                                }
                                System.out.println("Goals size: " + goals.size());
                        System.out.println("Goals : "); printGoals(goals);
                        System.out.println();

                        for(int i=0; i<seqPrograms.size(); i++) {
                                        System.out.println("Program "+(i+1));
                                        printActions(seqPrograms.get(i));
                                }
                        }

                        if(dot) {

                                if(DEBUG) System.out.println("dirPath: " + dirPath);
                                try {
                                        File file = new File(dirPath + specName+".dot");

                                        // if file doesn't exists, then create it
                                        if (!file.exists()) {
                                                file.createNewFile();
                                        }

                                        fw = new FileWriter(file.getAbsoluteFile());
                                        bw = new BufferedWriter(fw);
                                        bw.write("digraph "+specName+" {\u005cn");
                                } catch (IOException e) {
                                        System.out.println("Error - IOExeption:");
                                        e.printStackTrace();
                                        System.exit(2);
                                }
                        }

                        Specification spec = buildProgramGraph(specName);

                        //MARCO remove useless paths
                        spec.removeDeadEdges();
                        ArrayList<ActionEdge> edgesToRemove=spec.getEdgesRemoved();
                        //MARCO

                        if(dot) {
                                try {
                                        bw.write("}");
                                        bw.close();

                                //MARCO
                                try{
                                        File dot = new File(dirPath + specName+".dot");
                                        File dot_tmp = new File(dirPath + specName+"_tmp.dot");
                                        BufferedReader reader = Files.newBufferedReader(FileSystems.getDefault().getPath(dot.getAbsolutePath()),Charset.defaultCharset());
                                        BufferedWriter writer = Files.newBufferedWriter(FileSystems.getDefault().getPath(dot_tmp.getAbsolutePath()), Charset.defaultCharset());
                                        String line = null;
                                        boolean toRemove=false;

                                        while ((line=reader.readLine())!=null) {
                                                for(int i=0;i<edgesToRemove.size();i++){
                                                        if(line.startsWith(edgesToRemove.get(i).getSource().print().substring(1)+" -> "+edgesToRemove.get(i).getTarget().print().substring(1))){
                                                                toRemove=true;
                                                                break;
                                                        }
                                                }

                                                if(toRemove){
                                                        toRemove=false;
                                                }else{
                                                        writer.write(line+"\u005cn");
                                                }

                                        }

                                        reader.close();
                                        writer.flush();
                                        writer.close();
                                        dot.delete();
                                        dot_tmp.renameTo(new File(dirPath + specName+".dot"));


                                }catch (Exception e){
                                        e.printStackTrace();
                                }
                                //MARCO

                                        String s = null;
                                        String cmds[] = { "dot", "-Tpng", dirPath + specName + ".dot", "-o", dirPath + specName +".png" };

                                        Process p = Runtime.getRuntime().exec(cmds);
                                        p.waitFor();

                                        if(DEBUG) {
                                                BufferedReader stdInput = new BufferedReader(new InputStreamReader(p.getInputStream()));
                                                BufferedReader stdError = new BufferedReader(new InputStreamReader(p.getErrorStream()));

                                                // read the output from the command
                                                System.out.println("Here is the standard output of the command:\u005cn");
                                                while ((s = stdInput.readLine()) != null) {
                                                        System.out.println(s);
                                                }

                                                // read any errors from the attempted command
                                                System.out.println("Here is the standard error of the command (if any):\u005cn");
                                                while ((s = stdError.readLine()) != null) {
                                                        System.out.println(s);
                                                }
                                        }
                                } catch (InterruptedException e) {
                                        System.out.println("Error - InterruptedException:");
                                        e.printStackTrace();
                                } catch (IOException e) {
                                        System.out.println("Error - IOExeption:");
                                        e.printStackTrace();
                                        System.exit(2);
                                }
                        }
                        return spec;
                }

                /**
		  * Adds an edge to the file .dot, which will be read later by GraphViz to build the image of the graph.
		  * @param from starting vertex.
		  * @param to ending vertex.
		  * @param label label of the edge.
		  */
                public void addToDot(int from, int to, String label) {
                        try {
                                bw.write(from + " -> " + to + " [label=\u005c""+ label +"\u005c", color=black] ;\u005cn");
                        } catch (IOException e) {
                                System.out.println("Error - IOExeption: addToDot");
                                e.printStackTrace();
                                System.exit(2);
                        }
                }

                /**
		  * Prints a HashMap<String, Message> object.
		  * @param map HashMap<String, Message> object to be printed.
		  */
                public void printHm(HashMap<String, Message> map) {
                        Iterator<String> iterator = map.keySet().iterator();

                        while (iterator.hasNext()) {
                                String key = iterator.next().toString();
                                Message value = map.get(key);

                                System.out.println(key + " : " + value.print());
                        }
                }

                /**
		  * Prints an ArrayList<Action> object.
		  * @param al ArrayList<Action> object to be printed.
		  */
                public void printGoals(ArrayList<Action> al)
                {
                  for(int i=0; i<al.size(); i++) {
                        if (al.get(i) instanceof FactAdd) System.out.println("| TempGoal |");
                        else System.out.println("| "+al.get(i).print()+" : "+ goalsActionsNumber.get(al.get(i)) +" |");
                        }
                }

                /**
		  * Prints an ArrayList<Action> object.
		  * @param al ArrayList<Action> object to be printed.
		  */
                public void printAl(ArrayList<Action> al)
                {
                  for(int i=0; i<al.size(); i++) {
                        if (al.get(i) instanceof FactAdd) System.out.println("| TempGoal |");
                        else System.out.println("| "+al.get(i).print()+" |");
                        }
                }

                /**
		  * Prints an ArrayList<Action> object.
		  * @param al ArrayList<Action> object to be printed.
		  */
                public void printActions(ArrayList<Action> actions)
                {
                  for(int i=0; i<actions.size(); i++) {
                        if (actions.get(i) instanceof FactAdd) System.out.println("| TempGoal |");
                        else System.out.println("| "+actions.get(i).toString()+" |");
                        }
                }

                /**
		  * Prints an ArrayList<String> object.
		  * @param al ArrayList<String> object to be printed.
		  */
                public void printStringAl(ArrayList<String> al)
                {
                        System.out.print("| ");
                        for(int i=0; i<al.size(); i++) System.out.print(al.get(i)+" | ");
                }

                /**
		  * Given an action, it returns the corresponding name.
		  * @param a Action.
		  * @return the name of the Action. 
		  */
                public String getActionName(Action a){
                        return a.print().split(":")[0];
                }

                /**
		  * Builds the program graph specification by inserting the actions that were stored.
		  * @param name name that will be given to the specification.
		  * @return the specification builded. 
		  */
        public Specification buildProgramGraph(String name) {
                spec = new Specification(name);
                DirectedGraph<String, DefaultEdge> directedGraph = new DefaultDirectedGraph<String, DefaultEdge>(DefaultEdge.class);
            String lastInserted = null;
            HashMap<String, Message> symbols = environment.getSymbols();
            Iterator<String> iterSymbols = symbols.keySet().iterator();

                // for each constant in environment 
            while (iterSymbols.hasNext()){
                String symbol = iterSymbols.next();
                if(DEBUG) System.out.println("aumento IK per "+symbol);
                // i knows the agents involved in the scenario 
                Message message = environment.getSymbol(symbol);
                if(message  instanceof AgentConstant){
                    MessageSet mS_ik_initAg = new MessageSet();
                    mS_ik_initAg.add( (AgentConstant)message );
                    IKUpdate ik_initAg = new IKUpdate(mS_ik_initAg, "IK := IK + " + mS_ik_initAg.toString() );

                    spec.step(globalCount++, globalCount, ik_initAg);
                    if(dot){
                        addToDot(globalCount-1, globalCount, ik_initAg.print().split(":")[0]
                        + "\u005c\u005cn" + ik_initAg.toString());
                    }

                    String actName = globalCount+"."+getActionName(ik_initAg);
                    directedGraph.addVertex(actName);
                                if(lastInserted != null) directedGraph.addEdge(lastInserted, actName);
                    lastInserted = actName;
                    if(DEBUG) System.out.println("locCount: "+globalCount);

                                // for each agent i knows his own public and private keys 
                    MessageSet mS_ik_initPk = new MessageSet();
                    mS_ik_initPk.add( ((AgentConstant)message).getPublicKey() );
                    IKUpdate ik_initPk = new IKUpdate(mS_ik_initPk, "IK := IK + " + mS_ik_initPk.toString() );

                    spec.step(globalCount++, globalCount, ik_initPk);
                    if(dot){
                        addToDot(globalCount-1, globalCount, ik_initPk.print().split(":")[0]
                        + "\u005c\u005cn" + ik_initPk.toString());
                    }

                    actName = globalCount+"."+getActionName(ik_initPk);
                    directedGraph.addVertex(actName);
                    directedGraph.addEdge(lastInserted, actName);
                    lastInserted = actName;
                    if(DEBUG) System.out.println("globalCount: "+globalCount);
                    // check if the symbol is the intruder
                    if( symbol.equals("i") ){
                        MessageSet mS_ik_initPrv = new MessageSet();
                        mS_ik_initPrv.add( ((AgentConstant)message).getPrivateKey() );
                        IKUpdate ik_initPrv = new IKUpdate(mS_ik_initPrv, "IK := IK + " + mS_ik_initPrv.toString() );

                        spec.step(globalCount++, globalCount, ik_initPrv);
                        if(dot){
                                addToDot(globalCount-1, globalCount, ik_initPrv.print().split(":")[0]
                                + "\u005c\u005cn" + ik_initPrv.toString() );
                        }

                        actName = globalCount+"."+getActionName(ik_initPk);
                        directedGraph.addVertex(actName);
                        directedGraph.addEdge(lastInserted, actName);
                        lastInserted = actName;
                        if(DEBUG) System.out.println("globalCount: "+globalCount+" initPrv");
                            }
                    }
                        else if(DEBUG) System.out.println("Il simbolo " + message + " in Environment non e' instanceof AgentConstant");
                }
                if(DEBUG) System.out.println("ending initialization...");

                // the initial	location of a block	is either the initial	location of the graph or an intruder location
                // the exit		location of a block is either the last 		location of the graph or the location before an intruder location 
            ArrayList<ArrayList<int[]>> boundaryOfSessionsBlock = new ArrayList<ArrayList<int[]>>();
                for(int i=0; i< seqPrograms.size(); i++){
                        ArrayList<Action> actionsOfASession = seqPrograms.get(i);
                        ArrayList<int[]> tempBoundOfBlocks = new ArrayList<int[]>();
                        int[] boundary = new int[2];
                        boundary[start]=0;
                        for(int j=0; j<actionsOfASession.size(); j++){
                                Action aj = actionsOfASession.get(j);
                                // the block ends when the action is about the reciving of a message
                                // an intruder location corresponds to an action which is IKConditional but not a goal
                                if(aj instanceof IKConditional && !goals.contains(aj)){
                                                boundary[end] = j-1;
                                                tempBoundOfBlocks.add(boundary.clone());
                                                boundary[start] = j;
                                }
                        }
                        boundary[end]=actionsOfASession.size()-1;
                        // if the last boundary is valid
                        // it is added, otherwise it is discarded
                        if(boundary[start] < actionsOfASession.size()){
                                tempBoundOfBlocks.add(boundary.clone());
                        }
                        boundaryOfSessionsBlock.add(tempBoundOfBlocks);
                }

                ArrayList<Integer> goalQueue = new ArrayList<Integer>();

                // since the initialization blocks can be swapped without changing the semantics,
                // the initialization blocks of every session are added before the method which combines the sessions
                for(int sessionIterator=0; sessionIterator<boundaryOfSessionsBlock.size(); sessionIterator++){
                        int blockStart = boundaryOfSessionsBlock.get(sessionIterator).get(0)[start];
                        // check that every block added here starts with 0
                        assert blockStart == 0;
                        int blockEnd = boundaryOfSessionsBlock.get(sessionIterator).get(0)[end];

                        for(int blockIterator = blockStart; blockIterator <= blockEnd ; blockIterator++){
/*
				spec.step(globalCount++, globalCount, seqPrograms.get(sessionIterator).get(blockIterator));
				if(dot){
					addToDot(globalCount-1, globalCount, seqPrograms.get(sessionIterator).get(blockIterator).print().split(":")[0]
					+ "\\n" + (sessionIterator+1) + "." + (blockIterator+1) + " " + seqPrograms.get(sessionIterator).get(blockIterator).toString() );
				}
*/
                                Action toBeInserted = seqPrograms.get(sessionIterator).get(blockIterator);
                                if(goals.contains(toBeInserted)){ // se l'azione e' un goal devo inserire uno step particolare
                                        if(goalsActionsNumber.get(toBeInserted) == 1) {

                                                spec.step(globalCount, ++globalCount, toBeInserted, goalString);
                                                if(dot){
                                                        addToDot(globalCount-1, globalCount, goalString+":"+toBeInserted.print().split(":")[0]
                                                        + "\u005c\u005cn" + (sessionIterator+1) + "." + (blockIterator+1) + " " + toBeInserted.toString() );
                                                }

                                                for(int k=0; k<goalQueue.size(); k++) {
                                                        spec.step(goalQueue.get(k), globalCount, toBeInserted, goalString);
                                                        if(dot){
                                                                addToDot(goalQueue.get(k), globalCount, goalString+":"+toBeInserted.print().split(":")[0]
                                                                + "\u005c\u005cn" + (sessionIterator+1) + "." + (blockIterator+1) + " " + toBeInserted.toString() );
                                                        }
                                                }

                                        } else if (goalsActionsNumber.get(toBeInserted) == 2) {

                                                spec.step(globalCount, ++globalCount, toBeInserted);
                                                if(dot){
                                                        addToDot(globalCount-1, globalCount, toBeInserted.print().split(":")[0]
                                                        + "\u005c\u005cn" + (sessionIterator+1) + "." + (blockIterator+1) + " " + toBeInserted.toString() );
                                                }

                                                for(int k=0; k<goalQueue.size(); k++) {
                                                        spec.step(goalQueue.get(k), globalCount, toBeInserted);
                                                        if(dot){
                                                                addToDot(goalQueue.get(k), globalCount, toBeInserted.print().split(":")[0]
                                                                + "\u005c\u005cn" + (sessionIterator+1) + "." + (blockIterator+1) + " " + toBeInserted.toString() );
                                                        }
                                                }

                                                blockIterator++;
                                                toBeInserted = seqPrograms.get(sessionIterator).get(blockIterator);

                                                spec.step(globalCount, ++globalCount, toBeInserted, goalString);
                                                if(dot){
                                                        addToDot(globalCount-1, globalCount, goalString+":"+toBeInserted.print().split(":")[0]
                                                        + "\u005c\u005cn" + (sessionIterator+1) + "." + (blockIterator+1) + " " + toBeInserted.toString() );
                                                }

                                        }

                                        goalQueue.add(globalCount);

                                        if(DEBUG) System.out.println("goal spec.step("+(globalCount-1)+","+globalCount+","+toBeInserted.print()+")");
                                } else {

                                        spec.step(globalCount, ++globalCount, toBeInserted);
                                        if(dot){
                                                addToDot(globalCount-1, globalCount, toBeInserted.print().split(":")[0]
                                                + "\u005c\u005cn" + (sessionIterator+1) + "." + (blockIterator+1) + " " + toBeInserted.toString());
                                        }

                                        for(int k=0; k<goalQueue.size(); k++) {
                                                spec.step(goalQueue.get(k), globalCount, toBeInserted);
                                                if(dot){
                                                        addToDot(goalQueue.get(k), globalCount, toBeInserted.print().split(":")[0]
                                                        + "\u005c\u005cn" + (sessionIterator+1) + "." + (blockIterator+1) + " " + toBeInserted.toString() );
                                                }
                                        }

                                        goalQueue.clear();
                                }


                                if(DEBUG) System.out.println(""+blockIterator+"INIT spec.step("+(globalCount-1)+","+globalCount+","+seqPrograms.get(sessionIterator).get(blockIterator).print()+")");
                        }
                }

                int indexes[] = new int[boundaryOfSessionsBlock.size()];
                // having already inserted the first blocks of every session, the indexes starts from 1
                for(int k=0; k<indexes.length; k++) { indexes[k] = 1; switchIntruderConstants.add(new SwitchIntruderConstant(""+(k+1) ) ); }

                assert indexes.length == boundaryOfSessionsBlock.size();

//		locationsMap.put( Arrays.toString(indexes) , globalCount); 
//		if(DEBUG) System.out.println("sIndexes "+Arrays.toString(indexes)+" messo nella locaionsMap con "+globalCount);

                combineSessions(boundaryOfSessionsBlock, indexes, globalCount, goalQueue);

                if(DEBUG) System.out.println();
                return spec;
        }

        /**
	  * Builds the graph by combining the blocks related to all the sessions in the scenario.
	  * The idea is that all instructions contained in seqPrograms are allowed to be executed, in the proper order within the corresponding block.
	  * The blocks are combined in all the possible ways, provided that the order of the blocks belonging to the same agent is preserved.
	  * @param boundaryOfSessionsBlock contains, for every session and every block, an array of integer representing the boundaries.
	  * @param indexes keeps track, for every session, of the inserted blocks. Contains the number of the block to be inserted. 
	  * @param localCount keeps track of the last splitting point
	  * @param xInputSwitchIntruderVariable represents the variable that will be set by the intruder as an input on the receiving of a message. The corresponding step is added at the beginning of every block.  
	  */
        public void combineSessions(ArrayList<ArrayList<int[]>> boundaryOfSessionsBlock, int[] indexes, int localCount, ArrayList<Integer> goalQueue){

                int tempIndexes[] = indexes.clone();
                String sTempIndexes = Arrays.toString(tempIndexes);

                if(locationsMap.containsKey(sTempIndexes)){

                        ArrayList<Object[]> steps = locationsMap.get(sTempIndexes);
                        for( int i = 0; i<steps.size(); i++ ) {
                                Object[] step = steps.get(i);

                                spec.step(localCount, (Integer)step[0], (EqOPConditional)step[1]);
                                if(dot){
                                        addToDot(localCount, (Integer)step[0], ((EqOPConditional)step[1]).print().split(":")[0]
                                        + "\u005c\u005cn" + ((EqOPConditional)step[1]).toString());
                                }

                                for(int k=0; k<goalQueue.size(); k++) {
                                        spec.step(goalQueue.get(k), (Integer)step[0], (EqOPConditional)step[1]);
                                        if(dot){
                                                addToDot(goalQueue.get(k), (Integer)step[0], ((EqOPConditional)step[1]).print().split(":")[0]
                                                + "\u005c\u005cn" + ((EqOPConditional)step[1]).toString());
                                        }
                                }

                        }

                        goalQueue.clear();

                } else {

                        ArrayList<Object[]> steps = new ArrayList<Object[]>();
                        locationsMap.put(sTempIndexes, steps);

                        if(DEBUG) System.out.println("AGGIUNGO: "+sTempIndexes);

                        InputSwitchIntruderVariable xInputSwitchIntruderVariable = new InputSwitchIntruderVariable("X_"+X_counter++);

                        int sameCount = localCount;

                        ArrayList<Integer> sameQueue = new ArrayList<Integer>();
                        for(int j = 0; j < goalQueue.size(); j++) {
                                sameQueue.add(goalQueue.get(j));
                        }

                        for(int sessionIterator = 0; sessionIterator < boundaryOfSessionsBlock.size(); sessionIterator++){
                                // if there are blocks remaining in this session
                                if(indexes[sessionIterator] < boundaryOfSessionsBlock.get(sessionIterator).size()){

                                        EqOPConditional xEqSession = new EqOPConditional(xInputSwitchIntruderVariable, switchIntruderConstants.get(sessionIterator), xInputSwitchIntruderVariable + " = " + switchIntruderConstants.get(sessionIterator));

                                        // the first action added always go from the previous split (tracked by localCount) to the globalCount, which has to be incremented
                                        spec.step(sameCount, ++globalCount, xEqSession);
                                        Object[] step = new Object[]{ globalCount, xEqSession };

                                        if(DEBUG) System.out.println("AGGIUNGO step: "+globalCount +" , "+ xEqSession);

                                        if(dot){
                                                addToDot(sameCount, globalCount, xEqSession.print().split(":")[0]
                                                + "\u005c\u005cn" + xEqSession.toString());
                                        }

                                        for(int k=0; k<sameQueue.size(); k++) {
                                                spec.step(sameQueue.get(k), globalCount, xEqSession);
                                                if(dot){
                                                        addToDot(sameQueue.get(k), globalCount, xEqSession.print().split(":")[0]
                                                        + "\u005c\u005cn" + xEqSession.toString());
                                                }
                                        }

                                        localCount = globalCount;
                                        goalQueue.clear();

                                        locationsMap.get(sTempIndexes).add(step);


                                        //if(DEBUG) System.out.println("XSWITCH spec.step("+localCount+","+globalCount+","+xEqSession.print()+")");

                                        int blockStart = boundaryOfSessionsBlock.get(sessionIterator).get(indexes[sessionIterator])[start];
                                        int blockEnd = boundaryOfSessionsBlock.get(sessionIterator).get(indexes[sessionIterator])[end];

                                        // the actions within the block are added
                                        for(int blockIterator = blockStart; blockIterator <= blockEnd ; blockIterator++){
                                                Action toBeInserted = seqPrograms.get(sessionIterator).get(blockIterator);
                                                if(goals.contains(toBeInserted)){ // se l'azione e' un goal devo inserire uno step particolare
                                                        if(goalsActionsNumber.get(toBeInserted) == 1) {

                                                                spec.step(localCount, ++globalCount, toBeInserted, goalString);
                                                                if(dot){
                                                                        addToDot(localCount, globalCount, goalString+":"+toBeInserted.print().split(":")[0]
                                                                        + "\u005c\u005cn" + (sessionIterator+1) + "." + (blockIterator+1) + " " + toBeInserted.toString() );
                                                                }

                                                                for(int k=0; k<goalQueue.size(); k++) {
                                                                        spec.step(goalQueue.get(k), globalCount, toBeInserted, goalString);
                                                                        if(dot){
                                                                                addToDot(goalQueue.get(k), globalCount, goalString+":"+toBeInserted.print().split(":")[0]
                                                                                + "\u005c\u005cn" + (sessionIterator+1) + "." + (blockIterator+1) + " " + toBeInserted.toString() );
                                                                        }
                                                                }

                                                        } else if (goalsActionsNumber.get(toBeInserted) == 2) {

                                                                spec.step(localCount, ++globalCount, toBeInserted);
                                                                if(dot){
                                                                        addToDot(localCount, globalCount, toBeInserted.print().split(":")[0]
                                                                        + "\u005c\u005cn" + (sessionIterator+1) + "." + (blockIterator+1) + " " + toBeInserted.toString() );
                                                                }

                                                                for(int k=0; k<goalQueue.size(); k++) {
                                                                        spec.step(goalQueue.get(k), globalCount, toBeInserted);
                                                                        if(dot){
                                                                                addToDot(goalQueue.get(k), globalCount, toBeInserted.print().split(":")[0]
                                                                                + "\u005c\u005cn" + (sessionIterator+1) + "." + (blockIterator+1) + " " + toBeInserted.toString() );
                                                                        }
                                                                }

                                                                blockIterator++;
                                                                toBeInserted = seqPrograms.get(sessionIterator).get(blockIterator);

                                                                spec.step(globalCount, ++globalCount, toBeInserted, goalString);
                                                                if(dot){
                                                                        addToDot(globalCount-1, globalCount, goalString+":"+toBeInserted.print().split(":")[0]
                                                                        + "\u005c\u005cn" + (sessionIterator+1) + "." + (blockIterator+1) + " " + toBeInserted.toString() );
                                                                }

                                                        }

                                                        goalQueue.add(globalCount);

                                                        if(DEBUG) System.out.println("goal spec.step("+(globalCount-1)+","+globalCount+","+toBeInserted.print()+")");
                                                } else {

                                                        spec.step(localCount, ++globalCount, toBeInserted);
                                                        if(dot){
                                                                addToDot(localCount, globalCount, toBeInserted.print().split(":")[0]
                                                                + "\u005c\u005cn" + (sessionIterator+1) + "." + (blockIterator+1) + " " + toBeInserted.toString());
                                                        }

                                                        for(int k=0; k<goalQueue.size(); k++) {
                                                                spec.step(goalQueue.get(k), globalCount, toBeInserted);
                                                                if(dot){
                                                                        addToDot(goalQueue.get(k), globalCount, toBeInserted.print().split(":")[0]
                                                                        + "\u005c\u005cn" + (sessionIterator+1) + "." + (blockIterator+1) + " " + toBeInserted.toString() );
                                                                }
                                                        }

                                                        localCount = globalCount;

                                                        goalQueue.clear();

                                                        //if(DEBUG) System.out.println("spec.step("+(globalCount-1)+","+globalCount+","+seqPrograms.get(sessionIterator).get(blockIterator).print()+")");
                                                }
                                        }

                                        // the index has to be incremented for the next recursion call 
                                        tempIndexes[sessionIterator]++;
                                        combineSessions(boundaryOfSessionsBlock, tempIndexes, localCount, goalQueue);
                                        tempIndexes[sessionIterator]--;
                                }
                        }
                }
        }

/*
 * Productions block.
 */

/**
 * Recognizes the specification.
 */
  final public void MainSpec() throws ParseException {
  String name;
    jj_consume_token(7);
    name = Name();
          specName = name;
    jj_consume_token(8);
    jj_consume_token(9);
    EntityDecl(null);
    jj_consume_token(0);
  }

/**
 * Recognizes an entity declaration.
 * @param	parent	Parent entity of the current entity.
 */
  final public void EntityDecl(Entity parent) throws ParseException {
        Token token;
        Entity entity = new Entity(null, null);
    jj_consume_token(10);
    token = jj_consume_token(UpperName);
        if(parent == null) { // Root entity
                environment = new Environment(parent, token.toString());
                entity = environment;
        }
        else if (parent instanceof Environment) {
                session = new Session(parent, token.toString());
                entity = session;
        }
        else if (parent instanceof Session) entity = new Entity(parent, token.toString());
        else { System.out.println("Error - Entity "+ token.toString() +": Wrong nesting."); System.exit(2); }
        entities.put(token.toString(), entity);
        /* Reset counters */
        Y_Counter = 1;
//    c_Counter = 1;

    switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
    case 14:
      Params(entity);
      break;
    default:
      jj_la1[0] = jj_gen;
      ;
    }
    jj_consume_token(11);
    label_1:
    while (true) {
      switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
      case 17:
        ;
        break;
      default:
        jj_la1[1] = jj_gen;
        break label_1;
      }
      Decls(entity);
    }
    label_2:
    while (true) {
      switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
      case 10:
        ;
        break;
      default:
        jj_la1[2] = jj_gen;
        break label_2;
      }
      EntityDecl(entity);
    }
    switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
    case 26:
      Body(entity);
      break;
    default:
      jj_la1[3] = jj_gen;
      ;
    }
    switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
    case 32:
      GoalDecls(entity);
      break;
    default:
      jj_la1[4] = jj_gen;
      ;
    }
    jj_consume_token(12);
  }

/**
 * Recognizes a name.
 * @return	The name.
 */
  final public String Name() throws ParseException {
  Token token;
    switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
    case LowerName:
      token = jj_consume_token(LowerName);
      break;
    case UpperName:
      token = jj_consume_token(UpperName);
      break;
    default:
      jj_la1[5] = jj_gen;
      jj_consume_token(-1);
      throw new ParseException();
    }
   {if (true) return token.toString();}
    throw new Error("Missing return statement in function");
  }

/**
 * Recognizes a variable name.
 * @return	The variable name.
 */
  final public String Var() throws ParseException {
  Token token;
    token = jj_consume_token(UpperName);
          {if (true) return token.toString();}
    throw new Error("Missing return statement in function");
  }

/**
 * Recognizes a list of variables.
 * @return	ArrayList containing the list of variables.
 */
  final public ArrayList<String > Vars() throws ParseException {
  ArrayList<String > vars = new ArrayList<String >(); String var;
    var = Var();
                vars.add(var);
    label_3:
    while (true) {
      switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
      case 13:
        ;
        break;
      default:
        jj_la1[6] = jj_gen;
        break label_3;
      }
      jj_consume_token(13);
      var = Var();
                     vars.add(var);
    }
          {if (true) return vars;}
    throw new Error("Missing return statement in function");
  }

/**
 * Recognizes a list of parameters and adds them to the current entity data structures.
 * @param	entity	The current entity.
 */
  final public void Params(Entity entity) throws ParseException {
        String type;
        ArrayList<String > vars;
    jj_consume_token(14);
    vars = Vars();
    jj_consume_token(15);
    type = Type();
        if(type.equals("agent")) {
                        for(int i=0; i<vars.size(); i++) {
                                entity.addSymbol(vars.get(i), new AgentVariable( entity.getName() + "." + vars.get(i) ));
                                entity.addFormalParam(vars.get(i));
                        }
        }
                else { System.out.println("Error - Type not supported: "+ type); System.exit(2); }
    label_4:
    while (true) {
      switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
      case 13:
        ;
        break;
      default:
        jj_la1[7] = jj_gen;
        break label_4;
      }
      jj_consume_token(13);
      vars = Vars();
      jj_consume_token(15);
      type = Type();
        if(type.equals("agent")) {
                        for(int i=0; i<vars.size(); i++) {
                                entity.addSymbol(vars.get(i), new AgentVariable( entity.getName() + "." + vars.get(i) ));
                                entity.addFormalParam(vars.get(i));
                        }
        }
                else { System.out.println("Error - Type not supported: "+ type); System.exit(2); }
    }
    jj_consume_token(16);
  }

/*
void Imports():
{}
{
  "import" { constructNotSupported("import"); } < UpperName > ("," < UpperName >)*
}
*/

/**
 * Recognizes a declaration.
 * @param	entity	The current entity.
 */
  final public void Decls(Entity entity) throws ParseException {
    SymbolDecls(entity);
  }

/*
void TypeDecls():
{}
{
  "types" { constructNotSupported("types"); } TypeDecl() ("," TypeDecl())*
}

void TypeDecl():
{}
{
  < LowerName > ("<" < LowerName >)?	// note the optional supertype
}
*/

/**
 * Recognizes a type name.
 * @return	The type name.
 */
  final public String Type() throws ParseException {
  Token token;
  String type;
    switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
    case LowerName:
      token = jj_consume_token(LowerName);
          type = token.toString();
      break;
    case 14:
      jj_consume_token(14);
      type = Type();
      jj_consume_token(16);
      break;
    default:
      jj_la1[8] = jj_gen;
      jj_consume_token(-1);
      throw new ParseException();
    }
          {if (true) return type;}
    throw new Error("Missing return statement in function");
  }

/*
void Types():
{}
{
  Type() ("," Type())*
}
*/

/**
 * Recognizes a list of symbol declarations.
 * @param	entity	The current entity.
 */
  final public void SymbolDecls(Entity entity) throws ParseException {
    jj_consume_token(17);
    label_5:
    while (true) {
      SymbolDecl(entity);
      jj_consume_token(18);
      switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
      case UpperName:
      case LowerName:
        ;
        break;
      default:
        jj_la1[9] = jj_gen;
        break label_5;
      }
    }
  }

/**
 * Recognizes a symbol declaration and adds it to the <i>symbol</i> data structure of the current entity.
 * @param	entity	The current entity.
 */
  final public void SymbolDecl(Entity entity) throws ParseException {
        String type;
        ArrayList< String > vars = new ArrayList< String >();
        Token token;
    switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
    case UpperName:
      vars = Vars();
      jj_consume_token(15);
      type = Type();
                if(type.equals("agent")) { // Agent variable
                        for(int i=0; i<vars.size(); i++) entity.addSymbol(vars.get(i), new AgentVariable(entity.getName()+"."+vars.get(i)));
        } else if(type.equals("text")) { // Nonce variable
                        for(int i=0; i<vars.size(); i++) entity.addSymbol(vars.get(i), new NonceVariable(entity.getName()+"."+vars.get(i)));
                } else if(type.equals("symmetric_key")) { // Symmetric key variable
                        for(int i=0; i<vars.size(); i++) entity.addSymbol(vars.get(i), new SymKeyVariable(entity.getName()+"."+vars.get(i)));
                } else if(type.equals("public_key")) { // Public key variable, considered as an agent variable
                        for(int i=0; i<vars.size(); i++) entity.addSymbol(vars.get(i), new AgentVariable(entity.getName()+"."+vars.get(i))); // extracts the agent's name and adds the public key 
                                                                                                                                                                                                                // symbols HashMap, according to the convention:
                                                                                                                                                                                                                // KIdP = pk(IdP) 
                } else { // Type unknown
                        for(int i=0; i<vars.size(); i++) entity.addSymbol(vars.get(i), new GenericVariable(entity.getName()+"."+vars.get(i)));
                        //System.out.println("Type not supported: "+s);
                }
      break;
    case LowerName:
      /* ("nonpublic")? */ token = jj_consume_token(LowerName);
                                               vars.add(token.toString());
      label_6:
      while (true) {
        switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
        case 13:
          ;
          break;
        default:
          jj_la1[10] = jj_gen;
          break label_6;
        }
        jj_consume_token(13);
        token = jj_consume_token(LowerName);
                               vars.add(token.toString());
      }
      jj_consume_token(15);
      type = Type();
                if(type.equals("agent")) { // Agent constant
                        for(int i=0; i<vars.size(); i++) entity.addSymbol(vars.get(i), new AgentConstant(entity.getName()+"."+vars.get(i)));
        } else if(type.equals("text")) { // Nonce constant
                        for(int i=0; i<vars.size(); i++) entity.addSymbol(vars.get(i), new NonceConstant(entity.getName()+"."+vars.get(i)));
                } else if(type.equals("symmetric_key")) { // Symmetric key constant
                        for(int i=0; i<vars.size(); i++) entity.addSymbol(vars.get(i), new SymKeyConstant(entity.getName()+"."+vars.get(i)));
                //} else if(s.equals("public_key")) {
                        //for(int i=0; i<al.size(); i++) e.addSymbol(al.get(i), new PublicKey(al.get(i).substring(1))); // extracts the agent's name and adds the public key 
                                                                                                                                                                                                                // symbols HashMap, according to the convention:
                                                                                                                                                                                                                // KIdP = pk(IdP) 
                } else { // Type unknown
                        for(int i=0; i<vars.size(); i++) entity.addSymbol(vars.get(i), new GenericConstant(entity.getName()+"."+vars.get(i)));
                        //System.out.println("Type not supported: "+s);
                }
      break;
    default:
      jj_la1[11] = jj_gen;
      jj_consume_token(-1);
      throw new ParseException();
    }
  }

/*
void MacroDecls():
{}
{
  "macros" { constructNotSupported("macros"); } (MacroDecl() ";")+
}

void MacroDecl():
{}
{
  (Var() "->")? < LowerName > ("(" Vars() ")")? "=" Term()
}

void ClauseDecls():
{}
{
  "clauses" { constructNotSupported("clauses"); } (ClauseDecl() ";")+
}

void ClauseDecl():
{}
{
  < LowerName > ("(" Vars() ")")? ":"	// name and parameters
  ("forall" (Var())+ ".")? Term() (":-" Condition() ("&" Condition())*)?
}

void Condition():
{}
{
  Term() | Term() "=" Term() | Term() "!=" Term() | "!" Condition()
}

void EquationDecls():
{}
{
  "equations" { constructNotSupported("equations"); } (EquationDecl() ";")+	// only allowed in the root entity
}

void EquationDecl():
{}
{
  Term() "=" Term()
}
*/

/**
 * Recognizes a term and returns it. 
 * @param	entity	The current entity.
 * @return	The recognized term.
 */
  final public Message Term(Entity entity) throws ParseException {
  Message term_1, term_2, term_3;
 Token token;
 String name;
 String var;
    if (jj_2_1(2147483647)) {
      name = Name();
      jj_consume_token(19);
      term_1 = Term(entity);
      jj_consume_token(16);
        /* 
  	 * If goal labels are present a temporary goal action is added to the tempActions,
  	 * entityGoals and goalParameter data structures of the current entity.  
 	 */
        //MARCO

        this.thereisagoal=true;

        if(this.itisarealgoal)
        {
        this.inlineGoals.add(new Pair<String,Message>(name,term_1));
        Message[] w2w_entry=new Message[3];
    w2w_entry[0]=this.tmpSender;
    w2w_entry[1]=this.tmpReceiver;
    w2w_entry[2]=term_1;
    this.who2who.add(w2w_entry);
        //MARCO
        //System.out.println("DEBUG real goal!");
        Action action = new FactAdd(null, null); // Temporary action //InEqOPConditional(null, null);	
        entity.addGoalTempAction(action); // on top o normale?
        entity.addToEntityGoals(action);
        entity.addToGoalParameter(action, new Object[]{ entity.getTempSender(), entity.getTempReceiver(), name, term_1 } );
        //MARCO
        }
        //else
        //{
         //System.out.println("DEBUG not real");
        //}
        this.itisarealgoal=true;


        //MARCO

      // nuova action aggiunta atempaction... 
        //la aggiungo anche ai goal e all'hashmap passando un array con tempsender, tempreceiver, s ed e_1 
      
                                                                                              term_3 = TermNoLR(entity);
    if(term_3 == null) {if (true) return term_1;} else {if (true) return new ConcatenatedMessage(term_1, term_3, term_1.toString()+","+term_3.toString());}
    } else if (jj_2_2(2147483647)) {
      term_1 = FuncApp(entity);
      term_3 = TermNoLR(entity);
          if(term_3 == null) {if (true) return term_1;} else {if (true) return new ConcatenatedMessage(term_1, term_3, term_1.toString()+","+term_3.toString());}
    } else if (jj_2_3(2147483647)) {
      token = jj_consume_token(LowerName);
                          term_1 = entity.getSymbol(token.toString());
      term_3 = TermNoLR(entity);
          if(term_3 == null) {if (true) return term_1;} else {if (true) return new ConcatenatedMessage(term_1, term_3, term_1.toString()+","+term_3.toString());}
    } else if (jj_2_4(2147483647)) {
      var = Var();
      term_3 = TermNoLR(entity);
          if(term_3 == null) {if (true) return entity.getSymbol(var);} else {if (true) return new ConcatenatedMessage(entity.getSymbol(var), term_3, entity.getSymbol(var).toString()+","+term_3.toString());}
    } else if (jj_2_5(2147483647)) {
      jj_consume_token(20);
      var = Var();
          term_2 = entity.getSymbol(var);
        if(entity.containsSymbol("?"+var)) term_1 = entity.getSymbol("?"+var);
        else {
                if (term_2 instanceof AgentVariable) { term_1 = new InputAgentVariable( entity.getName() + "." + "Y_"+Y_Counter++ ); entity.addTempAction(new Assignment((AgentVariable)term_2, (InputAgentVariable)term_1, term_2.toString() + " := " + term_1.toString() )); }
                else if(term_2 instanceof NonceVariable) { term_1 = new InputNonceVariable( entity.getName() + "." + "Y_"+Y_Counter++ ); entity.addTempAction(new Assignment((NonceVariable)term_2, (InputNonceVariable)term_1, term_2.toString() + " := " + term_1.toString())); }
                else if(term_2 instanceof SymKeyVariable) { term_1 = new InputSymKeyVariable( entity.getName() + "." + "Y_"+Y_Counter++ ); entity.addTempAction(new Assignment((SymKeyVariable)term_2, (InputSymKeyVariable)term_1, term_2.toString() + " := " + term_1.toString())); }
                else if(term_2 instanceof GenericVariable) { term_1 = new InputGenericVariable( entity.getName() + "." + "Y_"+Y_Counter++ ); entity.addTempAction(new Assignment((GenericVariable)term_2, (InputGenericVariable)term_1, term_2.toString() + " := " + term_1.toString())); }
                else { term_1 = null; System.out.println("Error - Symbol not declared: " + var); System.exit(2);  /*term_1 = null; System.out.println("else da togliere, questa stampa non dovrebbe mai essere visualizzata");*/ } } // errore da gestire con un'eccezione
                entity.addSymbol("?"+var, term_1);
      term_3 = TermNoLR(entity);
          if(term_3 == null) {if (true) return term_1;} else {if (true) return new ConcatenatedMessage(term_1, term_3, term_1.toString()+","+term_3.toString());}
    } else if (jj_2_6(2147483647)) {
      jj_consume_token(20);
          {if (true) return null;}
    } else {
      switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
      case 21:
        jj_consume_token(21);
        term_1 = Term(entity);
        jj_consume_token(22);
        term_2 = Term(entity);
        term_3 = TermNoLR(entity);
        if(term_2 instanceof SymKey) term_1 = new SymEncryptedMessage(term_1, (SymKey)term_2, "{|"+term_1.toString()+"|}_"+term_2.toString());
        else { System.out.println("Error - " + term_2.print() + " is not insance of SymKey"); System.exit(2); }
        if(term_3 == null) {if (true) return term_1;} else {if (true) return new ConcatenatedMessage(term_1, term_3, term_1.toString()+","+term_3.toString());}
        break;
      default:
        jj_la1[12] = jj_gen;
        if (jj_2_7(2147483647)) {
          jj_consume_token(11);
          term_1 = Term(entity);
          jj_consume_token(23);
          term_2 = Term(entity);
          term_3 = TermNoLR(entity);
        if(term_2 instanceof PublicKey) term_1 = new EncryptedMessage(term_1, ((PublicKey)term_2).getAgent().getPrivateKey(), "{|"+term_1.toString()+"|}_inv("+term_2.toString()+")");
        else if(term_2 instanceof Agent) term_1 = new EncryptedMessage(term_1, ((Agent)term_2).getPrivateKey(), "{|"+term_1.toString()+"|}_inv("+term_2.toString()+")");
        else { System.out.println("Error - " + term_2.print() + " is not insance of PublicKey or Agent"); System.exit(2); }
        if(term_3 == null) {if (true) return term_1;} else {if (true) return new ConcatenatedMessage(term_1, term_3, term_1.toString()+","+term_3.toString());}
        } else if (jj_2_8(2147483647)) {
          jj_consume_token(11);
          term_1 = Term(entity);
          jj_consume_token(24);
          term_2 = Term(entity);
          term_3 = TermNoLR(entity);
        if(term_2 instanceof PublicKey) term_1 = new EncryptedMessage(term_1, (PublicKey)term_2, "{|"+term_1.toString()+"|}_"+term_2.toString());
        else if(term_2 instanceof Agent) term_1 = new EncryptedMessage(term_1, ((Agent)term_2).getPublicKey(), "{|"+term_1.toString()+"|}_"+term_2.toString());
        else { System.out.println("Error - " + term_2.print() + " is not insance of PublicKey or Agent"); System.exit(2); }
        if(term_3 == null) {if (true) return term_1;} else {if (true) return new ConcatenatedMessage(term_1, term_3, term_1.toString()+","+term_3.toString());}
        } else {
          switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
          case 14:
            jj_consume_token(14);
            term_1 = Term(entity);
            jj_consume_token(16);
            term_3 = TermNoLR(entity);
          if(term_3 == null) {if (true) return term_1;} else {if (true) return new ConcatenatedMessage(term_1, term_3, term_1.toString()+","+term_3.toString());}
            break;
          default:
            jj_la1[13] = jj_gen;
            jj_consume_token(-1);
            throw new ParseException();
          }
        }
      }
    }
    throw new Error("Missing return statement in function");
  }

/**
 * Recognizes a term after a "." and returns it. 
 * @param	entity	The current entity.
 * @return	The recognized term.
 */
  final public Message TermNoLR(Entity entity) throws ParseException {
  Message term_1, term_2;
    switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
    case 25:
      jj_consume_token(25);
      term_1 = Term(entity);
      /* +*/     term_2 = TermNoLR(entity);
   if(term_2 == null) {if (true) return term_1;} else {if (true) return new ConcatenatedMessage(term_1, term_2, term_1.toString()+","+term_2.toString());}
      break;
    default:
      jj_la1[14] = jj_gen;
    {if (true) return null;}
    }
    throw new Error("Missing return statement in function");
  }

/**
 * Recognizes a list of terms and returns it. 
 * @param	entity	The current entity.
 * @return	ArrayList containing the list of terms.
 */
  final public ArrayList<Message> Terms(Entity entity) throws ParseException {
        Message term;
        ArrayList<Message> terms = new ArrayList<Message>();
    term = Term(entity);
                              terms.add(term);
    label_7:
    while (true) {
      switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
      case 13:
        ;
        break;
      default:
        jj_la1[15] = jj_gen;
        break label_7;
      }
      jj_consume_token(13);
      term = Term(entity);
                                    terms.add(term);
    }
          {if (true) return terms;}
    throw new Error("Missing return statement in function");
  }

/**
 * Recognizes a function application and returns the term representing it; at the moment only the <i>pk</i> function is supported. 
 * @param	entity	The current entity.
 * @return	The term representing the function application.
 */
  final public Message FuncApp(Entity entity) throws ParseException {
        Token token;
        ArrayList<Message > terms;
    /*( Term() "->" )?*/ token = jj_consume_token(LowerName);
    jj_consume_token(14);
    terms = Terms(entity);
    jj_consume_token(16);
        if(token.toString().equals("pk") && terms.size() == 1 && terms.get(0) instanceof AgentVariable) {if (true) return ((AgentVariable)terms.get(0)).getPublicKey();}
        else { System.out.println("Error - Only \u005c"pk\u005c" function is supported and it must have one argument of type \u005c"agent\u005c""); System.exit(2); }
        // TODO bisognerebbe gestire meglio l'errore in caso di casting
        {if (true) return null;}
    throw new Error("Missing return statement in function");
  }

/**
 * Recognizes the body of the current entity.
 * @param	entity	The current entity.
 */
  final public void Body(Entity entity) throws ParseException {
    jj_consume_token(26);
    Stmt(entity);
  }

/*
void BreakDecl():
{}
{
	"breakpoints" "{" ( < LowerName > ("," < LowerName > )* )+ "}"
}
*/
/*
void Guard():
{}
{
  Transmission()			GuardNoLR() // sending transmission not allowed here
|FuncApp()					GuardNoLR()
| "!" Guard()				GuardNoLR()
| Term() "!=" Term()		GuardNoLR()
| Term() "=" Term()			GuardNoLR()
| "(" Guard() ")"			GuardNoLR()
}

void GuardNoLR():
{}
{
 	("&" | "|" | "=>") Guard()	GuardNoLR()
}


void GuardNoRcv():
{}
{
	Guard() // any Guard but no transmission allowed here
}
*/

/**
 * Recognizes a statement.
 * @param	entity	The current entity.
 */
  final public void Stmt(Entity entity) throws ParseException {
  Action action;
    if (jj_2_9(2147483647)) {
      action = FreshGen(entity);
        if(action!=null) {
                entity.addAction(action); // The action representing the statement is added to the actions ArrayList of the current entity.
                entity.getActions().addAll(entity.getGoalTempActions()); // Temporary goal actions of the current entity are added to its actions ArrayList.
                entity.getGoalTempActions().clear();
        }
    } else if (jj_2_10(2147483647)) {
      action = Assign(entity);
        if(action!=null) {
                entity.addAction(action); // The action representing the statement is added to the actions ArrayList of the current entity.
                entity.getActions().addAll(entity.getGoalTempActions()); // Temporary goal actions of the current entity are added to its actions ArrayList.
                entity.getGoalTempActions().clear();
        }
    } else {
      switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
      case 29:
        EntityGen(entity);
        break;
      case 30:
        SymbEntityGen(entity);
        break;
      default:
        jj_la1[17] = jj_gen;
        if (jj_2_11(2147483647)) {
          action = Transmission(entity);
        if(action!=null) {
                entity.addAction(action); // The action representing the statement is added to the actions ArrayList of the current entity.
                entity.getActions().addAll(entity.getTempActions()); // Temporary actions of the current entity are added to its actions ArrayList.
                entity.getActions().addAll(entity.getGoalTempActions()); // Temporary goal actions of the current entity are added to its actions ArrayList.
                entity.getTempActions().clear();
                entity.getGoalTempActions().clear();
                if(this.anotherRound){
                        //Stmt(entity);
                        entity.addAction(this.witness);
                        this.anotherRound=false;
                }
        }
        } else if (jj_2_12(2147483647)) {
          jj_consume_token(11);
          label_8:
          while (true) {
            switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
            case 11:
            case 14:
            case 20:
            case 21:
            case 29:
            case 30:
            case UpperName:
            case LowerName:
              ;
              break;
            default:
              jj_la1[16] = jj_gen;
              break label_8;
            }
            Stmt(entity);
          }
          jj_consume_token(12);
        } else {
          jj_consume_token(-1);
          throw new ParseException();
        }
      }
    }
  }

/**
 * Recognizes an assignment statement and returns an Action representing it.
 * @param	entity	The current entity.
 * @return	Action representing the assignment statement.
 */
  final public Action Assign(Entity entity) throws ParseException {
  String var, name;
  Message term;
    if (jj_2_13(2147483647)) {
      name = Name();
      jj_consume_token(19);
      var = Var();
      jj_consume_token(16);
                Action action = new FactAdd(null, null);//InEqOPConditional(null, null);	
                entity.addGoalTempAction(action);
                entity.addToEntityGoals(action);
                entity.addToGoalParameter(action, new Object[]{ entity.getTempSender(), entity.getTempReceiver(), name, entity.getSymbol(var) } );
    } else if (jj_2_14(2147483647)) {
      var = Var();
    } else {
      jj_consume_token(-1);
      throw new ParseException();
    }
    jj_consume_token(27);
    term = Term(entity);
    jj_consume_token(18);
                {if (true) return new Assignment((Variable)entity.getSymbol(var), term, entity.getSymbol(var)+" := "+term.toString());}
    throw new Error("Missing return statement in function");
  }

/**
 * Recognizes a fresh value generation statement and returns an Action representing it.
 * @param	entity	The current entity.
 * @return	Action representing the fresh value generation statement.
 */
  final public Action FreshGen(Entity entity) throws ParseException {
  String var, name;
    if (jj_2_15(2147483647)) {
      name = Name();
      jj_consume_token(19);
      var = Var();
      jj_consume_token(16);
                Action action = new FactAdd(null, null);//new InEqOPConditional(null, null);	
                entity.addGoalTempAction(action);
                entity.addToEntityGoals(action);
                entity.addToGoalParameter(action, new Object[]{ entity.getTempSender(), entity.getTempReceiver(), name, entity.getSymbol(var) } );
    } else if (jj_2_16(2147483647)) {
      var = Var();
    } else {
      jj_consume_token(-1);
      throw new ParseException();
    }
    jj_consume_token(27);
    jj_consume_token(28);
    jj_consume_token(18);
                if(entity.getSymbol(var) instanceof NonceVariable) {
                        {if (true) return new Assignment((NonceVariable)entity.getSymbol(var), new NonceConstant("c_"+c_Counter), entity.getSymbol(var)+" := c_"+c_Counter++);}
                } else if(entity.getSymbol(var) instanceof AgentVariable) {
                        {if (true) return new Assignment((AgentVariable)entity.getSymbol(var), new AgentConstant("c_"+c_Counter), entity.getSymbol(var)+" := c_"+c_Counter++);}
                } else if(entity.getSymbol(var) instanceof SymKeyVariable) {
                        {if (true) return new Assignment((SymKeyVariable)entity.getSymbol(var), new SymKeyConstant("c_"+c_Counter), entity.getSymbol(var)+" := c_"+c_Counter++);}
                } else if(entity.getSymbol(var) instanceof GenericVariable) {
                        {if (true) return new Assignment((GenericVariable)entity.getSymbol(var), new GenericConstant("c_"+c_Counter), entity.getSymbol(var)+" := c_"+c_Counter++);}
                } else {if (true) return null;}
    throw new Error("Missing return statement in function");
  }

/**
 * Recognizes new entity instantiation.
 * @param	entity	The current entity.
 */
  final public void EntityGen(Entity entity) throws ParseException {
        ArrayList<Object> newEntityObjects = new ArrayList<Object>();
          if(! (entity instanceof Session) ) { System.out.println("Error - Entity "+entity.getName()+" must be instance of Session"); System.exit(2); /*System.out.println("Error: l'entity "+e+" non e' instanceof di Session");*/}
    jj_consume_token(29);
    newEntityObjects = Entity(entity);
    jj_consume_token(18);
                if(newEntityObjects.size() != 2) {
                        System.out.println("Error - wrong size: " + newEntityObjects.size()); System.exit(2);
                }
                String newEntityName = (String)newEntityObjects.get(0);
                Entity newEntity = entities.get(newEntityName);
                ArrayList<Message> actualParams = (ArrayList<Message>)newEntityObjects.get(1);
                newEntity.setActualParams(actualParams);
                // li salviamo nell'entità per poterli poi recuperare nell'analisi dei goal
                // contiene le variabili della sessione che andranno mappate nell'instanziare delle entita' (es: A,B)
                // signature = entity.getSignature();
                /*for (int i=0; i< signature.size(); i++){
			m_1 = entity.getSymbol(signature.get(i));
			m_2 = al.get(i);
			if(signature.get(i).equals("Actor")) entity.setSessionActor(m_2);
			if(m_1 instanceof Variable && m_2 instanceof Variable)
				entity.addActionOnTop(new Assignment((Variable)m_1, (Variable)m_2));
			else System.out.println("Error: casting - m_1 "+m_1+" o m_2 "+m_2+" non erano Variable");
		}*/

  }

/**
 * Recognizes new symbolic session instantiation.
 * @param	entity	The current entity.
 */
  final public void SymbEntityGen(Entity entity) throws ParseException {
        /* 
	 * ArrayList containing:
	 * Symbolic session entity name at position 0.
	 * Symbolic session entity actual parameters at position 1.
	*/
        ArrayList<Object> anyEntityObjects = new ArrayList<Object>();
        // Signature of the current symbolic session entity.
        ArrayList<String> sessionFormalParams = new ArrayList<String>(); //MR quanti agenti per sessione

        ArrayList<String> iterationEntityFormalParams = new ArrayList<String>();
        ArrayList<Action> initSession = new ArrayList<Action>();
        //ArrayList<Action> sessionInstance = new ArrayList<Action>();
        Message message_1, message_2;
          if(! (entity instanceof Environment) ) { System.out.println("Error - Entity "+entity.getName()+" is not instance of Environment"); System.exit(2); }
    jj_consume_token(30);
    /*( ( Var() )+ ".")?*/ anyEntityObjects = Entity(entity);
    jj_consume_token(18);
                //MARCO
                ArrayList<Pair<Message, Message>> mex_newMex=new ArrayList<Pair<Message, Message>>();
                this.howManySessions++;
                //MARCO
                if(anyEntityObjects.size() != 2) {      System.out.println("Error - Wrong size."); System.exit(2); }

                String anyEntityName = (String)anyEntityObjects.get(0);
                Session sessionEntity=null;

                if(entities.get(anyEntityName) instanceof Session)      sessionEntity = (Session)entities.get(anyEntityName);
                else { System.out.println("Error - Entity "+ anyEntityName + " is not instance of Session."); System.exit(2); }
                ArrayList<Message> anyEntityParameters = (ArrayList<Message>)anyEntityObjects.get(1); //MR qui ci sono gli agenti 
                sessionFormalParams = sessionEntity.getFormalParams();
                /*for (int i=0; i< signature.size(); i++){
			m_1 = sessionEntity.getSymbol(signature.get(i));
			m_2 = al.get(i);
			if(al.get(i) == e.getSymbol("i")) sessionIntruderVariable = m_1;
			if(m_1 instanceof Variable && m_2 instanceof Constant)
				initSession.add(new Assignment((Variable)m_1, (Constant)m_2)); // assegnamento alle variabili di sessione che sono assegnate alle variabili iniziali
			else System.out.println("Error: casting - m_1 "+m_1+" o m_2 "+m_2+" non erano Variable");
			// nell'else dovremmo lanciare un'eccezione: se uno dei 2 è null non è stata definita precedentemente
			// altrimenti vuol dire che m1 o m2 è di tipo sbagliato per qualche altra ragione
			// si potrebbe anche modificare la getsymbol in modo che, se il parametro richiesto è una costante (1 carattere minuscolo) 
			// e non è presente la crea e la restituisce 
		}*/
                Iterator<String> entityIterator = entities.keySet().iterator();
                String Key;
                Entity iterationEntity;
                while(entityIterator.hasNext()){ // For each entity
                        Key = (String)entityIterator.next();
                        /* Entity that iterates on all the entities. */
                        iterationEntity = entities.get(Key);
                        //System.out.println("1key: "+key);
                        if(iterationEntity.getParent() instanceof Session) { // If the entity is in a session
                                /* if(sessionIntruderVariable != entity.getSessionActor()) {// controlla che Actor != i
					ArrayList<Action> tempAl = new ArrayList<Action>();
					tempAl.addAll(initSession);
					tempAl.addAll(entity.getActions());
					sessionsActions.add(tempAl);
					}*/
                                        /* Temporary collection of actions, will contain the initActions and the iterationEntity actions */
                                        ArrayList<Action> tempActions = new ArrayList<Action>();
                                        /* Collection of initialization actions that are at the beginning of the sequential SiL program */
                                        ArrayList<Action> initActions = new ArrayList<Action>();
                                        //System.out.println("2key: "+key);
                                        /* Signature of the iteration entity. */
                                        iterationEntityFormalParams = iterationEntity.getFormalParams();
                                        /*
					 * ArrayList containing:
					 * ArrayList that will contain the signature of the iteration entity at position 0.
					 * ArrayList that will contain the parameters of the symbolic entity instantiation at position 1.
					 * So sessionInstance.get(1).get(i) contains the object representing the role for sessionInstance.get(0).get(i)
					 * for entity iterationEntity in this symbolic session instantiation.
					 */
                                        ArrayList<ArrayList<Message>> sessionInstance = new ArrayList<ArrayList<Message>>();
                                        sessionInstance.add(new ArrayList<Message>());
                                        sessionInstance.add(new ArrayList<Message>());
                                        /* Boolean variable that indicates if si(iterationEntity) = i */
                                        boolean isActorIntruder=false;

                                        for(int i=0; i < iterationEntityFormalParams.size(); i++){ // Iterate on signature of iterationEntity
                                                for(int j=0; j < sessionFormalParams.size(); j++){ // Iterate of signature of the session
                                                        if(iterationEntity.getActualParams().get(i) == sessionEntity.getSymbol(sessionFormalParams.get(j))){ // If the symbol corresponds
                                                                        message_1 = iterationEntity.getSymbol(iterationEntityFormalParams.get(i));
                                                                        message_2 = anyEntityParameters.get(j);
                                                                        sessionInstance.get(0).add(message_1);
                                                                        sessionInstance.get(1).add(message_2);
                                                                        initActions.add(new Assignment((Variable)message_1, (Constant)message_2, message_1.toString() + " := " + message_2.toString())); // Add the new initialization assignment.
                                                                        /* Check if the Actor parameter of iterationEntity is "i" in this symbolic entity instantiation */
                                                                        if(iterationEntity.getFormalParams().get(i).equals("Actor") && message_2 == entity.getSymbol("i")){
                                                                                        //new Assignment(m_1, m_2);
                                                                                        isActorIntruder=true;
                                                                                        //System.out.println("assegnamento di "+entity.getSignature().get(i)+" = "+ m_2.print());
                                                                                }
                                                                }
                                                }
                                        }
                                        if(!isActorIntruder) { // If Actor!=i then generate a new SiL sequential program
                                                /* Check  goals */
                                                ArrayList<Action> entityGoals = iterationEntity.getEntityGoals();
                                                ArrayList<Action> iterationEntityActions = new ArrayList<Action>();
                                                iterationEntityActions.addAll(iterationEntity.getActions());
                                                Action goal; // Generic goal
                                                // Object[] parameters;
                                                // Object[] sessionGoalParameters;
                                                // System.out.println("goals.size(): "+goals.size());

                                                /*
						System.out.println("Entity: "+iterationEntity.getName() );
						System.out.println("Goals size: " + entityGoals.size());
	        			System.out.println("Goals : "); printAl(entityGoals);
	        			System.out.println();
						*/

                                                for(int k=0; k<entityGoals.size(); k++) {
                                                        goal = entityGoals.get(k);
                                                        //parameters = entity.getGoalParameters(goal);
                                                        //sessionGoalParameters = sessionEntity.getSessionGoalParameters( entity.getGoalLabel(goal) );
                                                        //sessionGoalParameters = sessionEntity.getSessionGoalParameters( (String)parameters[2] );
                                                        //if( ((String)sessionGoalParameters[0]).equals("auth") ) {
                                                        String sessionGoalType = sessionEntity.getSessionGoalType(iterationEntity.getGoalLabel(goal));
                                                        if( sessionGoalType.equals("auth") ){ // Authentication goal
                                                                Message sender=null; // questa parte di controllo potrebbe essere omessa
                                                                //if( parameters[0] instanceof Message) sender = (Message)parameters[0];
                                                                sender = iterationEntity.getGoalSender(goal); // Sender of authentication message.
                                                                //else System.out.println("Sender non era di tipo Message");
                                                                /* Index of object representing the sender in this symbolic session */
                                                                int sessionIndex = sessionInstance.get(0).indexOf(sender);
                                                                /* Prove action */
                                                                //Action prove = new InEqOPConditional(sender, sessionInstance.get(1).get(sessionIndex), sender.toString() + " != " + sessionInstance.get(1).get(sessionIndex).toString() );
                                                                //MARCO (old stuff above)
                                                                //Action prove = new InEqOPConditional(sender, environment.getSymbol("i"), sender.toString() + " != "+environment.getSymbol("i").toString());
                                                                //for(int i=0;i<this.inlineGoals.size();i++)
                                                                //{
                                                                //	System.out.println("DEBUG "+this.inlineGoals.get(i).getElement0()+" , "+this.inlineGoals.get(i).getElement1().toString()); 
                                                                //}

                                                                Message[] args=new Message[3];
                                                                args[2]=this.inlineGoals.get(0).getElement1();
                                                                //for(int db=0;db<this.inlineGoals.size();db++)
                                                                //{
                                                                //  System.out.println("DEBUG inlineGoal["+db+"] "+this.inlineGoals.get(db).getElement1());//+" "+this.inlineGoals.get(db).getElement1());
                                                                //}
                                                                //for(int db=0;db<this.who2who.size();db++){
                                                                //	System.out.println("DEBUG who2who ["+db+"] "+this.who2who.get(db)[0]+" "+this.who2who.get(db)[1]+" and "+this.who2who.get(db)[2]+" != "+args[2]);
                                                                //}
                                                                //searches who sends the message to who - to crate the correct triple
                                                                for(int i=0;i<this.who2who.size();i++){
                                                                        if(this.who2who.get(i)[2].equals(args[2])){
                                                                          args[0]=this.who2who.get(i)[0];
                                                                          args[1]=this.who2who.get(i)[1];
                                                                          break;
                                                                        }
                                                                }

                                                                //Tuple tp=new Tuple(args,this.inlineGoals.get(0).getElement1().toString())
                                                                Tuple tp=new Tuple(args,"<"+args[0].toString()+", "+args[1].toString()+", "+args[2].toString()+" >");

                                                                Action prove = new ConjunctiveConditional(
                                                                  //new NotMInIKConditional(this.inlineGoals.get(0).getElement1()),
                                                                  new FactConditional(witness_fact,tp,FalseConstant.getInstance(),witness_fact.getName()+"("+tp.toString()+")=false"),
                                                                  new InEqOPConditional(sender, environment.getSymbol("i"), sender.toString() + " != "+environment.getSymbol("i").toString())
                                                                  );
                                                                //MARCO
                                                                iterationEntityActions.set(iterationEntityActions.indexOf(goal), prove); // Substitute the temporary goal action with the prove action.

                                                                goals.add(prove); // Add to goals collection.
                                                                goalsActionsNumber.put(prove, 1);

                                                        //}else if( ((String)sessionGoalParameters[0]).equals("secr") ){
                                                        } else if( sessionGoalType.equals("secr") ){ // Secrecy goal	

                                                                ArrayList<Message> entityActualParams = iterationEntity.getActualParams(); // iterationEntity actual parameters
                                                                ArrayList<Message> confSet = sessionEntity.getSessionGoalParameters(iterationEntity.getGoalLabel(goal)); // Confidentiality set
                                                                Message confMessage = iterationEntity.getGoalArgument(goal); // Confidential message
                                                                /* Prove conditional: IK |- confMessage */
                                                                Conditional prove = new IKConditional(confMessage, "IK |- " + confMessage.toString());

                                                                /* for(int i=0; i < entityMapping.size(); i++) {
									
									if(	confSet.contains(entityMapping.get(i)) && 
										sessionInstance.get(1).get(i) == environment.getSymbol("i") ){
											//prove = new ConjunctiveConditional( prove, 
											//		new InEqOPConditional(sessionInstance.get(0).get(i), environment.getSymbol("i")) );
											// si aspetta che prove sia di tipo stdconditional, occorre modificare i tipi
										}
								}
								//System.out.println("index:" + entity.getActions().indexOf(goal));	
								entity.getActions().set(entity.getActions().indexOf(goal), prove);*/
                                                                iterationEntityActions.add(prove); // Add goal action at the end of entity action

                                                                goals.add(prove); // Add to goals collection.
                                                                goalsActionsNumber.put(prove, 1);

                                                                StdConditional condProve = null;
                                                                boolean first = true; // First InEqOPConditional
                                                                for(int i=0; i < entityActualParams.size(); i++) { // For all the actual parameters.

                                                                        if(     confSet.contains(entityActualParams.get(i)) &&
                                                                                sessionInstance.get(1).get(i) == environment.getSymbol("i") ){
                                                                                        if(first) { // If first conditional
                                                                                                condProve = new InEqOPConditional(sessionInstance.get(0).get(i), environment.getSymbol("i"), sessionInstance.get(0).get(i).toString() + " != i" );
                                                                                                first = false;
                                                                                        }
                                                                                        else { // If not first conditional create a conjunction with the previous one.
                                                                                                InEqOPConditional inEq = new InEqOPConditional(sessionInstance.get(0).get(i), environment.getSymbol("i"), sessionInstance.get(0).get(i).toString() + " != i");
                                                                                                condProve = new ConjunctiveConditional( condProve, inEq, condProve.toString() +" and "+ inEq.toString() );
                                                                                        }


                                                                                        //prove = new ConjunctiveConditional( prove, 
                                                                                        //		new InEqOPConditional(sessionInstance.get(0).get(i), environment.getSymbol("i")) );
                                                                                        // si aspetta che prove sia di tipo stdconditional, occorre modificare i tipi
                                                                                }
                                                                }
                                                                iterationEntityActions.remove(goal); // Remove the temporary goal
                                                                if(condProve != null) { // Add the Conjunctive conditional if not empty.
                                                                        iterationEntityActions.add(condProve);
                                                                        goals.add(condProve);
                                                                        goalsActionsNumber.put(prove, 2);
                                                                }
                                                        } else { System.out.println("Error - Goal type not supported: "+ sessionGoalType ); System.exit(2); /*System.out.println("Errore: Goal " + sessionGoalType + " non supportato, non dovrebbe mai capitare");*/}
                                                }
                                                /* Create a SiL sequential program from initialization actions followed by entity actions. */
                                                tempActions.addAll(initActions);
                                                tempActions.addAll(iterationEntityActions);

                                                //MARCO sostituisco tutte le variabili (in tempAction)
                                               // ArrayList<Pair<Message, Message>> mex_newMex=new ArrayList<Pair<Message, Message>>();
                                                Iterator<String> iter_symb=iterationEntity.getSymbols().keySet().iterator();

                                                while(iter_symb.hasNext()){
                                                        String currentSymb=iter_symb.next();
//                                                	System.out.println(iterationEntity.getName()+"."+this.howManySessions+"_"+currentSymb.toString());
                                                        Message m=iterationEntity.getSymbols().get(currentSymb);
                                               // 	System.out.println("message to string "+m.toString());
                                                        mex_newMex.add(new Pair<Message, Message>(m,m.getAnotherOfTheSameType("S"+this.howManySessions+"_"+m.toString())));
                                                }

                                               // for(int i=0;i<mex_newMex.size();i++)
                                               // 	System.out.print("("+mex_newMex.get(i).getElement0().toString()+", "+mex_newMex.get(i).getElement1().toString()+")");
                                               // System.out.println("\n");

                                                                                                //for(int i=0;i<sessionInstance.size();i++)
                                                                                                //  for(int j=0;j<sessionInstance.size();j++){
                                                                                                 // 	System.out.println("DEBUG sess inst ["+i+", "+j+"] "+sessionInstance.get(i).get(j));
                                                                                                 // }
                                                //System.out.println("DEBUG sess inst (1): "+sessionInstance.get(0).get(0)+" = "+sessionInstance.get(1).get(0));
                                                //System.out.println("DEBUG sess inst (1): "+sessionInstance.get(0).get(1)+" = "+sessionInstance.get(1).get(1));


                                                                                                ArrayList<Action> tmp_tempActions = new ArrayList<Action>();
                                                for(int i=0;i<tempActions.size();i++){
                                                    //System.out.println("DEBUG action pre "+tempActions.get(i).toString());

                                                    //System.out.println();
                                                    //for(int a=0;a<mex_newMex.size();a++)
                                                        //	System.out.print("("+mex_newMex.get(a).getElement0().toString()+", "+mex_newMex.get(a).getElement1().toString()+")");
                                                        //System.out.println();

                                                        Action tmp_act=tempActions.get(i).substitute(mex_newMex);
                                                        tmp_tempActions.add(tmp_act);
                                                        //System.out.println("DEBUG action post "+tmp_act.toString());

                                                        if(goals.contains(tempActions.get(i))){
                                                                this.goals.add(tmp_act);
                                                                this.goalsActionsNumber.put(tmp_act,1);
                                                        }
                                                }


                                                                                                seqPrograms.add(tmp_tempActions); // Add the sequential program to the sequential program collection
                                                                                                //MARCO

                                        }

                        }
                }
        //MARCO
        //this.setHowManySessions(sessionFormalParams.size());
        this.setHowManyAgentsPerSessions(anyEntityParameters.size());
        //MARCO

  }

/**
 * Recognizes a transmission statement and returns an Action representing it.
 * @param	entity	The current entity.
 * @return	Action representing the transmission statement.
 */
  final public Action Transmission(Entity entity) throws ParseException {
        Message message_1, message_2, message_3;
    // (Term(e) "->")? ( "send" | "receive")  "(" Terms(e) ")" ("over" Term(e))? ";"	// where "Actor ->" is optional
    /*|*/ message_1 = Term(entity);
    jj_consume_token(31);
    message_2 = Term(entity);
                                entity.setTempSender(message_1); entity.setTempReceiver(message_2);
                                if(message_1 == entity.getSymbol("Actor") && message_2 != entity.getSymbol("Actor")){ // if it is a sending statement the goal label has to be used just to add the message to IKWitness and it is not the label we use to create the goal location (then it is not a "real" goal
                                        this.itisarealgoal=false;
                                }else{
                                  this.itisarealgoal=true;
                                  //add temporary sender and receiver
                                  this.tmpSender=message_1;
                                  this.tmpReceiver=message_2;
                                }
                                //System.out.println("DEBUG realgoal="+this.itisarealgoal);

    jj_consume_token(15);
    message_3 = Term(entity);
    jj_consume_token(18);
                                          entity.setTempSender(null); entity.setTempReceiver(null);
        if(message_1 == entity.getSymbol("Actor") && message_2 != entity.getSymbol("Actor")) { // Sending statement
                MessageSet messageSet = new MessageSet();
                messageSet.add(message_3);
                if( entity.getTempActions().size() > 0 ) { System.out.println("Error - \u005c"?\u005c" can not appear in a sending statement."); System.exit(2); /*System.out.println("Errore sintattico: invio messaggio con '?' ");*/}
                        //MARCO
                        if (!this.thereisagoal){
                                //System.out.println("DEBUG IKUpdate");
                                this.thereisagoal=false;
                {if (true) return new IKUpdate(messageSet, "IK := IK + " + messageSet.toString());}
            }
                        else{
                //System.out.println("DEBUG IKWitness");
                this.thereisagoal=false;
        //    	if(!this.anotherRound){
                        this.anotherRound=true;

                        Iterator<Message> iter=messageSet.iterator();
                        Message[] tuple=new Message[3];
                        tuple[0]=message_1;
                        tuple[1]=message_2;
                        tuple[2]=iter.next();
                        this.witness=new FactAdd(witness_fact, new Tuple(tuple,"<"+tuple[0].toString()+", "+tuple[1].toString()+", "+tuple[2].toString()+">")," witness("+"<"+tuple[0].toString()+", "+tuple[1].toString()+", "+tuple[2].toString()+">"+")");
                        //Message[] w2w_entry=new Message[3];
                        //w2w_entry[0]=tuple[0];
                        //w2w_entry[1]=tuple[1];
                        //w2w_entry[2]=tuple[2];
                        //this.who2who.add(w2w_entry);
                        {if (true) return new IKUpdate(messageSet, "IK := IK + " + messageSet.toString());}//+" & witness ("+tuple.toString()+")");

          //    	}else{
           //   	  	this.anotherRound=false;
                        //Iterator<Message> iter=messageSet.iterator();
                //  	Message[] tuple=new Message[3];
              //	  	tuple[0]=message_1;
              //	  	tuple[1]=message_2;
              //	  	tuple[2]=iter.next();
                // 	return new FactAdd(witness_fact, new Tuple(tuple));
                //}
            }

            //MARCO
                //return new IKUpdate(messageSet, "IK := IK + " + messageSet.toString() );
        }
        else if (message_1 != entity.getSymbol("Actor") && message_2 == entity.getSymbol("Actor")) {if (true) return new IKConditional(message_3, "IK |- "+message_3.toString());} // Receiving statement
        else { System.out.println("Error - Wrong transmission statement: At least the sender or the receiver must be Actor."); System.exit(2); {if (true) return null;}}
    throw new Error("Missing return statement in function");
  }

/*
void ChannelProps(): 
{}
{								  // li consideriamo solo nei goal, 
	("*")? ("->" | "->>") ("*")? // assumiamo che gli scambi dei messaggi nei protocolli avvengano solo con '->'
}
*/

/**
 * Recognizes an entity name followed by its parameters.
 * @param	entity	The current entity.
 * @return	An ArrayList of objects that contains: The entity name in position 0; an ArrayList containing the entity instantiation actual parameters in position 1.
 */
  final public ArrayList<Object> Entity(Entity entity) throws ParseException {
        ArrayList<Message> terms = new ArrayList<Message>();
        ArrayList<Object> entityObjects = new ArrayList<Object>();
        Token token;
    token = jj_consume_token(UpperName);
    switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
    case 14:
      jj_consume_token(14);
      terms = Terms(entity);
      jj_consume_token(16);
      break;
    default:
      jj_la1[18] = jj_gen;
      ;
    }
                entityObjects.add(token.toString());
                entityObjects.add(terms);
                //System.out.println("Entity "+t.toString());
                //for(int i=0; i<al.size(); i++) System.out.println(al.get(i).print());
                {if (true) return entityObjects;}
    throw new Error("Missing return statement in function");
  }

/*
void IntroduceFact():
{}
{
	FuncApp() ";" 	// only for "iknows" and facts not in any clause head
}

void RetractFact():
{}
{
	"retract" FuncApp() ";"		// only for "iknows" and facts not in any clause head
}

void Branch():
{}
{
	"if" "(" GuardNoRcv() ")" Stmt() ( "else" Stmt() )? ";"
}

void Loop():
{}
{
	"while" "(" GuardNoRcv() ")" Stmt()
}

void Select():
{}
{
	"select" "{" ("on" "(" Guard() ")" ":" Stmt() )+ "}"	// Guard may include receive
}

void Assert():
{}
{
	"assert" GoalDecl() ";"
}

void ConstrDecls():
{}
{
	"constraints" ( ConstrDecl() ";" )+		// Only allowed in the root entity
}

void ConstrDecl():
{}
{
	Name() ":" Formula()
}
*/
/**
 * Recognizes a sequence of goal declarations.
 * @param	entity	The current entity.
 */
  final public void GoalDecls(Entity entity) throws ParseException {
    jj_consume_token(32);
    label_9:
    while (true) {
      GoalDecl(entity);
      jj_consume_token(18);
      switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
      case UpperName:
      case LowerName:
        ;
        break;
      default:
        jj_la1[19] = jj_gen;
        break label_9;
      }
    }
  }

/**
 * Recognizes a goal declaration.
 * @param	entity	The current entity.
 */
  final public void GoalDecl(Entity entity) throws ParseException {
    if (jj_2_17(2147483647)) {
      ChanGoal(entity);
    } else if (jj_2_18(2147483647)) {
      SecrGoal(entity);
    } else {
      jj_consume_token(-1);
      throw new ParseException();
    }
  }

/**
 * Recognizes a channel goal declaration.
 * @param	entity	The current entity.
 */
  final public void ChanGoal(Entity entity) throws ParseException {
  String name; // Represents the goal label.
  Message term_1, term_2; // Represent the sender and the receiver.
  ArrayList<Message> agentTerms = new ArrayList<Message>();
    name = Name();
    jj_consume_token(33);
    term_1 = Term(entity);
    jj_consume_token(34);
    term_2 = Term(entity);
                agentTerms.add(term_1);
                agentTerms.add(term_2);
                if(entity instanceof Session) { // If goals are declared in a Session entity
                        ((Session)entity).addToGoalMap(name, new Object[]{"auth", agentTerms}); // Keep track of the goal declaration by adding the goal name and agents.
                } else{ // If goals are not declared in a Session entity
                        //System.out.println("Errore: non e' possibile dichiarare un goal in un'entita' che non sia di tipo Session");
                        System.out.println("Error - Goals can only be declared in Session entities."); System.exit(2);
                }
  }

/**
 * Recognizes a secrecy goal declaration.
 * @param	entity	The current entity.
 */
  final public void SecrGoal(Entity entity) throws ParseException {
  String name; // Represents the goal label.
  Message term_1; // Represents a generic member of the confidentiality set.
  ArrayList<Message> confSet = new ArrayList<Message>();
    name = Name();
    jj_consume_token(33);
    jj_consume_token(11);
    label_10:
    while (true) {
      term_1 = Term(entity);
                                                                  confSet.add(term_1);
      label_11:
      while (true) {
        switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
        case 13:
          ;
          break;
        default:
          jj_la1[20] = jj_gen;
          break label_11;
        }
        jj_consume_token(13);
        term_1 = Term(entity);
                                                                                                                      confSet.add(term_1);
      }
      switch ((jj_ntk==-1)?jj_ntk():jj_ntk) {
      case 11:
      case 14:
      case 20:
      case 21:
      case UpperName:
      case LowerName:
        ;
        break;
      default:
        jj_la1[21] = jj_gen;
        break label_10;
      }
    }
    jj_consume_token(12);
                if(entity instanceof Session) { // If goals are declared in a Session entity.
                        ((Session)entity).addToGoalMap(name, new Object[]{"secr", confSet});
                } else{ // If goals are not declared in a Session entity
                        //System.out.println("Errore: non e' possibile dichiarare un goal in un'entita' che non sia di tipo Session");
                        System.out.println("Error - Goals can only be declared in Session entities."); System.exit(2);
                }
  }

  private boolean jj_2_1(int xla) {
    jj_la = xla; jj_lastpos = jj_scanpos = token;
    try { return !jj_3_1(); }
    catch(LookaheadSuccess ls) { return true; }
    finally { jj_save(0, xla); }
  }

  private boolean jj_2_2(int xla) {
    jj_la = xla; jj_lastpos = jj_scanpos = token;
    try { return !jj_3_2(); }
    catch(LookaheadSuccess ls) { return true; }
    finally { jj_save(1, xla); }
  }

  private boolean jj_2_3(int xla) {
    jj_la = xla; jj_lastpos = jj_scanpos = token;
    try { return !jj_3_3(); }
    catch(LookaheadSuccess ls) { return true; }
    finally { jj_save(2, xla); }
  }

  private boolean jj_2_4(int xla) {
    jj_la = xla; jj_lastpos = jj_scanpos = token;
    try { return !jj_3_4(); }
    catch(LookaheadSuccess ls) { return true; }
    finally { jj_save(3, xla); }
  }

  private boolean jj_2_5(int xla) {
    jj_la = xla; jj_lastpos = jj_scanpos = token;
    try { return !jj_3_5(); }
    catch(LookaheadSuccess ls) { return true; }
    finally { jj_save(4, xla); }
  }

  private boolean jj_2_6(int xla) {
    jj_la = xla; jj_lastpos = jj_scanpos = token;
    try { return !jj_3_6(); }
    catch(LookaheadSuccess ls) { return true; }
    finally { jj_save(5, xla); }
  }

  private boolean jj_2_7(int xla) {
    jj_la = xla; jj_lastpos = jj_scanpos = token;
    try { return !jj_3_7(); }
    catch(LookaheadSuccess ls) { return true; }
    finally { jj_save(6, xla); }
  }

  private boolean jj_2_8(int xla) {
    jj_la = xla; jj_lastpos = jj_scanpos = token;
    try { return !jj_3_8(); }
    catch(LookaheadSuccess ls) { return true; }
    finally { jj_save(7, xla); }
  }

  private boolean jj_2_9(int xla) {
    jj_la = xla; jj_lastpos = jj_scanpos = token;
    try { return !jj_3_9(); }
    catch(LookaheadSuccess ls) { return true; }
    finally { jj_save(8, xla); }
  }

  private boolean jj_2_10(int xla) {
    jj_la = xla; jj_lastpos = jj_scanpos = token;
    try { return !jj_3_10(); }
    catch(LookaheadSuccess ls) { return true; }
    finally { jj_save(9, xla); }
  }

  private boolean jj_2_11(int xla) {
    jj_la = xla; jj_lastpos = jj_scanpos = token;
    try { return !jj_3_11(); }
    catch(LookaheadSuccess ls) { return true; }
    finally { jj_save(10, xla); }
  }

  private boolean jj_2_12(int xla) {
    jj_la = xla; jj_lastpos = jj_scanpos = token;
    try { return !jj_3_12(); }
    catch(LookaheadSuccess ls) { return true; }
    finally { jj_save(11, xla); }
  }

  private boolean jj_2_13(int xla) {
    jj_la = xla; jj_lastpos = jj_scanpos = token;
    try { return !jj_3_13(); }
    catch(LookaheadSuccess ls) { return true; }
    finally { jj_save(12, xla); }
  }

  private boolean jj_2_14(int xla) {
    jj_la = xla; jj_lastpos = jj_scanpos = token;
    try { return !jj_3_14(); }
    catch(LookaheadSuccess ls) { return true; }
    finally { jj_save(13, xla); }
  }

  private boolean jj_2_15(int xla) {
    jj_la = xla; jj_lastpos = jj_scanpos = token;
    try { return !jj_3_15(); }
    catch(LookaheadSuccess ls) { return true; }
    finally { jj_save(14, xla); }
  }

  private boolean jj_2_16(int xla) {
    jj_la = xla; jj_lastpos = jj_scanpos = token;
    try { return !jj_3_16(); }
    catch(LookaheadSuccess ls) { return true; }
    finally { jj_save(15, xla); }
  }

  private boolean jj_2_17(int xla) {
    jj_la = xla; jj_lastpos = jj_scanpos = token;
    try { return !jj_3_17(); }
    catch(LookaheadSuccess ls) { return true; }
    finally { jj_save(16, xla); }
  }

  private boolean jj_2_18(int xla) {
    jj_la = xla; jj_lastpos = jj_scanpos = token;
    try { return !jj_3_18(); }
    catch(LookaheadSuccess ls) { return true; }
    finally { jj_save(17, xla); }
  }

  private boolean jj_3R_36() {
    if (jj_3R_15()) return true;
    return false;
  }

  private boolean jj_3R_21() {
    if (jj_3R_12()) return true;
    if (jj_scan_token(33)) return true;
    if (jj_scan_token(11)) return true;
    Token xsp;
    if (jj_3R_38()) return true;
    while (true) {
      xsp = jj_scanpos;
      if (jj_3R_38()) { jj_scanpos = xsp; break; }
    }
    if (jj_scan_token(12)) return true;
    return false;
  }

  private boolean jj_3R_51() {
    if (jj_scan_token(30)) return true;
    if (jj_3R_53()) return true;
    if (jj_scan_token(18)) return true;
    return false;
  }

  private boolean jj_3R_29() {
    if (jj_scan_token(11)) return true;
    if (jj_3R_13()) return true;
    if (jj_scan_token(23)) return true;
    if (jj_3R_13()) return true;
    if (jj_3R_39()) return true;
    return false;
  }

  private boolean jj_3_6() {
    if (jj_scan_token(20)) return true;
    return false;
  }

  private boolean jj_3R_35() {
    if (jj_3R_12()) return true;
    if (jj_scan_token(19)) return true;
    if (jj_3R_15()) return true;
    if (jj_scan_token(16)) return true;
    return false;
  }

  private boolean jj_3R_28() {
    if (jj_scan_token(21)) return true;
    if (jj_3R_13()) return true;
    if (jj_scan_token(22)) return true;
    if (jj_3R_13()) return true;
    if (jj_3R_39()) return true;
    return false;
  }

  private boolean jj_3R_18() {
    if (jj_3R_13()) return true;
    if (jj_scan_token(31)) return true;
    if (jj_3R_13()) return true;
    if (jj_scan_token(15)) return true;
    if (jj_3R_13()) return true;
    if (jj_scan_token(18)) return true;
    return false;
  }

  private boolean jj_3R_17() {
    Token xsp;
    xsp = jj_scanpos;
    if (jj_3R_35()) {
    jj_scanpos = xsp;
    if (jj_3R_36()) return true;
    }
    if (jj_scan_token(27)) return true;
    if (jj_3R_13()) return true;
    if (jj_scan_token(18)) return true;
    return false;
  }

  private boolean jj_3R_14() {
    if (jj_scan_token(LowerName)) return true;
    if (jj_scan_token(14)) return true;
    if (jj_3R_32()) return true;
    if (jj_scan_token(16)) return true;
    return false;
  }

  private boolean jj_3R_47() {
    if (jj_scan_token(13)) return true;
    if (jj_3R_13()) return true;
    return false;
  }

  private boolean jj_3R_19() {
    if (jj_3R_37()) return true;
    return false;
  }

  private boolean jj_3R_15() {
    if (jj_scan_token(UpperName)) return true;
    return false;
  }

  private boolean jj_3R_27() {
    if (jj_scan_token(20)) return true;
    return false;
  }

  private boolean jj_3R_53() {
    if (jj_scan_token(UpperName)) return true;
    Token xsp;
    xsp = jj_scanpos;
    if (jj_3R_54()) jj_scanpos = xsp;
    return false;
  }

  private boolean jj_3_5() {
    if (jj_scan_token(20)) return true;
    if (jj_3R_15()) return true;
    return false;
  }

  private boolean jj_3_12() {
    if (jj_scan_token(11)) return true;
    Token xsp;
    while (true) {
      xsp = jj_scanpos;
      if (jj_3R_19()) { jj_scanpos = xsp; break; }
    }
    if (jj_scan_token(12)) return true;
    return false;
  }

  private boolean jj_3_4() {
    if (jj_3R_15()) return true;
    return false;
  }

  private boolean jj_3R_52() {
    if (jj_3R_37()) return true;
    return false;
  }

  private boolean jj_3R_20() {
    if (jj_3R_12()) return true;
    if (jj_scan_token(33)) return true;
    if (jj_3R_13()) return true;
    if (jj_scan_token(34)) return true;
    if (jj_3R_13()) return true;
    return false;
  }

  private boolean jj_3_3() {
    if (jj_scan_token(LowerName)) return true;
    return false;
  }

  private boolean jj_3_2() {
    if (jj_3R_14()) return true;
    return false;
  }

  private boolean jj_3R_26() {
    if (jj_scan_token(20)) return true;
    if (jj_3R_15()) return true;
    if (jj_3R_39()) return true;
    return false;
  }

  private boolean jj_3R_46() {
    if (jj_scan_token(11)) return true;
    Token xsp;
    while (true) {
      xsp = jj_scanpos;
      if (jj_3R_52()) { jj_scanpos = xsp; break; }
    }
    if (jj_scan_token(12)) return true;
    return false;
  }

  private boolean jj_3R_25() {
    if (jj_3R_15()) return true;
    if (jj_3R_39()) return true;
    return false;
  }

  private boolean jj_3R_12() {
    Token xsp;
    xsp = jj_scanpos;
    if (jj_scan_token(42)) {
    jj_scanpos = xsp;
    if (jj_scan_token(41)) return true;
    }
    return false;
  }

  private boolean jj_3R_24() {
    if (jj_scan_token(LowerName)) return true;
    if (jj_3R_39()) return true;
    return false;
  }

  private boolean jj_3R_23() {
    if (jj_3R_14()) return true;
    if (jj_3R_39()) return true;
    return false;
  }

  private boolean jj_3_18() {
    if (jj_3R_21()) return true;
    return false;
  }

  private boolean jj_3_11() {
    if (jj_3R_18()) return true;
    return false;
  }

  private boolean jj_3R_40() {
    if (jj_scan_token(13)) return true;
    if (jj_3R_13()) return true;
    return false;
  }

  private boolean jj_3R_50() {
    if (jj_scan_token(29)) return true;
    if (jj_3R_53()) return true;
    if (jj_scan_token(18)) return true;
    return false;
  }

  private boolean jj_3R_32() {
    if (jj_3R_13()) return true;
    Token xsp;
    while (true) {
      xsp = jj_scanpos;
      if (jj_3R_40()) { jj_scanpos = xsp; break; }
    }
    return false;
  }

  private boolean jj_3R_45() {
    if (jj_3R_18()) return true;
    return false;
  }

  private boolean jj_3_10() {
    if (jj_3R_17()) return true;
    return false;
  }

  private boolean jj_3R_44() {
    if (jj_3R_51()) return true;
    return false;
  }

  private boolean jj_3R_43() {
    if (jj_3R_50()) return true;
    return false;
  }

  private boolean jj_3_16() {
    if (jj_3R_15()) return true;
    return false;
  }

  private boolean jj_3R_38() {
    if (jj_3R_13()) return true;
    Token xsp;
    while (true) {
      xsp = jj_scanpos;
      if (jj_3R_47()) { jj_scanpos = xsp; break; }
    }
    return false;
  }

  private boolean jj_3_9() {
    if (jj_3R_16()) return true;
    return false;
  }

  private boolean jj_3R_42() {
    if (jj_3R_17()) return true;
    return false;
  }

  private boolean jj_3R_49() {
    return false;
  }

  private boolean jj_3_15() {
    if (jj_3R_12()) return true;
    if (jj_scan_token(19)) return true;
    if (jj_3R_15()) return true;
    if (jj_scan_token(16)) return true;
    return false;
  }

  private boolean jj_3R_34() {
    if (jj_3R_15()) return true;
    return false;
  }

  private boolean jj_3_1() {
    if (jj_3R_12()) return true;
    if (jj_scan_token(19)) return true;
    if (jj_3R_13()) return true;
    if (jj_scan_token(16)) return true;
    return false;
  }

  private boolean jj_3R_48() {
    if (jj_scan_token(25)) return true;
    if (jj_3R_13()) return true;
    if (jj_3R_39()) return true;
    return false;
  }

  private boolean jj_3R_39() {
    Token xsp;
    xsp = jj_scanpos;
    if (jj_3R_48()) {
    jj_scanpos = xsp;
    if (jj_3R_49()) return true;
    }
    return false;
  }

  private boolean jj_3R_41() {
    if (jj_3R_16()) return true;
    return false;
  }

  private boolean jj_3R_37() {
    Token xsp;
    xsp = jj_scanpos;
    if (jj_3R_41()) {
    jj_scanpos = xsp;
    if (jj_3R_42()) {
    jj_scanpos = xsp;
    if (jj_3R_43()) {
    jj_scanpos = xsp;
    if (jj_3R_44()) {
    jj_scanpos = xsp;
    if (jj_3R_45()) {
    jj_scanpos = xsp;
    if (jj_3R_46()) return true;
    }
    }
    }
    }
    }
    return false;
  }

  private boolean jj_3R_33() {
    if (jj_3R_12()) return true;
    if (jj_scan_token(19)) return true;
    if (jj_3R_15()) return true;
    if (jj_scan_token(16)) return true;
    return false;
  }

  private boolean jj_3_8() {
    if (jj_scan_token(11)) return true;
    if (jj_3R_13()) return true;
    if (jj_scan_token(24)) return true;
    if (jj_3R_13()) return true;
    return false;
  }

  private boolean jj_3R_16() {
    Token xsp;
    xsp = jj_scanpos;
    if (jj_3R_33()) {
    jj_scanpos = xsp;
    if (jj_3R_34()) return true;
    }
    if (jj_scan_token(27)) return true;
    if (jj_scan_token(28)) return true;
    if (jj_scan_token(18)) return true;
    return false;
  }

  private boolean jj_3R_13() {
    Token xsp;
    xsp = jj_scanpos;
    if (jj_3R_22()) {
    jj_scanpos = xsp;
    if (jj_3R_23()) {
    jj_scanpos = xsp;
    if (jj_3R_24()) {
    jj_scanpos = xsp;
    if (jj_3R_25()) {
    jj_scanpos = xsp;
    if (jj_3R_26()) {
    jj_scanpos = xsp;
    if (jj_3R_27()) {
    jj_scanpos = xsp;
    if (jj_3R_28()) {
    jj_scanpos = xsp;
    if (jj_3R_29()) {
    jj_scanpos = xsp;
    if (jj_3R_30()) {
    jj_scanpos = xsp;
    if (jj_3R_31()) return true;
    }
    }
    }
    }
    }
    }
    }
    }
    }
    return false;
  }

  private boolean jj_3R_22() {
    if (jj_3R_12()) return true;
    if (jj_scan_token(19)) return true;
    if (jj_3R_13()) return true;
    if (jj_scan_token(16)) return true;
    if (jj_3R_39()) return true;
    return false;
  }

  private boolean jj_3R_31() {
    if (jj_scan_token(14)) return true;
    if (jj_3R_13()) return true;
    if (jj_scan_token(16)) return true;
    if (jj_3R_39()) return true;
    return false;
  }

  private boolean jj_3_14() {
    if (jj_3R_15()) return true;
    return false;
  }

  private boolean jj_3_17() {
    if (jj_3R_20()) return true;
    return false;
  }

  private boolean jj_3_7() {
    if (jj_scan_token(11)) return true;
    if (jj_3R_13()) return true;
    if (jj_scan_token(23)) return true;
    if (jj_3R_13()) return true;
    return false;
  }

  private boolean jj_3R_54() {
    if (jj_scan_token(14)) return true;
    if (jj_3R_32()) return true;
    if (jj_scan_token(16)) return true;
    return false;
  }

  private boolean jj_3R_30() {
    if (jj_scan_token(11)) return true;
    if (jj_3R_13()) return true;
    if (jj_scan_token(24)) return true;
    if (jj_3R_13()) return true;
    if (jj_3R_39()) return true;
    return false;
  }

  private boolean jj_3_13() {
    if (jj_3R_12()) return true;
    if (jj_scan_token(19)) return true;
    return false;
  }

  /** Generated Token Manager. */
  public ParserTokenManager token_source;
  SimpleCharStream jj_input_stream;
  /** Current token. */
  public Token token;
  /** Next token. */
  public Token jj_nt;
  private int jj_ntk;
  private Token jj_scanpos, jj_lastpos;
  private int jj_la;
  private int jj_gen;
  final private int[] jj_la1 = new int[22];
  static private int[] jj_la1_0;
  static private int[] jj_la1_1;
  static {
      jj_la1_init_0();
      jj_la1_init_1();
   }
   private static void jj_la1_init_0() {
      jj_la1_0 = new int[] {0x4000,0x20000,0x400,0x4000000,0x0,0x0,0x2000,0x2000,0x4000,0x0,0x2000,0x0,0x200000,0x4000,0x2000000,0x2000,0x60304800,0x60000000,0x4000,0x0,0x2000,0x304800,};
   }
   private static void jj_la1_init_1() {
      jj_la1_1 = new int[] {0x0,0x0,0x0,0x0,0x1,0x600,0x0,0x0,0x400,0x600,0x0,0x600,0x0,0x0,0x0,0x0,0x600,0x0,0x0,0x600,0x0,0x600,};
   }
  final private JJCalls[] jj_2_rtns = new JJCalls[18];
  private boolean jj_rescan = false;
  private int jj_gc = 0;

  /** Constructor with InputStream. */
  public Parser(java.io.InputStream stream) {
     this(stream, null);
  }
  /** Constructor with InputStream and supplied encoding */
  public Parser(java.io.InputStream stream, String encoding) {
    try { jj_input_stream = new SimpleCharStream(stream, encoding, 1, 1); } catch(java.io.UnsupportedEncodingException e) { throw new RuntimeException(e); }
    token_source = new ParserTokenManager(jj_input_stream);
    token = new Token();
    jj_ntk = -1;
    jj_gen = 0;
    for (int i = 0; i < 22; i++) jj_la1[i] = -1;
    for (int i = 0; i < jj_2_rtns.length; i++) jj_2_rtns[i] = new JJCalls();
  }

  /** Reinitialise. */
  public void ReInit(java.io.InputStream stream) {
     ReInit(stream, null);
  }
  /** Reinitialise. */
  public void ReInit(java.io.InputStream stream, String encoding) {
    try { jj_input_stream.ReInit(stream, encoding, 1, 1); } catch(java.io.UnsupportedEncodingException e) { throw new RuntimeException(e); }
    token_source.ReInit(jj_input_stream);
    token = new Token();
    jj_ntk = -1;
    jj_gen = 0;
    for (int i = 0; i < 22; i++) jj_la1[i] = -1;
    for (int i = 0; i < jj_2_rtns.length; i++) jj_2_rtns[i] = new JJCalls();
  }

  /** Constructor. */
  public Parser(java.io.Reader stream) {
    jj_input_stream = new SimpleCharStream(stream, 1, 1);
    token_source = new ParserTokenManager(jj_input_stream);
    token = new Token();
    jj_ntk = -1;
    jj_gen = 0;
    for (int i = 0; i < 22; i++) jj_la1[i] = -1;
    for (int i = 0; i < jj_2_rtns.length; i++) jj_2_rtns[i] = new JJCalls();
  }

  /** Reinitialise. */
  public void ReInit(java.io.Reader stream) {
    jj_input_stream.ReInit(stream, 1, 1);
    token_source.ReInit(jj_input_stream);
    token = new Token();
    jj_ntk = -1;
    jj_gen = 0;
    for (int i = 0; i < 22; i++) jj_la1[i] = -1;
    for (int i = 0; i < jj_2_rtns.length; i++) jj_2_rtns[i] = new JJCalls();
  }

  /** Constructor with generated Token Manager. */
  public Parser(ParserTokenManager tm) {
    token_source = tm;
    token = new Token();
    jj_ntk = -1;
    jj_gen = 0;
    for (int i = 0; i < 22; i++) jj_la1[i] = -1;
    for (int i = 0; i < jj_2_rtns.length; i++) jj_2_rtns[i] = new JJCalls();
  }

  /** Reinitialise. */
  public void ReInit(ParserTokenManager tm) {
    token_source = tm;
    token = new Token();
    jj_ntk = -1;
    jj_gen = 0;
    for (int i = 0; i < 22; i++) jj_la1[i] = -1;
    for (int i = 0; i < jj_2_rtns.length; i++) jj_2_rtns[i] = new JJCalls();
  }

  private Token jj_consume_token(int kind) throws ParseException {
    Token oldToken;
    if ((oldToken = token).next != null) token = token.next;
    else token = token.next = token_source.getNextToken();
    jj_ntk = -1;
    if (token.kind == kind) {
      jj_gen++;
      if (++jj_gc > 100) {
        jj_gc = 0;
        for (int i = 0; i < jj_2_rtns.length; i++) {
          JJCalls c = jj_2_rtns[i];
          while (c != null) {
            if (c.gen < jj_gen) c.first = null;
            c = c.next;
          }
        }
      }
      return token;
    }
    token = oldToken;
    jj_kind = kind;
    throw generateParseException();
  }

  static private final class LookaheadSuccess extends java.lang.Error { }
  final private LookaheadSuccess jj_ls = new LookaheadSuccess();
  private boolean jj_scan_token(int kind) {
    if (jj_scanpos == jj_lastpos) {
      jj_la--;
      if (jj_scanpos.next == null) {
        jj_lastpos = jj_scanpos = jj_scanpos.next = token_source.getNextToken();
      } else {
        jj_lastpos = jj_scanpos = jj_scanpos.next;
      }
    } else {
      jj_scanpos = jj_scanpos.next;
    }
    if (jj_rescan) {
      int i = 0; Token tok = token;
      while (tok != null && tok != jj_scanpos) { i++; tok = tok.next; }
      if (tok != null) jj_add_error_token(kind, i);
    }
    if (jj_scanpos.kind != kind) return true;
    if (jj_la == 0 && jj_scanpos == jj_lastpos) throw jj_ls;
    return false;
  }


/** Get the next Token. */
  final public Token getNextToken() {
    if (token.next != null) token = token.next;
    else token = token.next = token_source.getNextToken();
    jj_ntk = -1;
    jj_gen++;
    return token;
  }

/** Get the specific Token. */
  final public Token getToken(int index) {
    Token t = token;
    for (int i = 0; i < index; i++) {
      if (t.next != null) t = t.next;
      else t = t.next = token_source.getNextToken();
    }
    return t;
  }

  private int jj_ntk() {
    if ((jj_nt=token.next) == null)
      return (jj_ntk = (token.next=token_source.getNextToken()).kind);
    else
      return (jj_ntk = jj_nt.kind);
  }

  private java.util.List<int[]> jj_expentries = new java.util.ArrayList<int[]>();
  private int[] jj_expentry;
  private int jj_kind = -1;
  private int[] jj_lasttokens = new int[100];
  private int jj_endpos;

  private void jj_add_error_token(int kind, int pos) {
    if (pos >= 100) return;
    if (pos == jj_endpos + 1) {
      jj_lasttokens[jj_endpos++] = kind;
    } else if (jj_endpos != 0) {
      jj_expentry = new int[jj_endpos];
      for (int i = 0; i < jj_endpos; i++) {
        jj_expentry[i] = jj_lasttokens[i];
      }
      jj_entries_loop: for (java.util.Iterator<?> it = jj_expentries.iterator(); it.hasNext();) {
        int[] oldentry = (int[])(it.next());
        if (oldentry.length == jj_expentry.length) {
          for (int i = 0; i < jj_expentry.length; i++) {
            if (oldentry[i] != jj_expentry[i]) {
              continue jj_entries_loop;
            }
          }
          jj_expentries.add(jj_expentry);
          break jj_entries_loop;
        }
      }
      if (pos != 0) jj_lasttokens[(jj_endpos = pos) - 1] = kind;
    }
  }

  /** Generate ParseException. */
  public ParseException generateParseException() {
    jj_expentries.clear();
    boolean[] la1tokens = new boolean[43];
    if (jj_kind >= 0) {
      la1tokens[jj_kind] = true;
      jj_kind = -1;
    }
    for (int i = 0; i < 22; i++) {
      if (jj_la1[i] == jj_gen) {
        for (int j = 0; j < 32; j++) {
          if ((jj_la1_0[i] & (1<<j)) != 0) {
            la1tokens[j] = true;
          }
          if ((jj_la1_1[i] & (1<<j)) != 0) {
            la1tokens[32+j] = true;
          }
        }
      }
    }
    for (int i = 0; i < 43; i++) {
      if (la1tokens[i]) {
        jj_expentry = new int[1];
        jj_expentry[0] = i;
        jj_expentries.add(jj_expentry);
      }
    }
    jj_endpos = 0;
    jj_rescan_token();
    jj_add_error_token(0, 0);
    int[][] exptokseq = new int[jj_expentries.size()][];
    for (int i = 0; i < jj_expentries.size(); i++) {
      exptokseq[i] = jj_expentries.get(i);
    }
    return new ParseException(token, exptokseq, tokenImage);
  }

  /** Enable tracing. */
  final public void enable_tracing() {
  }

  /** Disable tracing. */
  final public void disable_tracing() {
  }

  private void jj_rescan_token() {
    jj_rescan = true;
    for (int i = 0; i < 18; i++) {
    try {
      JJCalls p = jj_2_rtns[i];
      do {
        if (p.gen > jj_gen) {
          jj_la = p.arg; jj_lastpos = jj_scanpos = p.first;
          switch (i) {
            case 0: jj_3_1(); break;
            case 1: jj_3_2(); break;
            case 2: jj_3_3(); break;
            case 3: jj_3_4(); break;
            case 4: jj_3_5(); break;
            case 5: jj_3_6(); break;
            case 6: jj_3_7(); break;
            case 7: jj_3_8(); break;
            case 8: jj_3_9(); break;
            case 9: jj_3_10(); break;
            case 10: jj_3_11(); break;
            case 11: jj_3_12(); break;
            case 12: jj_3_13(); break;
            case 13: jj_3_14(); break;
            case 14: jj_3_15(); break;
            case 15: jj_3_16(); break;
            case 16: jj_3_17(); break;
            case 17: jj_3_18(); break;
          }
        }
        p = p.next;
      } while (p != null);
      } catch(LookaheadSuccess ls) { }
    }
    jj_rescan = false;
  }

  private void jj_save(int index, int xla) {
    JJCalls p = jj_2_rtns[index];
    while (p.gen > jj_gen) {
      if (p.next == null) { p = p.next = new JJCalls(); break; }
      p = p.next;
    }
    p.gen = jj_gen + xla - jj_la; p.first = token; p.arg = xla;
  }

  static final class JJCalls {
    int gen;
    Token first;
    int arg;
    JJCalls next;
  }

}
