package translator;
import java.util.ArrayList;
import java.util.HashMap;

import se.Message;

/**
 * Class representing a symbolic session entity.
 * 
 * @author Fabio Pettenuzzo
 * @author Marco Palamà
 * @version 1.0, 12/13
 */
public class Session extends Entity {	
	
	
	private HashMap<String, ArrayList<String>> signatures;
	/** Maps a goal name to an object containing:
	 * The type in position 0: "auth" for authentication, "secr" for confidentiality.
	 * The goal parameters.
	 */
	private HashMap<String, Object[]> goalMap;
	
	/**
	 * Constructor of the Session class.
	 * @param parent The parent entity, this value should be instance of Environment
	 * @param name The session entity name
	 */
	Session(Entity parent, String name){
		super(parent, name);
		if(!(parent instanceof Environment)) { System.out.println("Error - Entity "+ getName() +": Wrong nesting."); System.exit(3); }
		goalMap = new HashMap<String, Object[]>();
		
	}

	public HashMap<String, ArrayList<String>> getSignatures() {
		return signatures;
	}

	/**
	 * Adds a mapping between a goal name and its parameters.
	 * @param name The goal name
	 * @param parameters The goal parameters
	 */
	public void addToGoalMap(String name, Object[] parameters){
		goalMap.put(name, parameters);
	}
	
	/**
	 * Getter of goalMap
	 * @return goalMap
	 */
	public HashMap<String, Object[]> getGoalMap(){
		return goalMap;
	}
	
	/**
	 * Returns the type of a goal given its name.
	 * @param name The goal name
	 * @return The type of the goal 
	 */
	public String getSessionGoalType(String name){
		if(goalMap.get(name)[0] instanceof String)
			return (String) goalMap.get(name)[0];
		else { System.out.println("Error - The session goal type "+ goalMap.get(name)[0] +" is not an instance of String"); System.exit(3); }
			return null;
	}
	
	/**
	 * Returns the parameters of a goal given its name.
	 * @param name The goal name
	 * @return The parameters of the goal 
	 */
	public ArrayList<Message> getSessionGoalParameters(String name){
		if(goalMap.get(name)[1] instanceof ArrayList<?>)
			return (ArrayList<Message>) goalMap.get(name)[1];
		else { System.out.println("Error - The session goal type "+ goalMap.get(name)[1] +" is not an instance of ArrayList<?>"); System.exit(3); }
			return null;
	}
	
	/**
	 * Setter of signatures
	 * @param signatures signatures
	 */
	public void setSignatures(HashMap<String, ArrayList<String>> signatures) {
		this.signatures = signatures;
	}
	
	
}
